Package: bartMachine
Type: Package
Title: Bayesian Additive Regression Trees
Version: 1.4.2
Date: 2026-05-03
Maintainer: Adam Kapelner <kapelner@qc.cuny.edu>
Authors@R: c(person(given = "Adam",
                  family = "Kapelner",
                  role = c("aut", "cre"),
                  email = "kapelner@qc.cuny.edu",
                  comment = c(ORCID = "0000-0001-5985-6792")),
             person(given = "Justin",
                      family = "Bleich",
                      role = "aut"))
Description: An advanced implementation of Bayesian Additive Regression Trees with expanded features for data analysis and visualization.
License: GPL-3
Encoding: UTF-8
Depends: R (>= 2.14.0)
Imports: bartMachineJARs (>= 1.2.2), rJava, randomForest, missForest,
        stats, ggplot2, grid, checkmate, matrixStats, grDevices
Suggests: testthat (>= 3.0.0), withr
SystemRequirements: Java (>= 21.0), TornadoVM (if you want GPU
        optimization)
BugReports: https://github.com/kapelner/bartMachine/issues
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-03 21:07:08 UTC; kapelner
Author: Adam Kapelner [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5985-6792>),
  Justin Bleich [aut]
Repository: CRAN
Date/Publication: 2026-05-04 05:20:02 UTC
Built: R 4.5.2; ; 2026-05-04 05:58:50 UTC; unix

Package: bartMachine
Type: Package
Title: Bayesian Additive Regression Trees
Version: 1.4.2
Date: 2026-05-03
Maintainer: Adam Kapelner <kapelner@qc.cuny.edu>
Authors@R: c(person(given = "Adam",
                  family = "Kapelner",
                  role = c("aut", "cre"),
                  email = "kapelner@qc.cuny.edu",
                  comment = c(ORCID = "0000-0001-5985-6792")),
             person(given = "Justin",
                      family = "Bleich",
                      role = "aut"))
Description: An advanced implementation of Bayesian Additive Regression Trees with expanded features for data analysis and visualization.
License: GPL-3
Encoding: UTF-8
Depends: R (>= 2.14.0)
Imports: bartMachineJARs (>= 1.2.2), rJava, randomForest, missForest,
        stats, ggplot2, grid, checkmate, matrixStats, grDevices
Suggests: testthat (>= 3.0.0), withr
SystemRequirements: Java (>= 21.0), TornadoVM (if you want GPU
        optimization)
BugReports: https://github.com/kapelner/bartMachine/issues
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-03 21:07:08 UTC; kapelner
Author: Adam Kapelner [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5985-6792>),
  Justin Bleich [aut]
Repository: CRAN
Date/Publication: 2026-05-04 05:20:02 UTC
Built: R 4.5.2; ; 2026-05-04 07:47:56 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(predict,bartMachine)
S3method(print,bartMachine)
S3method(summary,bartMachine)
export(bartMachine)
export(bartMachineArr)
export(bartMachineCV)
export(bart_machine_get_posterior)
export(bart_machine_num_cores)
export(bart_predict_for_test_data)
export(build_bart_machine)
export(build_bart_machine_cv)
export(calc_credible_intervals)
export(calc_prediction_intervals)
export(check_bart_error_assumptions)
export(cov_importance_test)
export(dummify_data)
export(extract_raw_node_data)
export(get_projection_weights)
export(get_sigsqs)
export(get_var_counts_over_chain)
export(get_var_props_over_chain)
export(interaction_investigator)
export(investigate_var_importance)
export(k_fold_cv)
export(linearity_test)
export(node_prediction_training_data_indices)
export(pd_plot)
export(plot_convergence_diagnostics)
export(plot_y_vs_yhat)
export(predict_bartMachineArr)
export(rmse_by_num_trees)
export(set_bart_machine_num_cores)
export(var_selection_by_permute)
export(var_selection_by_permute_cv)
import(bartMachineJARs)
import(checkmate)
import(grid)
import(missForest)
import(randomForest)
import(stats)
importFrom(grDevices,rgb)
importFrom(rJava,.jarray)
importFrom(rJava,.jcache)
importFrom(rJava,.jcall)
importFrom(rJava,.jcast)
importFrom(rJava,.jevalArray)
importFrom(rJava,.jnew)
importFrom(rJava,.jnull)
importFrom(rJava,.jpackage)
importFrom(rJava,is.jnull)
importFrom(utils,packageVersion)
importFrom(utils,tail)

# Benchmark Suite: bartMachine vs randomForest

This benchmark suite runs repeated K-fold cross-validation and compares
out-of-sample performance between `bartMachine` and `randomForest` across
datasets drawn from standard benchmark libraries (e.g., `datasets`, `MASS`,
`mlbench`, `ISLR`).

## Quick start

1. Install the package and optional dataset libraries:

```r
install.packages(c("bartMachine", "randomForest", "mlbench", "ISLR", "pROC"))
```

2. Run the suite from the package root:

```bash
Rscript inst/benchmarks/run_benchmark_suite.R --folds=5 --repeats=3
```

3. Outputs are written to `inst/benchmarks/results`:
   - `benchmark_folds.csv`: per-fold metrics
   - `benchmark_summary.csv`: mean and standard deviation by dataset/model
   - `benchmark_skipped.csv`: datasets that were skipped and why
   - `benchmark_results.rds`: all results + config
   - `sessionInfo.txt`: session metadata

## Notes

- Set Java memory and other parameters *before* loading `bartMachine`. You can do this via:
  - `options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))`
- Use `--list` to view all dataset definitions and `--dry-run` to list the
  filtered selection without running.
- Large or slow datasets can be skipped with `--skip-tags=large`.
- `pROC` is optional; if installed, AUC is computed for classification tasks.

## Example filters

```bash
Rscript inst/benchmarks/run_benchmark_suite.R --packages=datasets,MASS --folds=3
Rscript inst/benchmarks/run_benchmark_suite.R --datasets=Boston,BostonHousing
Rscript inst/benchmarks/run_benchmark_suite.R --skip-tags=large --repeats=2
```

#!/usr/bin/env Rscript

options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))
suppressPackageStartupMessages({
  library(bartMachine)
})
set_bart_machine_num_cores(4)

plot_env <- asNamespace("bartMachine")
plot_source <- file.path("bartMachine", "R", "bart_package_plots.R")
if (file.exists(plot_source)) {
  plot_env <- new.env(parent = plot_env)
  sys.source(plot_source, envir = plot_env)
}
plot_fn <- function(name) {
  get(name, envir = plot_env, inherits = TRUE)
}

run_test <- function(name, expr) {
  message("Running ", name, "...")
  tryCatch(
    {
      expr
      TRUE
    },
    error = function(e) {
      stop(name, " failed: ", e$message)
    }
  )
}

set.seed(1984)
n <- 500
p <- 8
X <- data.frame(matrix(runif(n * p), ncol = p))
y <- 3 * X[, 1] + 2 * X[, 2] + rnorm(n, sd = 0.2)
num_replicates_for_avg = 4
num_trees_bottleneck = 10

bart_machine <- bartMachine(
  X,
  y,
  num_trees = 100,
  num_burn_in = 1000,
  num_iterations_after_burn_in = 1000,
  flush_indices_to_save_RAM = FALSE,
  verbose = FALSE
)

plot_file <- file.path("bartMachine", "inst", "scripts", "bartmachine_ggplot_tests.pdf")
grDevices::pdf(plot_file, width = 8, height = 6)
on.exit(grDevices::dev.off(), add = TRUE)

run_test("check_bart_error_assumptions", plot_fn("check_bart_error_assumptions")(bart_machine))
run_test("plot_y_vs_yhat_default", plot_fn("plot_y_vs_yhat")(bart_machine))
run_test("plot_y_vs_yhat_credible", plot_fn("plot_y_vs_yhat")(bart_machine, credible_intervals = TRUE))
run_test("get_sigsqs_plot", plot_fn("get_sigsqs")(bart_machine, plot_hist = TRUE))
run_test(
  "investigate_var_importance",
  plot_fn("investigate_var_importance")(
    bart_machine,
    plot = TRUE,
    num_replicates_for_avg = num_replicates_for_avg,
    num_trees_bottleneck = num_trees_bottleneck,
    num_var_plot = p
  )
)
run_test("plot_convergence_diagnostics", plot_fn("plot_convergence_diagnostics")(bart_machine))
run_test(
  "interaction_investigator",
  plot_fn("interaction_investigator")(
    bart_machine,
    plot = TRUE,
    num_replicates_for_avg = num_replicates_for_avg,
    num_trees_bottleneck = num_trees_bottleneck,
    num_var_plot = p
  )
)
run_test("pd_plot", plot_fn("pd_plot")(bart_machine, j = 1, prop_data = 0.3))
run_test("pd_plot", plot_fn("pd_plot")(bart_machine, j = 2, prop_data = 0.3))
run_test("pd_plot", plot_fn("pd_plot")(bart_machine, j = 3, prop_data = 0.3))
run_test(
  "rmse_by_num_trees",
  plot_fn("rmse_by_num_trees")(
    bart_machine,
    tree_list = c(10, 50),
    in_sample = TRUE,
    plot = TRUE,
    num_replicates = num_replicates_for_avg
  )
)

message("ggplot tests completed. Output PDF: ", plot_file)

benchmark_registry_version <- "1.1"

benchmark_registry <- list(
  list(name = "mtcars", package = "datasets", response = "mpg", task = "regression"),
  list(name = "airquality", package = "datasets", response = "Ozone", task = "regression"),
  list(name = "ToothGrowth", package = "datasets", response = "len", task = "regression"),
  list(name = "warpbreaks", package = "datasets", response = "breaks", task = "regression"),
  list(name = "swiss", package = "datasets", response = "Fertility", task = "regression"),
  list(name = "trees", package = "datasets", response = "Volume", task = "regression"),
  list(name = "faithful", package = "datasets", response = "eruptions", task = "regression"),
  list(name = "rock", package = "datasets", response = "perm", task = "regression"),
  list(name = "attitude", package = "datasets", response = "rating", task = "regression"),
  list(name = "USArrests", package = "datasets", response = "Murder", task = "regression"),
  list(name = "quakes", package = "datasets", response = "mag", task = "regression"),
  list(name = "CO2", package = "datasets", response = "uptake", task = "regression"),
  list(name = "infert", package = "datasets", response = "case", task = "classification"),
  list(name = "Boston", package = "MASS", response = "medv", task = "regression"),
  list(name = "Pima.tr", package = "MASS", response = "type", task = "classification"),
  list(name = "biopsy", package = "MASS", response = "class", task = "classification", drop = c("ID")),
  list(name = "birthwt", package = "MASS", response = "low", task = "classification"),
  list(name = "cats", package = "MASS", response = "Hwt", task = "regression"),
  list(name = "cement", package = "MASS", response = "y", task = "regression"),
  list(name = "cpus", package = "MASS", response = "perf", task = "regression", drop = c("name")),
  list(name = "Cars93", package = "MASS", response = "Price", task = "regression"),
  list(name = "BostonHousing", package = "mlbench", response = "medv", task = "regression"),
  list(name = "BostonHousing2", package = "mlbench", response = "medv", task = "regression"),
  list(name = "PimaIndiansDiabetes", package = "mlbench", response = "diabetes", task = "classification"),
  list(name = "PimaIndiansDiabetes2", package = "mlbench", response = "diabetes", task = "classification"),
  list(name = "BreastCancer", package = "mlbench", response = "Class", task = "classification", drop = c("Id")),
  list(name = "HouseVotes84", package = "mlbench", response = "Class", task = "classification"),
  list(name = "Ionosphere", package = "mlbench", response = "Class", task = "classification"),
  list(name = "Sonar", package = "mlbench", response = "Class", task = "classification"),
  list(name = "Servo", package = "mlbench", response = "Class", task = "regression"),
  list(name = "Auto", package = "ISLR", response = "mpg", task = "regression", drop = c("name")),
  list(name = "Carseats", package = "ISLR", response = "Sales", task = "regression"),
  list(name = "College", package = "ISLR", response = "Apps", task = "regression"),
  list(name = "Credit", package = "ISLR", response = "Balance", task = "regression", drop = c("ID")),
  list(name = "Hitters", package = "ISLR", response = "Salary", task = "regression", drop = c("Player")),
  list(name = "Portfolio", package = "ISLR", response = "Y", task = "regression"),
  list(name = "Wage", package = "ISLR", response = "wage", task = "regression"),
  list(name = "Default", package = "ISLR", response = "default", task = "classification"),
  list(name = "OJ", package = "ISLR", response = "Purchase", task = "classification"),
  list(name = "Weekly", package = "ISLR", response = "Direction", task = "classification"),
  list(name = "Smarket", package = "ISLR", response = "Direction", task = "classification"),
  list(name = "Caravan", package = "ISLR", response = "Purchase", task = "classification", tags = c("large"))
)

#!/usr/bin/env Rscript

#Rscript bartMachine/inst/benchmarks/run_benchmark_suite.R  --java-params="-Xmx10g"  --cores=4  --folds=5  --repeats=3


opts <- list(
  output_dir = file.path("inst", "benchmarks", "results"),
  folds = 5,
  repeats = 1,
  seed = 123,
  cores = 1,
  datasets = NULL,
  packages = NULL,
  max_datasets = Inf,
  skip_tags = character(0),
  list = FALSE,
  dry_run = FALSE,
  java_params = Sys.getenv("BART_JAVA_PARAMS", ""),
  use_xoshiro = FALSE,
  bart_num_trees = 200,
  bart_burn_in = 800,
  bart_iter = 500,
  rf_ntree = 200,
  rf_mtry = NA_real_
)

parse_args <- function(args) {
  opts <- list()
  for (arg in args) {
    if (arg == "--list") {
      opts$list <- TRUE
      next
    }
    if (arg == "--dry-run") {
      opts$dry_run <- TRUE
      next
    }
    if (arg == "--help") {
      opts$help <- TRUE
      next
    }
    if (arg == "--use-xoshiro") {
      opts$use_xoshiro <- TRUE
      next
    }
    if (grepl("^--", arg)) {
      key_val <- sub("^--", "", arg)
      parts <- strsplit(key_val, "=", fixed = TRUE)[[1]]
      key <- parts[1]
      val <- if (length(parts) > 1) parts[2] else ""
      opts[[key]] <- val
    }
  }
  opts
}

get_script_dir <- function() {
  args <- commandArgs(trailingOnly = FALSE)
  file_arg <- "--file="
  match <- grep(file_arg, args)
  if (length(match) == 0) {
    return(getwd())
  }
  file_path <- sub(file_arg, "", args[match[1]])
  dirname(normalizePath(file_path))
}

print_usage <- function() {
  cat("Usage: Rscript run_benchmark_suite.R [options]\n")
  cat("Options:\n")
  cat("  --list                    List available datasets and exit\n")
  cat("  --dry-run                 List selected datasets and exit\n")
  cat("  --output-dir=PATH         Output directory (default: inst/benchmarks/results)\n")
  cat("  --folds=N                 Number of folds (default: 5)\n")
  cat("  --repeats=N               Number of CV repeats (default: 1)\n")
  cat("  --seed=N                  Random seed (default: 123)\n")
  cat("  --datasets=A,B            Only run specific dataset names\n")
  cat("  --packages=A,B            Only run datasets from packages\n")
  cat("  --max-datasets=N          Limit the number of datasets\n")
  cat("  --skip-tags=A,B           Skip datasets with these tags\n")
  cat("  --cores=N                 bartMachine cores (default: 1)\n")
  cat("  --java-params=STRING      Set options(java.parameters)\n")
  cat("  --use-xoshiro             Use Xoshiro256PlusPlus RNG (JDK 17+)\n")
  cat("  --bart-num-trees=N        bartMachine num_trees\n")
  cat("  --bart-burn-in=N          bartMachine num_burn_in\n")
  cat("  --bart-iter=N             bartMachine num_iterations_after_burn_in\n")
  cat("  --rf-ntree=N              randomForest ntree\n")
  cat("  --rf-mtry=N               randomForest mtry\n")
}

as_int <- function(value, default) {
  if (is.null(value) || !nzchar(value)) {
    return(default)
  }
  as.integer(value)
}

as_char_vec <- function(value) {
  if (is.null(value) || !nzchar(value)) {
    return(NULL)
  }
  strsplit(value, ",", fixed = TRUE)[[1]]
}

script_dir <- get_script_dir()
registry_path <- file.path(script_dir, "benchmark_registry.R")
if (!file.exists(registry_path)) {
  stop("Cannot find benchmark registry at: ", registry_path)
}
invisible(source(registry_path, local = TRUE))

raw_opts <- parse_args(commandArgs(trailingOnly = TRUE))
if (isTRUE(raw_opts$help)) {
  print_usage()
  quit(status = 0)
}

if (!is.null(raw_opts$output_dir)) {
  opts$output_dir <- raw_opts$output_dir
}
opts$folds <- as_int(raw_opts$folds, opts$folds)
opts$repeats <- as_int(raw_opts$repeats, opts$repeats)
opts$seed <- as_int(raw_opts$seed, opts$seed)
opts$cores <- as_int(raw_opts$cores, opts$cores)
opts$datasets <- as_char_vec(raw_opts$datasets)
opts$packages <- as_char_vec(raw_opts$packages)
opts$max_datasets <- as_int(raw_opts$`max-datasets`, opts$max_datasets)
opts$skip_tags <- as_char_vec(raw_opts$`skip-tags`)
opts$java_params <- if (!is.null(raw_opts$`java-params`)) raw_opts$`java-params` else opts$java_params
opts$bart_num_trees <- as_int(raw_opts$`bart-num-trees`, opts$bart_num_trees)
opts$bart_burn_in <- as_int(raw_opts$`bart-burn-in`, opts$bart_burn_in)
opts$bart_iter <- as_int(raw_opts$`bart-iter`, opts$bart_iter)
opts$rf_ntree <- as_int(raw_opts$`rf-ntree`, opts$rf_ntree)
if (!is.null(raw_opts$`rf-mtry`) && nzchar(raw_opts$`rf-mtry`)) {
  opts$rf_mtry <- as.numeric(raw_opts$`rf-mtry`)
}
if (!is.null(raw_opts$use_xoshiro)) {
  opts$use_xoshiro <- isTRUE(raw_opts$use_xoshiro)
}
opts$list <- isTRUE(raw_opts$list)
opts$dry_run <- isTRUE(raw_opts$dry_run)

print_registry <- function(registry) {
  entries <- vapply(
    registry,
    function(item) paste(item$package, item$name, item$task, sep = " | "),
    character(1)
  )
  cat(paste(entries, collapse = "\n"), "\n")
}

registry <- benchmark_registry
if (!is.null(opts$datasets)) {
  registry <- Filter(function(item) item$name %in% opts$datasets, registry)
}
if (!is.null(opts$packages)) {
  registry <- Filter(function(item) item$package %in% opts$packages, registry)
}
if (!is.null(opts$skip_tags) && length(opts$skip_tags) > 0) {
  registry <- Filter(
    function(item) {
      tags <- item$tags
      if (is.null(tags)) {
        return(TRUE)
      }
      !any(tags %in% opts$skip_tags)
    },
    registry
  )
}
if (is.finite(opts$max_datasets)) {
  registry <- head(registry, opts$max_datasets)
}

if (opts$list) {
  print_registry(benchmark_registry)
  quit(status = 0)
}
if (opts$dry_run) {
  print_registry(registry)
  quit(status = 0)
}



if (nzchar(opts$java_params)) {
  options(java.parameters = c(opts$java_params, "--add-modules=jdk.incubator.vector"))
}

suppressPackageStartupMessages({
  if (!requireNamespace("bartMachine", quietly = TRUE)) {
    stop("bartMachine is not installed.")
  }
  if (!requireNamespace("randomForest", quietly = TRUE)) {
    stop("randomForest is not installed.")
  }
  library(bartMachine)
  library(randomForest)
})

set_bart_machine_num_cores(opts$cores, verbose = FALSE)

normalize_predictors <- function(df) {
  df[] <- lapply(df, function(col) {
    if (is.character(col) || is.logical(col)) {
      return(as.factor(col))
    }
    if (inherits(col, "ordered")) {
      return(factor(col, ordered = FALSE))
    }
    col
  })
  df
}

drop_unused_factor_levels <- function(df) {
  df[] <- lapply(df, function(col) {
    if (is.factor(col)) {
      return(droplevels(col))
    }
    col
  })
  df
}

align_factor_levels <- function(train_df, test_df, test_y) {
  drop_idx <- rep(FALSE, nrow(test_df))
  for (col_name in names(train_df)) {
    train_col <- train_df[[col_name]]
    test_col <- test_df[[col_name]]
    if (is.factor(train_col)) {
      test_col <- factor(as.character(test_col), levels = levels(train_col))
      missing_level <- is.na(test_col) & !is.na(test_df[[col_name]])
      drop_idx <- drop_idx | missing_level
      test_df[[col_name]] <- test_col
    }
  }
  if (any(drop_idx)) {
    test_df <- test_df[!drop_idx, , drop = FALSE]
    test_y <- test_y[!drop_idx]
  }
  list(train = train_df, test = test_df, test_y = test_y, dropped = sum(drop_idx))
}

drop_high_cardinality_factors <- function(train_df, test_df, max_levels) {
  factor_cols <- names(train_df)[vapply(train_df, is.factor, logical(1))]
  if (length(factor_cols) == 0) {
    return(list(train = train_df, test = test_df, dropped = character(0)))
  }
  level_counts <- vapply(train_df[factor_cols], nlevels, integer(1))
  drop_cols <- factor_cols[level_counts > max_levels]
  if (length(drop_cols) > 0) {
    keep_idx <- !(names(train_df) %in% drop_cols)
    train_df <- train_df[, keep_idx, drop = FALSE]
    test_df <- test_df[, keep_idx, drop = FALSE]
  }
  list(train = train_df, test = test_df, dropped = drop_cols)
}

load_dataset <- function(def) {
  if (!requireNamespace(def$package, quietly = TRUE)) {
    return(list(ok = FALSE, reason = paste("Package missing:", def$package)))
  }
  env <- new.env(parent = emptyenv())
  data(list = def$name, package = def$package, envir = env)
  if (!exists(def$name, envir = env)) {
    return(list(ok = FALSE, reason = "Dataset not found"))
  }
  dat <- get(def$name, envir = env)
  if (is.matrix(dat)) {
    dat <- as.data.frame(dat)
  } else if (is.list(dat) && !is.data.frame(dat)) {
    if (!is.null(dat$x) && !is.null(dat$y)) {
      dat <- data.frame(dat$x, y = dat$y)
    } else {
      return(list(ok = FALSE, reason = "Dataset format not supported"))
    }
  }
  dat <- as.data.frame(dat)
  if (!is.null(def$drop)) {
    dat <- dat[, !(names(dat) %in% def$drop), drop = FALSE]
  }
  if (!(def$response %in% names(dat))) {
    return(list(ok = FALSE, reason = "Response column not found"))
  }
  dat <- dat[complete.cases(dat), , drop = FALSE]
  y <- dat[[def$response]]
  X <- dat[, setdiff(names(dat), def$response), drop = FALSE]
  X <- normalize_predictors(X)
  if (def$task == "classification") {
    if (is.logical(y) || is.numeric(y) || is.integer(y)) {
      y <- as.factor(y)
    } else if (is.character(y)) {
      y <- as.factor(y)
    }
    y <- droplevels(y)
    if (!is.factor(y) || length(levels(y)) != 2) {
      return(list(ok = FALSE, reason = "Classification target is not binary"))
    }
    # Enforce uniformity: first level becomes "1", second becomes "0"
    original_levels <- levels(y)
    y <- factor(ifelse(y == original_levels[1], "1", "0"), levels = c("1", "0"))
  } else {
    if (!is.numeric(y) && !is.integer(y)) {
      return(list(ok = FALSE, reason = "Regression target is not numeric"))
    }
    y <- as.numeric(y)
  }
  list(ok = TRUE, X = X, y = y, data = dat)
}

make_folds <- function(y, k, seed, stratify) {
  set.seed(seed)
  n <- length(y)
  fold_ids <- integer(n)
  if (stratify) {
    for (lvl in levels(y)) {
      idx <- which(y == lvl)
      idx <- sample(idx)
      fold_ids[idx] <- rep(seq_len(k), length.out = length(idx))
    }
  } else {
    fold_ids <- sample(rep(seq_len(k), length.out = n))
  }
  fold_ids
}

score_regression <- function(pred, y) {
  rmse <- sqrt(mean((pred - y) ^ 2))
  mae <- mean(abs(pred - y))
  denom <- sum((y - mean(y)) ^ 2)
  rsq <- if (denom > 0) 1 - sum((pred - y) ^ 2) / denom else NA_real_
  c(rmse = rmse, mae = mae, rsq = rsq)
}

score_classification <- function(pred, y, prob = NULL) {
  pred <- factor(pred, levels = levels(y))
  misclass <- mean(pred != y)
  accuracy <- mean(pred == y)
  c(misclass = misclass, accuracy = accuracy)
}

fold_rows <- list()
skip_rows <- list()

metrics_cols <- c("rmse", "mae", "rsq", "misclass", "accuracy")
progress_cols <- c(
  "dataset", "package", "task", "model", "repeat_id", "fold", "n", "n_train", "n_test",
  "p", "train_seconds", "predict_seconds", "rmse", "mae", "rsq", "misclass", "accuracy"
)
fold_id <- 1

invisible(dir.create(opts$output_dir, recursive = TRUE, showWarnings = FALSE))
progress_path <- file.path(opts$output_dir, "benchmark_progress.csv")
if (file.exists(progress_path)) {
  invisible(file.remove(progress_path))
}
append_progress_rows <- function(rows) {
  if (nrow(rows) == 0) {
    return()
  }
  if (!all(progress_cols %in% colnames(rows))) {
    stop("Progress rows missing expected columns.")
  }
  write.table(
    rows[, progress_cols, drop = FALSE],
    progress_path,
    sep = ",",
    row.names = FALSE,
    col.names = !file.exists(progress_path),
    append = file.exists(progress_path),
    quote = TRUE
  )
}

create_summary_plots <- function(summary_df, fold_df, output_path) {
  if (nrow(summary_df) == 0 && nrow(fold_df) == 0) {
    return(invisible(NULL))
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    message("ggplot2 is required to generate benchmark plots.")
    return(invisible(NULL))
  }
  plot_list <- list()
  if (nrow(summary_df) > 0) {
    for (metric in unique(summary_df$metric)) {
      df <- subset(summary_df, metric == metric)
      if (nrow(df) == 0) {
        next
      }
      df$dataset <- factor(df$dataset, levels = unique(df$dataset[order(df$mean)]))
      p <- ggplot2::ggplot(df, ggplot2::aes(x = dataset, y = mean, fill = model)) +
        ggplot2::geom_col(position = ggplot2::position_dodge(width = 0.8)) +
        ggplot2::scale_y_log10() +
        ggplot2::labs(
          title = metric,
          x = "Dataset",
          y = "Mean (log scale)"
        ) +
        ggplot2::theme_minimal() +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
      plot_list[[length(plot_list) + 1]] <- p
    }
  }
  if (nrow(fold_df) > 0) {
    time_summary <- aggregate(cbind(train_seconds, predict_seconds) ~ dataset + model, data = fold_df, FUN = mean)
    if (nrow(time_summary) > 0) {
      time_long <- data.frame(
        dataset = rep(time_summary$dataset, times = 2),
        model = rep(time_summary$model, times = 2),
        stage = rep(c("train", "predict"), each = nrow(time_summary)),
        seconds = c(time_summary$train_seconds, time_summary$predict_seconds)
      )
      time_long$dataset <- factor(time_long$dataset, levels = unique(time_long$dataset))
      p <- ggplot2::ggplot(time_long, ggplot2::aes(x = dataset, y = seconds, fill = stage)) +
        ggplot2::geom_col(position = ggplot2::position_dodge(width = 0.8)) +
        ggplot2::scale_y_log10() +
        ggplot2::facet_wrap(~model, scales = "free_y", ncol = 1) +
        ggplot2::labs(title = "Average Timing by Dataset", x = "Dataset", y = "Seconds (log scale)") +
        ggplot2::theme_minimal() +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
      plot_list[[length(plot_list) + 1]] <- p
    }
  }
  if (length(plot_list) == 0) {
    return(invisible(NULL))
  }
  pdf(output_path, width = 8, height = 6)
  on.exit(grDevices::dev.off(), add = TRUE)
  for (plot_obj in plot_list) {
    print(plot_obj)
  }
  invisible(NULL)
}

for (dataset_index in seq_along(registry)) {
  def <- registry[[dataset_index]]
  message("Loading ", def$package, "::", def$name, "...")
  loaded <- load_dataset(def)
  if (!loaded$ok) {
    skip_rows[[length(skip_rows) + 1]] <- data.frame(
      dataset = def$name,
      package = def$package,
      reason = loaded$reason,
      stringsAsFactors = FALSE
    )
    message("  Skipping: ", loaded$reason)
    next
  }
  X <- loaded$X
  y <- loaded$y
  n <- nrow(X)
  p <- ncol(X)
  if (n < opts$folds) {
    skip_rows[[length(skip_rows) + 1]] <- data.frame(
      dataset = def$name,
      package = def$package,
      reason = "Not enough rows for folds",
      stringsAsFactors = FALSE
    )
    message("  Skipping: not enough rows for folds.")
    next
  }
  if (def$task == "classification") {
    class_counts <- table(y)
    if (min(class_counts) < 2) {
      skip_rows[[length(skip_rows) + 1]] <- data.frame(
        dataset = def$name,
        package = def$package,
        reason = "Insufficient class counts",
        stringsAsFactors = FALSE
      )
      message("  Skipping: insufficient class counts.")
      next
    }
  }

  for (rep_idx in seq_len(opts$repeats)) {
    fold_seed <- opts$seed + dataset_index * 1000 + rep_idx * 100
    fold_ids <- make_folds(y, opts$folds, fold_seed, def$task == "classification")

    for (fold in seq_len(opts$folds)) {
      test_idx <- which(fold_ids == fold)
      train_idx <- which(fold_ids != fold)
      train_X <- X[train_idx, , drop = FALSE]
      test_X <- X[test_idx, , drop = FALSE]
      train_y <- y[train_idx]
      test_y <- y[test_idx]

      train_X <- drop_unused_factor_levels(train_X)

      if (def$task == "classification") {
        train_y <- droplevels(train_y)
        if (length(levels(train_y)) < 2) {
          message("  Fold skipped: training data has one class only.")
          next
        }
        test_y <- factor(test_y, levels = levels(train_y))
        if (any(is.na(test_y))) {
          message("  Fold skipped: test data has unseen response levels.")
          next
        }
      }

      aligned <- align_factor_levels(train_X, test_X, test_y)
      train_X <- aligned$train
      test_X <- aligned$test
      test_y <- aligned$test_y
      if (aligned$dropped > 0) {
        message("  Dropped ", aligned$dropped, " test rows with unseen factor levels.")
      }
      if (nrow(test_X) == 0) {
        message("  Fold skipped: no test rows after factor alignment.")
        next
      }

      dropped_factors <- drop_high_cardinality_factors(train_X, test_X, 53L)
      train_X <- dropped_factors$train
      test_X <- dropped_factors$test
      if (length(dropped_factors$dropped) > 0) {
        message(
          "  Dropped high-cardinality factors for randomForest compatibility: ",
          paste(dropped_factors$dropped, collapse = ", ")
        )
      }
      if (ncol(train_X) == 0) {
        message("  Fold skipped: no predictors after dropping high-cardinality factors.")
        next
      }

      n_train <- nrow(train_X)
      n_test <- nrow(test_X)
      fold_p <- ncol(train_X)

      fold_seed <- fold_seed + fold
      bart_params <- list(
        num_trees = opts$bart_num_trees,
        num_burn_in = opts$bart_burn_in,
        num_iterations_after_burn_in = opts$bart_iter,
        use_xoshiro = opts$use_xoshiro,
        verbose = FALSE
      )
      rf_params <- list(
        ntree = opts$rf_ntree
      )
      if (!is.na(opts$rf_mtry)) {
        rf_params$mtry <- opts$rf_mtry
      }

      set.seed(fold_seed)
      bart_time <- system.time({
        bart_fit <- do.call(bartMachine, c(list(train_X, train_y), bart_params))
      })["elapsed"]
      set.seed(fold_seed)
      rf_time <- system.time({
        rf_fit <- do.call(randomForest, c(list(train_X, train_y), rf_params))
      })["elapsed"]

      if (def$task == "classification") {
        bart_pred_time <- system.time({
          bart_prob <- predict(bart_fit, test_X, type = "prob", verbose = FALSE)
        })["elapsed"]
        bart_pred <- ifelse(bart_prob >= 0.5, levels(train_y)[1], levels(train_y)[2])
        bart_pred <- factor(bart_pred, levels = levels(train_y))
        bart_metrics <- score_classification(bart_pred, test_y, bart_prob)

        rf_pred_time <- system.time({
          rf_prob <- predict(rf_fit, test_X, type = "prob")
        })["elapsed"]
        rf_prob <- rf_prob[, levels(train_y)[1]]
        rf_pred <- predict(rf_fit, test_X, type = "response")
        rf_metrics <- score_classification(rf_pred, test_y, rf_prob)
      } else {
        bart_pred_time <- system.time({
          bart_pred <- predict(bart_fit, test_X, verbose = FALSE)
        })["elapsed"]
        bart_metrics <- score_regression(bart_pred, test_y)

        rf_pred_time <- system.time({
          rf_pred <- predict(rf_fit, test_X)
        })["elapsed"]
        rf_metrics <- score_regression(rf_pred, test_y)
      }

      bart_row <- data.frame(
        dataset = def$name,
        package = def$package,
        task = def$task,
        model = "bartMachine",
        repeat_id = rep_idx,
        fold = fold,
        n = n,
        n_train = n_train,
        n_test = n_test,
        p = fold_p,
        train_seconds = as.numeric(bart_time),
        predict_seconds = as.numeric(bart_pred_time),
        rmse = ifelse("rmse" %in% names(bart_metrics), bart_metrics["rmse"], NA_real_),
        mae = ifelse("mae" %in% names(bart_metrics), bart_metrics["mae"], NA_real_),
        rsq = ifelse("rsq" %in% names(bart_metrics), bart_metrics["rsq"], NA_real_),
        misclass = ifelse("misclass" %in% names(bart_metrics), bart_metrics["misclass"], NA_real_),
        accuracy = ifelse("accuracy" %in% names(bart_metrics), bart_metrics["accuracy"], NA_real_),
        stringsAsFactors = FALSE
      )
      fold_rows[[fold_id]] <- bart_row
      append_progress_rows(bart_row)
      fold_id <- fold_id + 1

      rf_row <- data.frame(
        dataset = def$name,
        package = def$package,
        task = def$task,
        model = "randomForest",
        repeat_id = rep_idx,
        fold = fold,
        n = n,
        n_train = n_train,
        n_test = n_test,
        p = fold_p,
        train_seconds = as.numeric(rf_time),
        predict_seconds = as.numeric(rf_pred_time),
        rmse = ifelse("rmse" %in% names(rf_metrics), rf_metrics["rmse"], NA_real_),
        mae = ifelse("mae" %in% names(rf_metrics), rf_metrics["mae"], NA_real_),
        rsq = ifelse("rsq" %in% names(rf_metrics), rf_metrics["rsq"], NA_real_),
        misclass = ifelse("misclass" %in% names(rf_metrics), rf_metrics["misclass"], NA_real_),
        accuracy = ifelse("accuracy" %in% names(rf_metrics), rf_metrics["accuracy"], NA_real_),
        stringsAsFactors = FALSE
      )
      fold_rows[[fold_id]] <- rf_row
      append_progress_rows(rf_row)
      fold_id <- fold_id + 1
    }
  }
}

fold_results <- if (length(fold_rows) == 0) {
  data.frame()
} else {
  do.call(rbind, fold_rows)
}

summary_results <- data.frame()
if (nrow(fold_results) > 0) {
  summary_rows <- list()
  summary_id <- 1
  split_key <- paste(fold_results$dataset, fold_results$model, fold_results$task)
  for (key in unique(split_key)) {
    subset_idx <- which(split_key == key)
    subset_df <- fold_results[subset_idx, , drop = FALSE]
    for (metric in metrics_cols) {
      metric_vals <- subset_df[[metric]]
      if (all(is.na(metric_vals))) {
        next
      }
      summary_rows[[summary_id]] <- data.frame(
        dataset = subset_df$dataset[1],
        package = subset_df$package[1],
        task = subset_df$task[1],
        model = subset_df$model[1],
        metric = metric,
        mean = mean(metric_vals, na.rm = TRUE),
        sd = sd(metric_vals, na.rm = TRUE),
        n = sum(!is.na(metric_vals)),
        stringsAsFactors = FALSE
      )
      summary_id <- summary_id + 1
    }
  }
  if (length(summary_rows) > 0) {
    summary_results <- do.call(rbind, summary_rows)
  }
}

comparison_results <- data.frame()
if (nrow(fold_results) > 0 && "bartMachine" %in% fold_results$model && "randomForest" %in% fold_results$model) {
  comp_rows <- list()
  comp_id <- 1
  # Datasets that have both models
  datasets <- unique(fold_results$dataset)
  
  for (ds in datasets) {
    ds_df <- fold_results[fold_results$dataset == ds, , drop = FALSE]
    # Check if we have both models for this dataset
    if (!all(c("bartMachine", "randomForest") %in% ds_df$model)) {
      next
    }
    
    for (metric in metrics_cols) {
      # Extract vectors for each model
      # Ensure sorting by repeat and fold to align pairs
      bart_df <- ds_df[ds_df$model == "bartMachine", , drop = FALSE]
      rf_df <- ds_df[ds_df$model == "randomForest", , drop = FALSE]
      
      # Merge to ensure we only compare valid pairs
      # Using repeat_id and fold as keys
      merged <- merge(
        bart_df[, c("repeat_id", "fold", metric)], 
        rf_df[, c("repeat_id", "fold", metric)], 
        by = c("repeat_id", "fold"), 
        suffixes = c("_bart", "_rf")
      )
      
      val_bart <- merged[[paste0(metric, "_bart")]]
      val_rf <- merged[[paste0(metric, "_rf")]]
      
      # Filter NAs
      valid_idx <- !is.na(val_bart) & !is.na(val_rf)
      val_bart <- val_bart[valid_idx]
      val_rf <- val_rf[valid_idx]
      
      if (length(val_bart) < 2) {
        next
      }
      
      # T-test
      # We want to test difference
      # Default: two.sided, paired=TRUE
      test_res <- tryCatch(
        t.test(val_bart, val_rf, paired = TRUE, alternative = "two.sided"),
        error = function(e) NULL
      )
      
      pval <- if (!is.null(test_res)) test_res$p.value else NA_real_
      
      comp_rows[[comp_id]] <- data.frame(
        dataset = ds,
        package = ds_df$package[1],
        task = ds_df$task[1],
        metric = metric,
        bart_mean = mean(val_bart),
        rf_mean = mean(val_rf),
        diff_mean = mean(val_bart - val_rf),
        p_value = pval,
        n = length(val_bart),
        stringsAsFactors = FALSE
      )
      comp_id <- comp_id + 1
    }
  }
  
  if (length(comp_rows) > 0) {
    comparison_results <- do.call(rbind, comp_rows)
    # Add winner and order results
    comparison_results$winner <- ifelse(
      comparison_results$metric %in% c("rmse", "mae", "misclass"),
      ifelse(comparison_results$bart_mean <= comparison_results$rf_mean, "BART", "RF"),
      ifelse(comparison_results$bart_mean >= comparison_results$rf_mean, "BART", "RF")
    )
    comparison_results <- comparison_results[order(
      comparison_results$metric,
      comparison_results$dataset,
      comparison_results$p_value
    ), ]
  }
}

skip_results <- if (length(skip_rows) == 0) {
  data.frame()
} else {
  do.call(rbind, skip_rows)
}

fold_path <- file.path(opts$output_dir, "benchmark_folds.csv")
summary_path <- file.path(opts$output_dir, "benchmark_summary.csv")
comparison_path <- file.path(opts$output_dir, "benchmark_comparisons.csv")
skip_path <- file.path(opts$output_dir, "benchmark_skipped.csv")
rds_path <- file.path(opts$output_dir, "benchmark_results.rds")
session_path <- file.path(opts$output_dir, "sessionInfo.txt")

if (nrow(fold_results) > 0) {
  write.csv(fold_results, fold_path, row.names = FALSE)
}
if (nrow(summary_results) > 0) {
  write.csv(summary_results, summary_path, row.names = FALSE)
}
if (nrow(comparison_results) > 0) {
  write.csv(comparison_results, comparison_path, row.names = FALSE)
}
if (nrow(skip_results) > 0) {
  write.csv(skip_results, skip_path, row.names = FALSE)
}
saveRDS(
  list(
    config = opts,
    registry_version = benchmark_registry_version,
    summary = summary_results,
    comparison = comparison_results,
    folds = fold_results,
    skipped = skip_results
  ),
  rds_path
)
capture.output(sessionInfo(), file = session_path)

plot_path <- file.path(opts$output_dir, "benchmark_summary_plots.pdf")
create_summary_plots(summary_results, fold_results, plot_path)

message("Benchmark complete.")
message("Summary: ", summary_path)

#!/usr/bin/env Rscript

# bartMachine vs BART vs dbarts benchmark
# Accuracy and Speed comparison

options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector"))

suppressPackageStartupMessages({
  library(bartMachine)
  library(BART)
  library(dbarts)
  library(mlbench)
  library(ggplot2)
})

# Benchmark Configuration
args <- commandArgs(trailingOnly = TRUE)

# Default values
N_REPS <- 5
N_SAMPLES <- 500

# Parse arguments
for (arg in args) {
  if (grepl("^--reps=", arg)) {
    N_REPS <- as.integer(sub("--reps=", "", arg))
  } else if (grepl("^--n=", arg)) {
    N_SAMPLES <- as.integer(sub("--n=", "", arg))
  }
}

N_TREES <- 200
N_BURN <- 500
N_POST <- 1000
CORES <- 4

set_bart_machine_num_cores(CORES)

# Datasets to benchmark
# We'll use some standard ones from mlbench and Friedman data
generate_friedman <- function(n, p = 10, sigma = 1.0) {
  X <- matrix(runif(n * p), nrow = n)
  y_mean <- 10 * sin(pi * X[, 1] * X[, 2]) + 20 * (X[, 3] - 0.5)^2 + 10 * X[, 4] + 5 * X[, 5]
  y <- y_mean + rnorm(n, 0, sigma)
  data.frame(X, y = y)
}

generate_friedman_class <- function(n, p = 10) {
  X <- matrix(runif(n * p), nrow = n)
  logit <- 1.5 * sin(pi * X[, 1] * X[, 2]) - 2 * (X[, 3] - 0.5) + 3 * X[, 4] * X[, 5] + 2 * (X[, 6] - 0.5)^2
  prob <- 1 / (1 + exp(-logit))
  y <- factor(rbinom(n, 1, prob), levels = c(0, 1), labels = c("c0", "c1"))
  data.frame(X, y = y)
}

run_benchmark <- function(data, task = "regression", dataset_name = "data") {
  results <- list()
  
  n <- nrow(data)
  train_idx <- sample(1:n, round(0.8 * n))
  train_data <- data[train_idx, ]
  test_data <- data[-train_idx, ]
  
  X_train <- train_data[, setdiff(names(train_data), "y")]
  y_train <- train_data$y
  X_test <- test_data[, setdiff(names(test_data), "y")]
  y_test <- test_data$y
  
  # For BART and dbarts, we need matrix input and dummies for factors
  # Ensure consistent columns by processing train and test together
  all_data <- rbind(train_data, test_data)
  X_all_mat <- model.matrix(y ~ . - 1, data = all_data)
  X_train_mat <- X_all_mat[1:nrow(train_data), , drop = FALSE]
  X_test_mat <- X_all_mat[(nrow(train_data) + 1):nrow(all_data), , drop = FALSE]
  
  if (task == "classification") {
    y_train_num <- as.numeric(y_train) - 1 # 0, 1
    y_test_num <- as.numeric(y_test) - 1
  }

  models <- c("bartMachine", "BART", "dbarts")
  
  for (m in models) {
    message(sprintf("Running %s on %s...", m, dataset_name))
    
    start_time <- Sys.time()
    
    if (m == "bartMachine") {
      fit <- bartMachine(X_train, y_train, num_trees = N_TREES, num_burn_in = N_BURN, 
                         num_iterations_after_burn_in = N_POST, verbose = FALSE, seed = 1,
                         use_xoshiro = TRUE)
      train_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
      
      start_pred <- Sys.time()
      if (task == "regression") {
        pred <- predict(fit, X_test)
        acc <- sqrt(mean((pred - y_test)^2)) # RMSE
      } else {
        pred <- predict(fit, X_test, type = "class")
        acc <- mean(pred != y_test) # Misclassification
      }
      pred_time <- as.numeric(difftime(Sys.time(), start_pred, units = "secs"))
      
    } else if (m == "BART") {
      if (task == "regression") {
        fit <- try(mc.wbart(X_train_mat, y_train, X_test_mat, ntree = N_TREES, nskip = N_BURN, ndpost = N_POST, 
                        mc.cores = as.integer(CORES), seed = 1, printevery = 10000L))
        if (inherits(fit, "try-error") || (!inherits(fit, "wbart") && is.list(fit))) {
           stop(sprintf("BART::mc.wbart failed on %s. Check for worker crashes or column inconsistencies.", dataset_name))
        }
        train_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
        pred_time <- 0 
        pred <- fit$yhat.test.mean
        acc <- sqrt(mean((pred - y_test)^2))
      } else {
        fit <- try(mc.pbart(X_train_mat, y_train_num, X_test_mat, ntree = N_TREES, nskip = N_BURN, ndpost = N_POST, 
                        mc.cores = as.integer(CORES), seed = 1, printevery = 10000L))
        if (inherits(fit, "try-error") || (!inherits(fit, "pbart") && is.list(fit))) {
           stop(sprintf("BART::mc.pbart failed on %s. Check for worker crashes or column inconsistencies.", dataset_name))
        }
        train_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
        pred_time <- 0
        prob <- fit$prob.test.mean
        pred <- factor(ifelse(prob > 0.5, levels(y_test)[2], levels(y_test)[1]), levels = levels(y_test))
        acc <- mean(pred != y_test)
      }
      
    } else if (m == "dbarts") {
      if (task == "regression") {
        fit <- dbarts::bart(X_train_mat, y_train, X_test_mat, ntree = N_TREES, nskip = N_BURN, ndpost = N_POST, 
                            nthread = as.integer(CORES), verbose = FALSE, seed = 1)
        train_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
        pred_time <- 0
        pred <- fit$yhat.test.mean
        acc <- sqrt(mean((pred - y_test)^2))
      } else {
        fit <- dbarts::bart(X_train_mat, y_train_num, X_test_mat, ntree = N_TREES, nskip = N_BURN, ndpost = N_POST, 
                            nthread = as.integer(CORES), verbose = FALSE, seed = 1)
        train_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
        pred_time <- 0
        prob <- colMeans(pnorm(fit$yhat.test))
        pred <- factor(ifelse(prob > 0.5, levels(y_test)[2], levels(y_test)[1]), levels = levels(y_test))
        acc <- mean(pred != y_test)
      }
    }
    
    results[[m]] <- data.frame(
      dataset = dataset_name,
      model = m,
      task = task,
      train_time = train_time,
      pred_time = pred_time,
      total_time = train_time + pred_time,
      accuracy = acc
    )
  }
  
  do.call(rbind, results)
}

# Run on several datasets
datasets <- list(
  list(name = "Friedman_Reg", task = "regression", data = generate_friedman(N_SAMPLES)),
  list(name = "Friedman_Class", task = "classification", data = generate_friedman_class(N_SAMPLES)),
  list(name = "BostonHousing", task = "regression", data = {
    data(BostonHousing)
    df <- BostonHousing
    names(df)[names(df) == "medv"] <- "y"
    df
  }),
  list(name = "PimaIndiansDiabetes", task = "classification", data = {
    data(PimaIndiansDiabetes)
    df <- PimaIndiansDiabetes
    names(df)[names(df) == "diabetes"] <- "y"
    df
  })
)

all_results <- list()
for (i in 1:N_REPS) {
  message(sprintf("\n--- Replicate %d ---", i))
  for (ds in datasets) {
    res <- run_benchmark(ds$data, ds$task, ds$name)
    res$replicate <- i
    all_results[[length(all_results) + 1]] <- res
  }
}

final_df <- do.call(rbind, all_results)

# Summary
summary_list <- list()
for (ds_name in unique(final_df$dataset)) {
  for (mdl_name in unique(final_df$model)) {
    sub_df <- final_df[final_df$dataset == ds_name & final_df$model == mdl_name, ]
    if (nrow(sub_df) == 0) next
    
    summary_list[[length(summary_list) + 1]] <- data.frame(
      dataset = ds_name,
      model = mdl_name,
      task = sub_df$task[1],
      train_time_mean = mean(sub_df$train_time),
      train_time_sd = if(nrow(sub_df) > 1) sd(sub_df$train_time) else NA_real_,
      pred_time_mean = mean(sub_df$pred_time),
      pred_time_sd = if(nrow(sub_df) > 1) sd(sub_df$pred_time) else NA_real_,
      total_time_mean = mean(sub_df$total_time),
      total_time_sd = if(nrow(sub_df) > 1) sd(sub_df$total_time) else NA_real_,
      accuracy_mean = mean(sub_df$accuracy),
      accuracy_sd = if(nrow(sub_df) > 1) sd(sub_df$accuracy) else NA_real_,
      n = nrow(sub_df)
    )
  }
}

summary_flat <- do.call(rbind, summary_list)

print(summary_flat)

write.csv(summary_flat, "bartMachine/inst/benchmarks/bart_packages_comparisons_results/summary.csv", row.names = FALSE)
write.csv(final_df, "bartMachine/inst/benchmarks/bart_packages_comparisons_results/raw_results.csv", row.names = FALSE)

# Plotting
pdf("bartMachine/inst/benchmarks/bart_packages_comparisons_results/plots.pdf", width = 10, height = 8)

# Time Plot
p1 <- ggplot(summary_flat, aes(x = model, y = total_time_mean, fill = model)) +
  geom_bar(stat = "identity") +
  geom_errorbar(aes(ymin = total_time_mean - total_time_sd, ymax = total_time_mean + total_time_sd), width = 0.2) +
  facet_wrap(~dataset, scales = "free_y") +
  theme_minimal() +
  labs(title = "Total Time Comparison (Training + Prediction)", y = "Time (seconds)")

print(p1)

# Accuracy Plot
p2 <- ggplot(summary_flat, aes(x = model, y = accuracy_mean, fill = model)) +
  geom_bar(stat = "identity") +
  geom_errorbar(aes(ymin = accuracy_mean - accuracy_sd, ymax = accuracy_mean + accuracy_sd), width = 0.2) +
  facet_wrap(~dataset, scales = "free_y") +
  theme_minimal() +
  labs(title = "Accuracy Comparison (RMSE for Reg, Misclass for Class)", y = "Error (Lower is better)")

print(p2)

dev.off()

message("\nBenchmark complete. Results saved in bartMachine/inst/benchmarks/bart_packages_comparisons_results/")

##performs out-of-sample error estimation for a BART model
#' Estimate Out-of-sample Error with K-fold Cross validation
#'
#' @description
#' Builds a BART model using a specified set of arguments to \code{build_bart_machine} and estimates the out-of-sample performance by using k-fold cross validation.
#'
#' @details
#' For each fold, a new BART model is trained (using the same set of arguments) and its performance is evaluated on the holdout piece of that fold.
#' @param X Data frame of predictors. Factors are automatically converted to dummies internally.
#' @param y Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.
#' @param k_folds Number of folds to cross-validate over. This argument is ignored if \code{folds_vec} is non-null.
#' @param folds_vec An integer vector of indices specifying which fold each observation belongs to.
#' @param verbose Prints information about progress of the algorithm to the screen.
#' @param \dots Additional arguments to be passed to \code{build_bart_machine}.
#'
#' @return
#' For regression models, a list with the following components is returned:
#'   \item{y_hat}{Predictions for the observations computed on the fold for which the observation was omitted from the training set.}
#'   \item{L1_err}{Aggregate L1 error across the folds.}
#'   \item{L2_err}{Aggregate L1 error across the folds.}
#'   \item{rmse}{Aggregate RMSE across the folds.}
#'   \item{folds}{Vector of indices specifying which fold each observation belonged to.}
#' 
#' For classification models, a list with the following components is returned:
#' 
#'   \item{y_hat}{Class predictions for the observations computed on the fold for which the observation was omitted from the training set.}
#'     \item{p_hat}{Probability estimates for the observations computed on the fold for which the observation was omitted from the training set.}
#'   \item{confusion_matrix}{Aggregate confusion matrix across the folds.}
#'     \item{misclassification_error}{Total misclassification error across the folds.}
#'     \item{folds}{Vector of indices specifying which fold each observation belonged to.}
#'
#' @seealso
#' \code{\link{bartMachine}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' #evaluate default BART on 5 folds
#' k_fold_val = k_fold_cv(X, y)
#' print(k_fold_val$rmse)
#' }
#' @export
k_fold_cv = function(X, y, k_folds = 5, folds_vec = NULL, verbose = FALSE, ...){
  # Validate arguments
  assert_data_frame(X)
  assert_atomic_vector(y)
  assert_number(k_folds, lower = 2) # Inf is allowed as it is > 2, but wait assert_number(Inf) works.
  assert_integerish(folds_vec, null.ok = TRUE)
  assert_flag(verbose)

	#we cannot afford the time sink of serialization during the grid search, so shut it off manually
	args = list(...)
	args$serialize = FALSE

	
	if (!inherits(X, "data.frame")){
		stop("The training data X must be a data frame.")
	}
	if (!(class(y) %in% c("numeric", "integer", "factor"))){
		stop("Your response must be either numeric, an integer or a factor with two levels.\n")
	}	
    if (!is.null(folds_vec) & !inherits(folds_vec, "integer")){
	  stop("folds_vec must be an a vector of integers specifying the indexes of each folds.")  
    }
	
	y_levels = levels(y)
	if (inherits(y, "numeric") || inherits(y, "integer")){ #if y is numeric, then it's a regression problem
		pred_type = "regression"
	} else if (inherits(y, "factor") & length(y_levels) == 2){ #if y is a factor and and binary, then it's a classification problem
		pred_type = "classification"
	}
	
	n = nrow(X)
	Xpreprocess = pre_process_training_data(X)$data
	
	p = ncol(Xpreprocess)
	
	#set up k folds
    if (is.null(folds_vec)){ ##if folds were not pre-set:
	  	if (k_folds == Inf){ #leave-one-out
	  		k_folds = n
	  	}
	  	
	  	if (k_folds <= 1 || k_folds > n){
	  		stop("The number of folds must be at least 2 and less than or equal to n, use \"Inf\" for leave one out")
	  	}
		
	  	temp = rnorm(n)
		
	  	folds_vec = cut(temp, breaks = quantile(temp, seq(0, 1, length.out = k_folds + 1)), 
  	                include.lowest= T, labels = F)
    } else {
    	k_folds = length(unique(folds_vec)) ##otherwise we know the folds, so just get k
    }	

	if (pred_type == "regression"){
		L1_err = 0
		L2_err = 0
		yhat_cv = numeric(n) ##store cv
	} else {
    phat_cv = numeric(n)
    yhat_cv = factor(n, levels = y_levels)
		confusion_matrix = matrix(0, nrow = 3, ncol = 3)
		rownames(confusion_matrix) = c(paste("actual", y_levels), "use errors")
		colnames(confusion_matrix) = c(paste("predicted", y_levels), "model errors")
	}

	Xy = data.frame(Xpreprocess, y) ##set up data

	if (pred_type == "regression") {
		# Java path: builds all k fold models in parallel, returns out-of-fold predictions
		bart_machine_template = do.call(build_bart_machine, c(list(
				X = as.data.frame(Xpreprocess[, 1 : p, drop = FALSE]),
				y = as.numeric(y),
				run_in_sample = FALSE,
				verbose = FALSE), args))
		yhat_cv = .jcall(bart_machine_template$java_bart_machine, "[D", "runKFoldCV",
				.jarray(as.integer(folds_vec)), as.integer(k_folds))
		y_num = as.numeric(y)
		L1_err = sum(abs(y_num - yhat_cv))
		L2_err = sum((y_num - yhat_cv)^2)
		list(y_hat = yhat_cv, L1_err = L1_err, L2_err = L2_err, rmse = sqrt(L2_err / n),
				PseudoRsq = 1 - L2_err / sum((y_num - mean(y_num))^2), folds = folds_vec)
	} else {
	for (k in 1 : k_folds){
		if (verbose){
			cat(".")
		}

	    test_idx = which(folds_vec == k)
		test_data_k = Xy[test_idx, ]
		training_data_k = Xy[-test_idx, ]

   		#build bart object
		bart_machine_cv = do.call(build_bart_machine, c(list(
							X = training_data_k[, 1 : p, drop = FALSE],
							y = training_data_k[, (p + 1)],
							run_in_sample = FALSE,
							verbose = verbose), args))
		predict_obj = bart_predict_for_test_data(
			bart_machine_cv,
			test_data_k[, 1 : p, drop = FALSE],
			test_data_k[, (p + 1)],
			verbose = verbose
		)

		#tabulate errors
        phat_cv[test_idx] = predict_obj$p_hat
        yhat_cv[test_idx] = predict_obj$y_hat
		tab = table(factor(test_data_k$y, levels = y_levels), factor(predict_obj$y_hat, levels = y_levels))
		confusion_matrix[1 : 2, 1 : 2] = confusion_matrix[1 : 2, 1 : 2] + tab
	}
	if (verbose){
		cat("\n")
	}
	#calculate the rest of the confusion matrix and return it plus the errors
	confusion_matrix[3, 1] = round(confusion_matrix[2, 1] / (confusion_matrix[1, 1] + confusion_matrix[2, 1]), 3)
	confusion_matrix[3, 2] = round(confusion_matrix[1, 2] / (confusion_matrix[1, 2] + confusion_matrix[2, 2]), 3)
	confusion_matrix[1, 3] = round(confusion_matrix[1, 2] / (confusion_matrix[1, 1] + confusion_matrix[1, 2]), 3)
	confusion_matrix[2, 3] = round(confusion_matrix[2, 1] / (confusion_matrix[2, 1] + confusion_matrix[2, 2]), 3)
	confusion_matrix[3, 3] = round((confusion_matrix[1, 2] + confusion_matrix[2, 1]) / sum(confusion_matrix[1 : 2, 1 : 2]), 3)
	list(y_hat = yhat_cv, phat = phat_cv, confusion_matrix = confusion_matrix, misclassification_error = confusion_matrix[3, 3], folds = folds_vec)
	}
	
}

#' Build a BART Model
#'
#' @description
#' Builds a BART model for regression or classification.
#' @param X Data frame of predictors. Factors are automatically converted to dummies internally.
#' @param y Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.
#' @param Xy A data frame of predictors and the response. The response column must be named ``y''.
#' @param num_trees The number of trees to be grown in the sum-of-trees model.
#' @param num_burn_in Number of MCMC samples to be discarded as ``burn-in''.
#' @param num_iterations_after_burn_in Number of MCMC samples to draw from the posterior distribution of \eqn{\hat{f}(x)}.
#' @param alpha Base hyperparameter in tree prior for whether a node is nonterminal or not.
#' @param beta Power hyperparameter in tree prior for whether a node is nonterminal or not.
#' @param k For regression, \code{k} determines the prior probability that \eqn{E(Y|X)} is contained in the interval \eqn{(y_{min}, y_{max})}, based on a normal distribution. For example, when \eqn{k=2}, the prior probability is 95\%. For classification, \code{k} determines the prior probability that \eqn{E(Y|X)} is between \eqn{(-3,3)}. Note that a larger value of \code{k} results in more shrinkage and a more conservative fit.
#' @param q Quantile of the prior on the error variance at which the data-based estimate is placed. Note that the larger the value of \code{q}, the more aggressive the fit as you are placing more prior weight on values lower than the data-based estimate. Not used for classification.
#' @param nu Degrees of freedom for the inverse \eqn{\chi^2} prior. Not used for classification.
#' @param prob_rule_class Threshold for classification. Any observation with a conditional probability greater than \code{prob_class_rule} is assigned the ``positive'' outcome. Note that the first level of the response is treated as the ``positive'' outcome and the second is treated as the ``negative'' outcome.
#' @param mh_prob_steps Vector of prior probabilities for proposing changes to the tree structures: (GROW, PRUNE, CHANGE)
#' @param debug_log If TRUE, additional information about the model construction are printed to a file in the working directory.
#' @param run_in_sample If TRUE, in-sample statistics such as \eqn{\hat{f}(x)}, Pseudo-\eqn{R^2}, and RMSE are computed. Setting this to FALSE when not needed can decrease computation time.
#' @param s_sq_y If ``mse'', a data-based estimated of the error variance is computed as the MSE from ordinary least squares regression. If ``var''., the data-based estimate is computed as the variance of the response. Not used in classification.
#' @param sig_sq_est Pass in an estimate of the maximum sig_sq of the model. This is useful to cache somewhere and then pass in during cross-validation since the default method of estimation is a linear model. In large dimensions, linear model estimation is slow.
#' @param print_tree_illustrations For every Gibbs iteration, print out an illustration of the trees side-by-side. This is excruciatingly SLOW!
#' @param cov_prior_vec Vector assigning relative weights to how often a particular variable should be proposed as a candidate for a split. The vector is internally normalized so that the weights sum to 1. Note that the length of this vector must equal the length of the design matrix after dummification and augmentation of indicators of missingness (if used). To see what the dummified matrix looks like, use \code{\link{dummify_data}}. See Bleich et al. (2013) for more details on when this feature is most appropriate.
#' @param interaction_constraints A list of vectors indicating where the vectors are sets of elements allowed to interact with one another. The elements in each
#'   vector correspond to features in the data frame \code{X} specified by either the column number as a numeric value or the column
#'   name as a string e.g. \code{list(c(1, 2), c("nox", "rm"))}. The elements of the vectors can be reused among components for any
#'   level of interaction complexity you wish. Default is \code{NULL} which corresponds to the vanilla modeling procedure where
#'   all interactions are legal. For a pure generalized added model, use \code{as.list(seq(1 : p))} where \code{p}
#'   is the number of columns in the design matrix \code{X}.
#' @param use_missing_data If TRUE, the missing data feature is used to automatically handle missing data without imputation. See Kapelner and Bleich (2013) for details.
#' @param covariates_to_permute Private argument for \code{\link{cov_importance_test}}. Not needed by user.
#' @param num_rand_samps_in_library Before building a BART model, samples from the Standard Normal and \eqn{\chi^2(\nu)} are drawn to be used in the MCMC steps. This parameter determines the number of samples to be taken.
#' @param use_missing_data_dummies_as_covars If TRUE, additional indicator variables for whether or not an observation in a particular column is missing are included. See Kapelner and Bleich (2013) for details.
#' @param replace_missing_data_with_x_j_bar If TRUE ,missing entries in \code{X} are imputed with average value or modal category.
#' @param impute_missingness_with_rf_impute If TRUE, missing entries are filled in using the rf.impute() function from the \code{randomForest} library.
#' @param impute_missingness_with_x_j_bar_for_lm If TRUE, when computing the data-based estimate of \eqn{\sigma^2}, missing entries are imputed with average value or modal category.
#' @param mem_cache_for_speed Speed enhancement that caches the predictors and the split values that are available at each node for selecting new rules. If the number
#'   of predictors is large, the memory requirements become large. We recommend keeping this on (default) and turning it off if you experience out-of-memory errors.
#' @param flush_indices_to_save_RAM Setting this flag to \code{TRUE} saves memory with the downside that you cannot use the functions \code{node_prediction_training_data_indices} nor \code{get_projection_weights}.
#' @param serialize Setting this option to \code{TRUE} will allow serialization of bartMachine objects which allows for persistence between
#'   R sessions if the object is saved and reloaded. Note that serialized objects can take up a large amount of memory.
#'   Thus, the default is \code{FALSE}.
#' @param seed Optional: sets the seed in both R and Java. Default is \code{NULL} which does not set the seed in R nor Java.
#'   Setting the seed enforces deterministic behavior only in the case when one core is used (the default before
#'   \code{set_bart_machine_num_cores() was invoked}.
#' @param use_xoshiro if TRUE, use the Xoshiro256PlusPlus random number generator; if FALSE, use the legacy MersenneTwister
#'   random number generator (default is FALSE)
#' @param verbose Prints information about progress of the algorithm to the screen.
#'
#' @return
#' Returns an object of class ``bartMachine''. The ``bartMachine'' object contains a list of the following components:
#' 
#'   \item{java_bart_machine}{A pointer to the BART Java object.}
#'   \item{train_data_features}{The names of the variables used in the training data.}
#'   \item{training_data_features_with_missing_features.}{The names of the variables used in the training data. If \code{use_missing_data_dummies_as_covars = TRUE}, this also includes dummies for any predictors that contain at least one missing entry (named ``M_<feature>'').}
#'   \item{y}{The values of the response for the training data.}
#'   \item{y_levels}{The levels of the response (for classification only).}
#'   \item{pred_type}{Whether the model was build for regression of classification.}
#'   \item{model_matrix_training_data}{The training data with factors converted to dummies.}
#'   \item{num_cores}{The number of cores used to build the BART model.}
#'   \item{sig_sq_est}{The data-based estimate of \eqn{\sigma^2} used to create the prior on the error variance for the BART model.}
#'   \item{time_to_build}{Total time to build the BART model.}
#'   \item{y_hat_train}{The posterior means of \eqn{\hat{f}(x)} for each observation. Only returned if \code{run_in_sample = TRUE}.}
#'   \item{residuals}{The model residuals given by \code{y} - \code{y_hat_train}. Only returned if \code{run_in_sample = TRUE}.}
#'   \item{L1_err_train}{L1 error on the training set. Only returned if \code{run_in_sample = TRUE}.}
#'   \item{L2_err_train}{L2 error on the training set. Only returned if \code{run_in_sample = TRUE}.}
#'   \item{PseudoRsq}{Calculated as 1 - SSE / SST where SSE is the sum of square errors in the training data and SST is the sample variance of the response times \eqn{n-1}. Only returned if \code{run_in_sample = TRUE}.}
#'   \item{rmse_train}{Root mean square error on the training set. Only returned if \code{run_in_sample = TRUE}.}
#' 
#' Additionally, the parameters passed to the function \code{bartMachine} are also components of the list.
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#' 
#' HA Chipman, EI George, and RE McCulloch. BART: Bayesian Additive Regressive Trees.
#' The Annals of Applied Statistics, 4(1): 266--298, 2010.
#' 
#' A Kapelner and J Bleich. Prediction with Missing Data via Bayesian Additive Regression
#' Trees. Canadian Journal of Statistics, 43(2): 224-239, 2015
#' 
#' J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
#' Additive Regression Trees. ArXiv e-prints, 2013.
#'
#' @seealso
#' \code{\link{bartMachineCV}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set by \code{\link{set_bart_machine_num_cores}}. Each core will create an
#' independent MCMC chain of size \cr
#' \code{num_burn_in} \eqn{+} \code{num_iterations_after_burn_in / bart_machine_num_cores}.
#'
#' @examples
#' \dontrun{
#' ##regression example
#' 
#' ##generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' summary(bart_machine)
#' 
#' ##Build another BART regression model
#' bart_machine = bartMachine(X,y, num_trees = 200, num_burn_in = 500,
#' num_iterations_after_burn_in = 1000)
#' 
#' ##Classification example
#' 
#' #get data and only use 2 factors
#' data(iris)
#' iris2 = iris[51:150,]
#' iris2$Species = factor(iris2$Species)
#' 
#' #build BART classification model
#' bart_machine = build_bart_machine(iris2[ ,1:4], iris2$Species)
#' 
#' ##get estimated probabilities
#' phat = bart_machine$p_hat_train
#' ##look at in-sample confusion matrix
#' bart_machine$confusion_matrix
#' }
#'
#' @aliases build_bart_machine
#' @export
bartMachine = function(
	X = NULL, 
	y = NULL, 
	Xy = NULL, 
    num_trees = 50, #found many times to not get better after this value... so let it be the default, it's faster too 
    num_burn_in = 250, 
    num_iterations_after_burn_in = 1000, 
    alpha = 0.95,
    beta = 2,
    k = 2,
    q = 0.9,
    nu = 3.0,
    prob_rule_class = 0.5,
    mh_prob_steps = c(2.5, 2.5, 4) / 9, #only the first two matter
    debug_log = FALSE,
    run_in_sample = TRUE,
    s_sq_y = "mse", # "mse" or "var"
	sig_sq_est = NULL,
    print_tree_illustrations = FALSE, #POWER USERS ONLY
    cov_prior_vec = NULL,
	interaction_constraints = NULL,
    use_missing_data = FALSE,
    covariates_to_permute = NULL, #PRIVATE
    num_rand_samps_in_library = 10000, #give the user the option to make a bigger library of random samples of normals and inv-gammas
    use_missing_data_dummies_as_covars = FALSE,
    replace_missing_data_with_x_j_bar = FALSE,
    impute_missingness_with_rf_impute = FALSE,
    impute_missingness_with_x_j_bar_for_lm = TRUE,
    mem_cache_for_speed = TRUE,
	flush_indices_to_save_RAM = TRUE,
	serialize = FALSE,
	seed = NULL,
	use_xoshiro = FALSE,
    verbose = TRUE
    	){
  # Validate arguments
  assert_data_frame(X, null.ok = TRUE)
  if (!is.null(y)) assert_atomic_vector(y)
  assert_data_frame(Xy, null.ok = TRUE)
  assert_int(num_trees, lower = 1)
  assert_int(num_burn_in, lower = 0)
  assert_int(num_iterations_after_burn_in, lower = 1)
  assert_number(alpha, lower = 0, upper = 1)
  assert_number(beta, lower = 0)
  assert_number(k, lower = .Machine$double.eps)
  assert_number(q, lower = .Machine$double.eps, upper = 1 - .Machine$double.eps)
  assert_number(nu, lower = .Machine$double.eps)
  assert_number(prob_rule_class, lower = 0, upper = 1)
  assert_numeric(mh_prob_steps, len = 3, lower = .Machine$double.eps)
  assert_flag(debug_log)
  assert_flag(run_in_sample)
  assert_choice(s_sq_y, c("mse", "var"))
  assert_number(sig_sq_est, lower = .Machine$double.eps, null.ok = TRUE)
  assert_flag(print_tree_illustrations)
  assert_numeric(cov_prior_vec, lower = .Machine$double.eps, null.ok = TRUE)
  assert_list(interaction_constraints, null.ok = TRUE)
  assert_flag(use_missing_data)
  assert_int(num_rand_samps_in_library, lower = 1)
  assert_flag(use_missing_data_dummies_as_covars)
  assert_flag(replace_missing_data_with_x_j_bar)
  assert_flag(impute_missingness_with_rf_impute)
  assert_flag(impute_missingness_with_x_j_bar_for_lm)
  assert_flag(mem_cache_for_speed)
  assert_flag(flush_indices_to_save_RAM)
  assert_flag(serialize)
  assert_int(seed, null.ok = TRUE)
  assert_flag(use_xoshiro)
  assert_flag(verbose)

    	build_bart_machine(    
			X = X, 
			y = y, 
			Xy = Xy, 
			num_trees = num_trees,
			num_burn_in = num_burn_in, 
			num_iterations_after_burn_in = num_iterations_after_burn_in, 
			alpha = alpha,
			beta = beta,
			k = k,
			q = q,
			nu = nu,
			prob_rule_class = prob_rule_class,
			mh_prob_steps = mh_prob_steps,
			debug_log = debug_log,
			run_in_sample = run_in_sample,
			s_sq_y = s_sq_y,
			sig_sq_est = sig_sq_est,
			print_tree_illustrations = print_tree_illustrations, 
			cov_prior_vec = cov_prior_vec,
			interaction_constraints = interaction_constraints,
			use_missing_data = use_missing_data,
			covariates_to_permute = covariates_to_permute, 
			num_rand_samps_in_library = num_rand_samps_in_library, #give the user the option to make a bigger library of random samples of normals and inv-gammas
			use_missing_data_dummies_as_covars = use_missing_data_dummies_as_covars,
			replace_missing_data_with_x_j_bar = replace_missing_data_with_x_j_bar,
			impute_missingness_with_rf_impute = impute_missingness_with_rf_impute,
			impute_missingness_with_x_j_bar_for_lm = impute_missingness_with_x_j_bar_for_lm,
			mem_cache_for_speed = mem_cache_for_speed,
			flush_indices_to_save_RAM = flush_indices_to_save_RAM,
			serialize = serialize,
			seed = seed,
			use_xoshiro = use_xoshiro,
			verbose = verbose			
	   )
}


#' Build BART-CV
#'
#' @description
#' Builds a BART-CV model by cross-validating over a grid of hyperparameter choices.
#' @param X Data frame of predictors. Factors are automatically converted to dummies internally.
#' @param y Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.
#' @param Xy A data frame of predictors and the response. The response column must be named ``y''.
#' @param num_tree_cvs Vector of sizes for the sum-of-trees models to cross-validate over.
#' @param k_cvs Vector of choices for the hyperparameter \code{k} to cross-validate over.
#' @param nu_q_cvs Only for regression. List of vectors containing (\code{nu}, \code{q}) ordered pair choices to cross-validate over. If \code{NULL}, then it defaults to the three values \code{list(c(3, 0.9), c(3, 0.99), c(10, 0.75))}.
#' @param k_folds Number of folds for cross-validation
#' @param folds_vec An integer vector of indices specifying which fold each observation belongs to.
#' @param use_xoshiro if TRUE, use the Xoshiro256PlusPlus random number generator; if FALSE, use the legacy MersenneTwister
#'   random number generator (default is FALSE)
#' @param verbose Prints information about progress of the algorithm to the screen.
#' @param \dots Additional arguments to be passed to \code{bartMachine}.
#'
#' @return
#' Returns an object of class ``bartMachine'' with the set of hyperparameters chosen via cross-validation. We also return a matrix ``cv_stats''
#' which contains the out-of-sample RMSE for each hyperparameter set tried and ``folds'' which gives the fold in which each observation fell across the k-folds.
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#'
#' @seealso
#' \code{\link{bartMachine}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function may require significant run-time.
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}} via calling \code{\link{bartMachine}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine_cv = bartMachineCV(X, y)
#' 
#' #information about cross-validated model
#' summary(bart_machine_cv)
#' }
#'
#' @aliases build_bart_machine_cv
#' @export
bartMachineCV = function(X = NULL, y = NULL, Xy = NULL, 
   num_tree_cvs = c(50, 200),
   k_cvs = c(2, 3, 5),
   nu_q_cvs = NULL,
   k_folds = 5, 
   folds_vec = NULL,  
   use_xoshiro = FALSE,   
   verbose = FALSE, 
   ...){
  # Validate arguments
  assert_data_frame(X, null.ok = TRUE)
  if (!is.null(y)) assert_atomic_vector(y)
  assert_data_frame(Xy, null.ok = TRUE)
  assert_integerish(num_tree_cvs, lower = 1, min.len = 1)
  assert_numeric(k_cvs, lower = 0, min.len = 1)
  assert_list(nu_q_cvs, null.ok = TRUE)
  assert_count(k_folds, positive = TRUE)
  assert_integerish(folds_vec, null.ok = TRUE)
  assert_flag(use_xoshiro)
  assert_flag(verbose)

  	build_bart_machine_cv(X, y, Xy, 
	   num_tree_cvs,
	   k_cvs,
	   nu_q_cvs,
	   k_folds, 
	   folds_vec, 
       use_xoshiro = use_xoshiro,
	   verbose = verbose,
	   ...)
}

test_that("bartMachine runs with custom tree hyperparameters", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)

  # Test alpha, beta, k
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    alpha = 0.5,
    beta = 3,
    k = 3,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with custom error variance prior", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)

  # Test q, nu, s_sq_y
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    q = 0.95,
    nu = 5.0,
    s_sq_y = "var",
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")

  # Test custom sig_sq_est
  model_custom_sig <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    sig_sq_est = 0.01,
    verbose = FALSE
  )
  expect_s3_class(model_custom_sig, "bartMachine")
  expect_equal(model_custom_sig$sig_sq_est, 0.01)
})

test_that("bartMachine runs with custom covariate priors", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 3), ncol = 3))
  y <- X[, 1] + X[, 2] + rnorm(BART_TESTS$small_data_n, sd = 0.1)

  # cov_prior_vec
  prior <- c(0.1, 0.8, 0.1) # Sums to 1
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    cov_prior_vec = prior,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with different MH steps", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)

  # mh_prob_steps: GROW, PRUNE, CHANGE
  steps <- c(1/3, 1/3, 1/3)
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    mh_prob_steps = steps,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with memory options", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)

  # mem_cache_for_speed = FALSE
  model_no_cache <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    mem_cache_for_speed = FALSE,
    verbose = FALSE
  )
  expect_s3_class(model_no_cache, "bartMachine")

  # flush_indices_to_save_RAM = TRUE
  model_flush <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    flush_indices_to_save_RAM = TRUE,
    verbose = FALSE
  )
  expect_s3_class(model_flush, "bartMachine")
})

test_that("bartMachine runs with custom random sample library size", {
    set.seed(BART_TESTS$seed)
    X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
    y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
    
    model <- bartMachine(
        X, y,
        num_trees = 10,
        num_burn_in = 5,
        num_iterations_after_burn_in = 5,
        num_rand_samps_in_library = 1000,
        verbose = FALSE
    )
    expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with Xy argument", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  Xy <- cbind(X, y)
  colnames(Xy)[3] <- "y"
  
  model <- bartMachine(
    Xy = Xy,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
  expect_equal(model$y, y)
})

test_that("bartMachine runs with prob_rule_class", {
  set.seed(BART_TESTS$seed)
  n <- BART_TESTS$small_data_n
  p <- 2
  X <- data.frame(matrix(runif(n * p), ncol = p))
  y <- factor(ifelse(X[, 1] > 0.5, "1", "0"))
  
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    prob_rule_class = 0.3,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
  # Note: The prob_rule_class is used during prediction, but stored in the model?
  # Let's check if it's stored. Based on documentation it might not be directly exposed 
  # in the list but used in internal Java object or classification logic.
  # However, it is an argument, so we test it doesn't crash.
})

test_that("bartMachine runs with run_in_sample = FALSE", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    run_in_sample = FALSE,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
  expect_null(model$y_hat_train)
  expect_null(model$residuals)
  expect_null(model$L1_err_train)
  expect_null(model$L2_err_train)
  expect_null(model$PseudoRsq)
  expect_null(model$rmse_train)
})

test_that("bartMachine runs with debug_log = TRUE", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # It writes to a file, so we just check it runs without error
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    debug_log = TRUE,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with seed argument", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # We should get identical results if we pass the same seed
  set.seed(BART_TESTS$seed)
  model1 <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    seed = BART_TESTS$seed,
    verbose = FALSE
  )
  
  set.seed(BART_TESTS$seed)
  model2 <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    seed = BART_TESTS$seed,
    verbose = FALSE
  )
  
  expect_equal(model1$y_hat_train, model2$y_hat_train)
})

test_that("bartMachine runs with replace_missing_data_with_x_j_bar", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  X[1:5, 1] <- NA
  y <- X[, 2] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    replace_missing_data_with_x_j_bar = TRUE,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with impute_missingness_with_rf_impute", {
  skip_if_not_installed("randomForest")
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  X[1:5, 1] <- NA
  y <- X[, 2] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    impute_missingness_with_rf_impute = TRUE,
    replace_missing_data_with_x_j_bar = TRUE,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with impute_missingness_with_x_j_bar_for_lm", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  X[1:5, 1] <- NA
  y <- X[, 2] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    use_missing_data = TRUE,
    impute_missingness_with_x_j_bar_for_lm = TRUE,
    verbose = FALSE
  )
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine runs with print_tree_illustrations", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # Capturing output because it prints to stdout
  capture.output({
    model <- bartMachine(
      X, y,
      num_trees = 2, # Small number to reduce output
      num_burn_in = 2,
      num_iterations_after_burn_in = 2,
      print_tree_illustrations = TRUE,
      verbose = FALSE
    )
  })
  expect_s3_class(model, "bartMachine")
})

test_that("bartMachine validates integer arguments via checkmate", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
          # num_trees
  
          expect_error(bartMachine(X, y, num_trees = 0, verbose = FALSE))
  
      
  
  
  expect_error(bartMachine(X, y, num_trees = 1.5, verbose = FALSE))
  
  # num_burn_in
  expect_error(bartMachine(X, y, num_burn_in = -1, verbose = FALSE))
  expect_error(bartMachine(X, y, num_burn_in = 10.5, verbose = FALSE))
  
  # num_iterations_after_burn_in
  expect_error(bartMachine(X, y, num_iterations_after_burn_in = 0, verbose = FALSE))
})

test_that("bartMachine validates numeric hyperparameter arguments via checkmate", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # alpha [0, 1]
  expect_error(bartMachine(X, y, alpha = -0.1, verbose = FALSE))
  expect_error(bartMachine(X, y, alpha = 1.1, verbose = FALSE))
  
  # beta >= 0
  expect_error(bartMachine(X, y, beta = -1, verbose = FALSE))
  
  # k >= 0
  expect_error(bartMachine(X, y, k = -2, verbose = FALSE))
  
  # q [0, 1]
  expect_error(bartMachine(X, y, q = 1.5, verbose = FALSE))
  
  # nu >= 0
  expect_error(bartMachine(X, y, nu = -1, verbose = FALSE))
  
  # prob_rule_class [0, 1]
  expect_error(bartMachine(X, y, prob_rule_class = 2, verbose = FALSE))
})

test_that("bartMachine validates vector arguments via checkmate", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # mh_prob_steps: length 3
  expect_error(bartMachine(X, y, mh_prob_steps = c(0.5, 0.5), verbose = FALSE))
  expect_error(bartMachine(X, y, mh_prob_steps = c(-1, 0.5, 0.5), verbose = FALSE))
})

test_that("bartMachine validates boolean flags via checkmate", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  expect_error(bartMachine(X, y, debug_log = "yes", verbose = FALSE))
  expect_error(bartMachine(X, y, verbose = NULL))
  expect_error(bartMachine(X, y, use_missing_data = 1, verbose = FALSE))
})

test_that("bartMachineCV validates arguments via checkmate", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  expect_error(bartMachineCV(X, y, num_tree_cvs = c(10, -5), verbose = FALSE))
  expect_error(bartMachineCV(X, y, k_cvs = c(-1), verbose = FALSE))
  expect_error(bartMachineCV(X, y, k_folds = 1, verbose = FALSE))
  expect_error(bartMachineCV(X, y, k_folds = 0, verbose = FALSE))
})

test_that("predict.bartMachine validates arguments via checkmate", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  model <- bartMachine(X, y, num_trees = 5, num_burn_in = 2, num_iterations_after_burn_in = 2, verbose = FALSE)
  
  expect_error(predict(model, new_data = as.matrix(X), verbose = FALSE))
  expect_error(predict(model, X, type = "foo", verbose = FALSE))
  expect_error(predict(model, X, prob_rule_class = 1.5, verbose = FALSE))
})

test_that("bartMachine serialization works when serialize = TRUE", {
  skip_on_cran() # Serialization tests can be slow or problematic on CRAN
  
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # Build model with serialization enabled
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    serialize = TRUE,
    verbose = FALSE
  )
  
  # Save to a temporary file
  tmp_path <- tempfile(fileext = ".rds")
  saveRDS(model, tmp_path)
  
  # Load it back
  model_loaded <- readRDS(tmp_path)
  
  # Check if the loaded model can predict
  # This verifies that the Java object was correctly restored
  pred <- predict(model_loaded, X, verbose = FALSE)
  expect_length(pred, nrow(X))
  expect_false(is.null(model_loaded$java_bart_machine))
  
  unlink(tmp_path)
})

test_that("bartMachine serialization fails/warns when serialize = FALSE", {
  skip_on_cran()
  
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  # Build model with serialization disabled (default)
  model <- bartMachine(
    X, y,
    num_trees = 10,
    num_burn_in = 5,
    num_iterations_after_burn_in = 5,
    serialize = FALSE,
    verbose = FALSE
  )
  
  # Save to a temporary file
  tmp_path <- tempfile(fileext = ".rds")
  saveRDS(model, tmp_path)
  
  # Load it back
  model_loaded <- readRDS(tmp_path)
  
  # The Java object pointer should be dead or null, or prediction should fail
  # Note: rJava objects that are not serialized become null or invalid pointers upon reload
  # attempting to use them usually throws an error.
  
  expect_error(predict(model_loaded, X, verbose = FALSE))
  
  unlink(tmp_path)
})

test_that("bartMachine handles missing data in both training and prediction", {
  set.seed(BART_TESTS$seed)
  n <- 100
  p <- 5
  X <- data.frame(matrix(runif(n * p), ncol = p))
  
  # Introduce missingness in training data
  X[1:10, 1] <- NA
  X[11:20, 2] <- NA
  
  y <- 5 * ifelse(is.na(X[, 1]), 0.5, X[, 1]) + 2 * X[, 3] + rnorm(n, sd = 0.1)
  
  # Train model with missing data support
  model <- bartMachine(
    X, 
    y,
    num_trees = 10,
    num_burn_in = 50,
    num_iterations_after_burn_in = 50,
    use_missing_data = TRUE,
    verbose = FALSE
  )
  
  expect_s3_class(model, "bartMachine")
  
  # Create prediction data with missing values
  X_pred <- data.frame(matrix(runif(20 * p), ncol = p))
  colnames(X_pred) <- colnames(X)
  X_pred[1:5, 1] <- NA
  X_pred[6:10, 3] <- NA
  
  # Predict on data with missing values
  preds <- predict(model, X_pred)
  expect_length(preds, 20)
  expect_false(any(is.na(preds)))
})

test_that("bartMachine handles missing data with mean imputation", {
  set.seed(BART_TESTS$seed)
  n <- 100
  X <- data.frame(matrix(runif(n * 3), ncol = 3))
  X[1:10, 1] <- NA
  y <- 3 * X[, 2] + rnorm(n, sd = 0.1)
  
  model <- bartMachine(
    X, 
    y,
    num_trees = 5,
    num_burn_in = 20,
    num_iterations_after_burn_in = 20,
    replace_missing_data_with_x_j_bar = TRUE,
    verbose = FALSE
  )
  
  expect_s3_class(model, "bartMachine")
  
  X_pred <- data.frame(matrix(runif(10 * 3), ncol = 3))
  colnames(X_pred) <- colnames(X)
  X_pred[1, 1] <- NA
  
  preds <- predict(model, X_pred)
  expect_length(preds, 10)
})

test_that("bartMachine warns on missing data in prediction when feature is off", {
  set.seed(BART_TESTS$seed)
  n <- 50
  X <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- 2 * X$x1 + rnorm(n, sd = 0.1)
  
  # Train model WITHOUT missing data support
  model <- bartMachine(
    X, 
    y,
    num_trees = 5,
    num_burn_in = 10,
    num_iterations_after_burn_in = 10,
    use_missing_data = FALSE,
    verbose = FALSE
  )
  
  X_pred <- data.frame(x1 = runif(5), x2 = runif(5))
  X_pred[1, 1] <- NA
  
  # Should issue a warning but still predict
  expect_warning(predict(model, X_pred), "rows omitted due to missing data")
})

test_that("bartMachine handles missing data in classification", {
  set.seed(BART_TESTS$seed)
  n <- 100
  X <- data.frame(x1 = runif(n), x2 = runif(n))
  X[1:10, 1] <- NA
  y <- factor(ifelse(ifelse(is.na(X$x1), 0.5, X$x1) + X$x2 > 1, "yes", "no"))
  
  model <- bartMachine(
    X, 
    y,
    num_trees = 5,
    num_burn_in = 20,
    num_iterations_after_burn_in = 20,
    use_missing_data = TRUE,
    verbose = FALSE
  )
  
  expect_s3_class(model, "bartMachine")
  
  X_pred <- data.frame(x1 = c(NA, 0.1, 0.9), x2 = c(0.5, 0.5, 0.5))
  preds <- predict(model, X_pred, type = "prob")
  expect_length(preds, 3)
  expect_false(any(is.na(preds)))
})

test_that("summary.bartMachine produces output", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(X, y, 
                       num_trees = 10, 
                       num_burn_in = 5, 
                       num_iterations_after_burn_in = 5, 
                       verbose = FALSE)
  
  # Capture output to ensure it prints something and doesn't crash
  out <- capture.output(summary(model))
  expect_true(length(out) > 0)
  expect_true(any(grepl("bartMachine", out, ignore.case = TRUE)))
})

test_that("print.bartMachine produces output", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(X, y, 
                       num_trees = 10, 
                       num_burn_in = 5, 
                       num_iterations_after_burn_in = 5, 
                       verbose = FALSE)
  
  out <- capture.output(print(model))
  expect_true(length(out) > 0)
  expect_true(any(grepl("bartMachine", out, ignore.case = TRUE)))
})

#S3 predict method
#' Make a prediction on data using a BART object
#'
#' @description
#' Makes a prediction on new data given a fitted BART model for regression or classification.
#' @param object An object of class ``bartMachine''.
#' @param new_data A data frame where each row is an observation to predict. The column names
#'   should be the same as the column names of the training data.
#' @param type Only relevant if the bartMachine model is classification. The type can be ``prob'' which will
#'   return the estimate of \eqn{P(Y = 1)}(the ``positive'' class) or ``class'' which will return the best guess as to the
#'   class of the object, in the original label, based on if the probability estimate is greater
#'   than \code{prob_rule_class}. Default is ``prob.''
#' @param prob_rule_class The rule to determine when the class estimate is \eqn{Y = 1} (the ``positive'' class) based on the probability estimate. This
#'   defaults to what was originally specified in the \code{bart_machine} object.
#' @param verbose Prints out prediction-related messages. Currently in use only for probability predictions to let the user know which class
#'   is being predicted. Default is \code{TRUE}.
#' @param ... Parameters that are ignored.
#'
#' @return
#' If regression, a numeric vector of \code{y_hat}, the best guess as to the response. If classification and \code{type = ``prob''},
#' a numeric vector of \code{p_hat}, the best guess as to the probability of the response class being  the ''positive'' class. If classification and
#' \code{type = ''class''}, a character vector of the best guess of the response's class labels.
#'
#' @seealso
#' \code{\link{bart_predict_for_test_data}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' #Regression example
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' ##make predictions on the training data
#' y_hat = predict(bart_machine, X)
#' 
#' #Classification example
#' data(iris)
#' iris2 = iris[51 : 150, ] #do not include the third type of flower for this example
#' iris2$Species = factor(iris2$Species)
#' bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)
#' 
#' ##make probability predictions on the training data
#' p_hat = predict(bart_machine, X)
#' 
#' ##make class predictions on test data
#' y_hat_class = predict(bart_machine, X, type = "class")
#' 
#' ##make class predictions on test data conservatively for ''versicolor''
#' y_hat_class_conservative = predict(bart_machine, X, type = "class", prob_rule_class = 0.9)
#' }
#' @export
predict.bartMachine = function(object, new_data, type = "prob", prob_rule_class = NULL, verbose = TRUE, ...){
  # Validate arguments
  assert_class(object, "bartMachine")
  assert_data_frame(new_data)
  assert_choice(type, c("prob", "class"))
  assert_number(prob_rule_class, lower = 0, upper = 1, null.ok = TRUE)
  assert_flag(verbose)

	check_serialization(object) #ensure the Java object exists and fire an error if not
	
	if(!(type %in% c("prob", "class"))){
		stop("For classification, type must be either \"prob\" or \"class\". ")
	}
  
	if (object$pred_type == "regression"){	
		bart_machine_get_posterior_mean(object, new_data, verbose = verbose)$y_hat
	} else { ##classification
	    if (type == "prob"){
			if (isTRUE(verbose)){
				cat("predicting probabilities where \"", object$y_levels[1], "\" is considered the target level...\n", sep = "")
			}			
	    	bart_machine_get_posterior_mean(object, new_data, verbose = verbose)$y_hat
	    } else {
	    	labels = bart_machine_get_posterior_mean(object, new_data, verbose = verbose)$y_hat > ifelse(is.null(prob_rule_class), object$prob_rule_class, prob_rule_class)
	      	#return whatever the raw y_levels were
	      	labels_to_y_levels(object, labels)      
	    }
	}	
}

##private function
labels_to_y_levels = function(bart_machine, labels){
	factor(ifelse(labels == TRUE, bart_machine$y_levels[1], bart_machine$y_levels[2]), levels = bart_machine$y_levels)
}

.bartMachine_pred_cache <- new.env(parent = emptyenv())
.bartMachine_pred_cache$processed_data_array <- NULL

prepare_prediction_data = function(bart_machine, new_data){
	if (!"data.frame"%in%class(new_data)){		
		stop("\"new_data\" needs to be a data frame with the same column names as the training data.")
	}
	clean_data = new_data
	if (!bart_machine$use_missing_data && !bart_machine$replace_missing_data_with_x_j_bar){
		if (anyNA(clean_data)){
			nrow_before = nrow(clean_data)
			clean_data = na.omit(clean_data)
			if (nrow_before > nrow(clean_data)){
				warning(nrow_before - nrow(clean_data), " rows omitted due to missing data. Try using the missing data feature in \"build_bart_machine\" to be able to predict on all observations.", call. = FALSE)
			}
		}
	}
	
	if (nrow(clean_data) == 0){
		stop("No rows to predict.\n")
	}
	
	cache = .bartMachine_pred_cache
	if (!is.null(cache$java_bart_machine) &&
		identical(cache$java_bart_machine, bart_machine$java_bart_machine) &&
		!is.null(cache$raw_data) &&
		identical(cache$raw_data, clean_data)) {
		return(list(data = cache$processed_data, array = cache$processed_data_array))
	}
	
	if (identical(clean_data, bart_machine$X) &&
		(bart_machine$use_missing_data || bart_machine$replace_missing_data_with_x_j_bar || !anyNA(clean_data))) {
		processed_data = bart_machine$model_matrix_training_data[, 1 : bart_machine$p, drop = FALSE]
	} else {
		processed_data = pre_process_new_data(clean_data, bart_machine)
	}
	
	cache$java_bart_machine = bart_machine$java_bart_machine
	cache$raw_data = clean_data
	cache$processed_data = processed_data
	cache$processed_data_array = .jarray(processed_data, dispatch = TRUE)
	list(data = processed_data, array = cache$processed_data_array)
}

bart_machine_get_posterior_mean = function(bart_machine, new_data, verbose = TRUE){
  # Validate arguments
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(new_data)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	prepared_data = prepare_prediction_data(bart_machine, new_data)
	new_data = prepared_data$data
		
	y_hat = .jcall(
		bart_machine$java_bart_machine,
		"[D",
		"getPosteriorMeanForPrediction",
		prepared_data$array,
		as.integer(bart_machine_num_cores()),
		simplify = TRUE
	)
	list(y_hat = y_hat, X = new_data)
}

##utility function for predicting when test outcomes are known
#' Predict for Test Data with Known Outcomes
#'
#' @description
#' Utility wrapper function for computing out-of-sample metrics for a BART model when the test set outcomes are known.
#' @param bart_machine An object of class ``bartMachine''.
#' @param Xtest Data frame for test data containing rows at which predictions are to be made. Colnames should match that of the training data.
#' @param ytest Actual outcomes for test data.
#' @param prob_rule_class Threshold for classification.
#' @param verbose If TRUE, prints prediction-related messages.
#'
#' @return
#' For regression models, a list with the following components is returned:
#' 
#'   \item{y_hat}{Predictions (as posterior means) for the test observations.}
#'   \item{L1_err}{L1 error for predictions.}
#'   \item{L2_err}{L2 error for predictions.}
#'   \item{rmse}{RMSE for predictions.}
#' 
#' For classification models, a list with the following components is returned:
#' 
#'   \item{y_hat}{Class predictions for the test observations.}
#'   \item{p_hat}{Probability estimates for the test observations.}
#'   \item{confusion_matrix}{A confusion matrix for the test observations.}
#' 
#' %% ...
#'
#' @seealso
#' \code{\link{predict}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 250
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##split into train and test
#' train_X = X[1 : 200, ]
#' test_X = X[201 : 250, ]
#' train_y = y[1 : 200]
#' test_y = y[201 : 250]
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(train_X, train_y)
#' 
#' #explore performance on test data
#' oos_perf = bart_predict_for_test_data(bart_machine, test_X, test_y)
#' print(oos_perf$rmse)
#' }
#' @export
bart_predict_for_test_data = function(bart_machine, Xtest, ytest, prob_rule_class = NULL, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(Xtest)
  assert_atomic_vector(ytest)
  assert_number(prob_rule_class, lower = 0, upper = 1, null.ok = TRUE)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (bart_machine$pred_type == "regression"){ #regression list
	  ytest_hat = predict(bart_machine, Xtest, verbose = verbose)
		n = nrow(Xtest)
		L2_err = sum((ytest - ytest_hat)^2)
		
		list(
				y_hat = ytest_hat,
				L1_err = sum(abs(ytest - ytest_hat)),
				L2_err = L2_err,
				rmse = sqrt(L2_err / n),
				e = ytest - ytest_hat
		)
	} else { ##classification list
	    if (!inherits(ytest, "factor")){
			stop("ytest must be a factor.")
		}
	    if (!all(levels(ytest) %in% bart_machine$y_levels)){
			stop("New factor level not seen in training introduced. Please remove.")
		}
		
	    ptest_hat = predict(bart_machine, Xtest, type = "prob", verbose = verbose)
	    ytest_labels = ptest_hat > ifelse(is.null(prob_rule_class), bart_machine$prob_rule_class, prob_rule_class)
	    ytest_hat = labels_to_y_levels(bart_machine, ytest_labels)
    
		confusion_matrix = as.data.frame(matrix(NA, nrow = 3, ncol = 3))
		rownames(confusion_matrix) = c(paste("actual", bart_machine$y_levels), "use errors")
		colnames(confusion_matrix) = c(paste("predicted", bart_machine$y_levels), "model errors")		
		confusion_matrix[1 : 2, 1 : 2] = as.integer(table(ytest, ytest_hat)) 
		confusion_matrix[3, 1] = round(confusion_matrix[2, 1] / (confusion_matrix[1, 1] + confusion_matrix[2, 1]), 3)
		confusion_matrix[3, 2] = round(confusion_matrix[1, 2] / (confusion_matrix[1, 2] + confusion_matrix[2, 2]), 3)
		confusion_matrix[1, 3] = round(confusion_matrix[1, 2] / (confusion_matrix[1, 1] + confusion_matrix[1, 2]), 3)
		confusion_matrix[2, 3] = round(confusion_matrix[2, 1] / (confusion_matrix[2, 1] + confusion_matrix[2, 2]), 3)
		confusion_matrix[3, 3] = round((confusion_matrix[1, 2] + confusion_matrix[2, 1]) / sum(confusion_matrix[1 : 2, 1 : 2]), 3)
		
		list(y_hat = ytest_hat, p_hat = ptest_hat, confusion_matrix = confusion_matrix)
	}
}

##get full set of samples from posterior distribution of f(x)
#' Get Full Posterior Distribution
#'
#' @description
#' Generates draws from posterior distribution of \eqn{\hat{f}(x)} for a specified set of observations.
#' @param bart_machine An object of class ``bartMachine''.
#' @param new_data A data frame containing observations at which draws from posterior distribution of \eqn{\hat{f}(x)} are to be obtained.
#' @param verbose If TRUE, prints preprocessing-related messages.
#'
#' @return
#' Returns a list with the following components:
#' %%  If it is a LIST, use
#'   \item{y_hat}{Posterior mean estimates. For regression, the estimates have the same units as the response. For classification, the estimates are probabilities.}
#'   \item{new_data}{The data frame with rows at which the posterior draws are to be generated. Column names should match that of the training data.}
#'   \item{y_hat_posterior_samples}{The full set of posterior samples of size \code{num_iterations_after_burn_in} for each observation. For regression, the estimates have the same units as the response. For classification, the estimates are probabilities.}
#' %% ...
#'
#' @seealso
#' \code{\link{calc_credible_intervals}}, \code{\link{calc_prediction_intervals}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #Regression example
#' 
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #get posterior distribution
#' posterior = bart_machine_get_posterior(bart_machine, X)
#' print(posterior$y_hat)
#' 
#' 
#' #Classification example
#' 
#' #get data and only use 2 factors
#' data(iris)
#' iris2 = iris[51:150,]
#' iris2$Species = factor(iris2$Species)
#' 
#' #build BART classification model
#' bart_machine = bartMachine(iris2[ ,1 : 4], iris2$Species)
#' 
#' #get posterior distribution
#' posterior = bart_machine_get_posterior(bart_machine, iris2[ ,1 : 4])
#' print(posterior$y_hat)
#' }
#' @export
bart_machine_get_posterior = function(bart_machine, new_data, verbose = TRUE){
  # Validate arguments
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(new_data)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	prepared_data = prepare_prediction_data(bart_machine, new_data)
	new_data = prepared_data$data
	
	y_hat_posterior_samples = 
		.jcall(
			bart_machine$java_bart_machine,
			"[[D",
			"getGibbsSamplesForPrediction",
			prepared_data$array,
			as.integer(bart_machine_num_cores()),
			simplify = TRUE
		)
	
	#to get y_hat.. just take straight mean of posterior samples, alternatively, we can let java do it if we want more bells and whistles
	y_hat = rowMeans(y_hat_posterior_samples)
	
	list(y_hat = y_hat, X = new_data, y_hat_posterior_samples = y_hat_posterior_samples)
}

##compute credible intervals
#' Calculate Credible Intervals
#'
#' @description
#' Generates credible intervals for \eqn{\hat{f}(x)} for a specified set of observations.
#'
#' @details
#' This interval is the appropriate quantiles  based on the confidence level, \code{ci_conf}, of the predictions
#' for each of the Gibbs samples post-burn in.
#' @param bart_machine An object of class ``bartMachine''.
#' @param new_data A data frame containing observations at which credible intervals for \eqn{\hat{f}(x)} are to be computed.
#' @param ci_conf Confidence level for the credible intervals. The default is 95\%.
#'
#' @return
#' Returns a matrix of the lower and upper bounds of the credible intervals for each observation in \code{new_data}.
#'
#' @seealso
#' \code{\link{calc_prediction_intervals}}, \code{\link{bart_machine_get_posterior}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #get credible interval
#' cred_int = calc_credible_intervals(bart_machine, X)
#' print(head(cred_int))
#' }
#' @export
calc_credible_intervals = function(bart_machine, new_data, ci_conf = 0.95){
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(new_data)
  assert_number(ci_conf, lower = 0, upper = 1)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	prepared_data = prepare_prediction_data(bart_machine, new_data)

	.jcall(
		bart_machine$java_bart_machine,
		"[[D",
		"getCredibleIntervalsForPrediction",
		prepared_data$array,
		as.double(ci_conf),
		as.integer(bart_machine_num_cores()),
		simplify = TRUE
	)
}

##compute prediction intervals
#' Calculate Prediction Intervals
#'
#' @description
#' Generates prediction intervals for \eqn{\hat{y}} for a specified set of observations.
#'
#' @details
#' Credible intervals (see \code{\link{calc_credible_intervals}}) are the appropriate quantiles of the prediction
#' for each of the Gibbs samples post-burn in. Prediction intervals also make use of the noise estimate at each Gibbs
#' sample and hence are wider. For each Gibbs sample, we record the \eqn{\hat{y}} estimate of the response and the
#' \eqn{\hat{\sigma^2}} estimate of the noise variance. We then sample \code{normal_samples_per_gibbs_sample} times
#' from a \eqn{N(\hat{y}, \hat{\sigma^2})} random variable to simulate many possible disturbances for that Gibbs sample.
#' Then, all \code{normal_samples_per_gibbs_sample} times the number of Gibbs sample post burn-in are collected and the
#' appropriate quantiles are taken based on the confidence level, \code{pi_conf}.
#' @param bart_machine An object of class ``bartMachine''.
#' @param new_data A data frame containing observations at which prediction intervals for \eqn{\hat{y}} are to be computed.
#' @param pi_conf Confidence level for the prediction intervals. The default is 95\%.
#' @param num_samples_per_data_point The number of samples taken from the predictive distribution. The default is 1000.
#'
#' @return
#' Returns a matrix of the lower and upper bounds of the prediction intervals for each observation in \code{new_data}.
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#'
#' @seealso
#' \code{\link{calc_credible_intervals}}, \code{\link{bart_machine_get_posterior}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #get prediction interval
#' pred_int = calc_prediction_intervals(bart_machine, X)
#' print(head(pred_int))
#' }
#' @export
calc_prediction_intervals = function(bart_machine, new_data, pi_conf = 0.95, num_samples_per_data_point = 1000){
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(new_data)
  assert_number(pi_conf, lower = 0, upper = 1)
  assert_int(num_samples_per_data_point, lower = 1)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (bart_machine$pred_type == "classification"){
		stop("Prediction intervals are not possible for classification.")
	}
  
	prepared_data = prepare_prediction_data(bart_machine, new_data)
	.jcall(
		bart_machine$java_bart_machine,
		"[[D",
		"getPredictionIntervalsForPrediction",
		prepared_data$array,
		as.double(pi_conf),
		as.integer(num_samples_per_data_point),
		as.integer(bart_machine_num_cores()),
		simplify = TRUE
	)
}

#' Create an array of BART models for the same data.
#'
#' @description
#' If BART creates models that are variable,
#' running many on the same dataset and averaging is a good strategy.
#' This function is a convenience method for this procedure.
#' @param bart_machine An object of class ``bartMachine''.
#' @param R The number of replicated BART models in the array.
#'
#' @return
#' A \code{bartMachineArr} object which is just a list of the \code{R} bartMachine models.
#'
#' @author
#' Adam Kapelner
#'
#' @examples
#' \dontrun{
#' #Regression example
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' bart_machine_arr = bartMachineArr(bart_machine)
#' 
#' #Classification example
#' data(iris)
#' iris2 = iris[51 : 150, ] #do not include the third type of flower for this example
#' iris2$Species = factor(iris2$Species)
#' bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)
#' bart_machine_arr = bartMachineArr(bart_machine)
#' }
#' @export
bartMachineArr = function(bart_machine, R = 10){
  assert_class(bart_machine, "bartMachine")
  assert_count(R, positive = TRUE)

	arr = list()
	arr[[1]] = bart_machine
	for (i in 2 : R){
		arr[[i]] = bart_machine_duplicate(bart_machine)
	}
	class(arr) = "bartMarchineArr"
	arr
}

#' Make a prediction on data using a BART array object
#'
#' @description
#' Makes a prediction on new data given an array of fitted BART model for
#' regression or classification. If BART creates models that are variable,
#' running many and averaging is a good strategy. It is well known that the
#' Gibbs sampler gets locked into local modes at times. This is a way
#' to average over many chains.
#' @param object An object of class ``bartMachineArr''.
#' @param new_data A data frame where each row is an observation to predict. The column names
#'   should be the same as the column names of the training data.
#' @param ... Not supported. Note that parameters \code{type} and \code{prob_rule_class} for
#'   \code{\link{predict.bartMachine}} are not supported.
#'
#' @return
#' If regression, a numeric vector of \code{y_hat}, the best guess as to the response. If classification and \code{type = ``prob''},
#' a numeric vector of \code{p_hat}, the best guess as to the probability of the response class being  the ''positive'' class. If classification and
#' \code{type = ''class''}, a character vector of the best guess of the response's class labels.
#'
#' @seealso
#' \code{\link{predict.bartMachine}}
#'
#' @author
#' Adam Kapelner
#'
#' @examples
#' \dontrun{
#' #Regression example
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' bart_machine_arr = bartMachineArr(bart_machine)
#' 
#' ##make predictions on the training data
#' y_hat = predict(bart_machine_arr, X)
#' 
#' #Classification example
#' data(iris)
#' iris2 = iris[51 : 150, ] #do not include the third type of flower for this example
#' iris2$Species = factor(iris2$Species)
#' bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)
#' bart_machine_arr = bartMachineArr(bart_machine)
#' 
#' ##make probability predictions on the training data
#' p_hat = predict_bartMachineArr(bart_machine_arr, iris2[ ,1:4])
#' }
#' @export
predict_bartMachineArr = function(object, new_data, ...){
  assert_class(object, "bartMarchineArr")
  assert_data_frame(new_data)

	R = length(object)
	n_star = nrow(new_data)
	predicts = matrix(NA, nrow = n_star, ncol = R)
	for (r in 1 : R){
		predicts[, r] = predict(object[[r]], new_data, ...)
	}
	rowMeans(predicts)
}

source("../testthat/helper-bartmachine-models.R")

test_that("bartMachine builds regression and classification models", {
  reg_model <- build_regression_model()
  expect_s3_class(reg_model, "bartMachine")
  expect_equal(reg_model$pred_type, "regression")

  class_model <- build_classification_model()
  expect_s3_class(class_model, "bartMachine")
  expect_equal(class_model$pred_type, "classification")
})

test_that("build_bart_machine handles X/y and Xy inputs", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(
    x1 = runif(BART_TESTS$small_data_n),
    x2 = runif(BART_TESTS$small_data_n)
  )
  y <- X$x1 + rnorm(BART_TESTS$small_data_n, sd = BART_TESTS$small_data_sd)
  model_xy <- build_bart_machine(X, y,
    num_trees = BART_TESTS$small_model_num_trees,
    num_burn_in = BART_TESTS$small_model_burn_in,
    num_iterations_after_burn_in = BART_TESTS$small_model_iter,
    flush_indices_to_save_RAM = FALSE,
    verbose = BART_TESTS$verbose
  )
  expect_s3_class(model_xy, "bartMachine")

  Xy <- data.frame(X, y = y)
  model_Xy <- build_bart_machine(Xy = Xy,
    num_trees = BART_TESTS$small_model_num_trees,
    num_burn_in = BART_TESTS$small_model_burn_in,
    num_iterations_after_burn_in = BART_TESTS$small_model_iter,
    flush_indices_to_save_RAM = FALSE,
    verbose = BART_TESTS$verbose
  )
  expect_s3_class(model_Xy, "bartMachine")
})

test_that("core helpers set and read cores", {
  set_bart_machine_num_cores(BART_TESTS$core_count, verbose = BART_TESTS$verbose)
  expect_equal(bart_machine_num_cores(), BART_TESTS$core_count)
  set_bart_machine_num_cores(1, verbose = BART_TESTS$verbose)
})

test_that("dummify_data converts factors", {
  df <- data.frame(a = factor(c("a", "b")), b = BART_TESTS$dummify_b_values)
  dummified <- dummify_data(df)
  expect_true(is.data.frame(dummified))
  expect_true(any(grepl("^a_", names(dummified))))
})

test_that("bartMachineArr and predict_bartMachineArr work", {
  reg_model <- build_regression_model()
  arr <- bartMachineArr(reg_model, R = BART_TESTS$arr_r)
  expect_true(is.list(arr))
  preds <- predict_bartMachineArr(arr, reg_model$X)
  expect_length(preds, nrow(reg_model$X))
})

test_that("predict methods return expected shapes", {
  reg_model <- build_regression_model()
  preds <- predict(reg_model, reg_model$X, verbose = BART_TESTS$verbose)
  expect_length(preds, nrow(reg_model$X))

  class_model <- build_classification_model()
  probs <- predict(class_model, class_model$X, type = "prob", verbose = BART_TESTS$verbose)
  expect_length(probs, nrow(class_model$X))
  classes <- predict(class_model, class_model$X, type = "class", verbose = BART_TESTS$verbose)
  expect_length(classes, nrow(class_model$X))
})

.onLoad = function(libname, pkgname) {
  .jpackage(pkgname, lib.loc = libname)
  assign("bartMachine_globals", new.env(), envir = parent.env(environment()))
}

# Returns the Java *feature* version as an integer (e.g., 8, 11, 17, 21),
# or NA_integer_ if it can't be parsed.
java_feature_version_from_runtime <- function(jv) {
  jv <- as.character(jv)[1L]
  if (is.na(jv) || !nzchar(jv)) return(NA_integer_)

  # Legacy: "1.8.0_402..." -> feature version is 8
  if (startsWith(jv, "1.")) {
    m <- regexec("^1\\.(\\d+)", jv)
    g <- regmatches(jv, m)[[1L]]
    if (length(g) >= 2L) return(as.integer(g[2L]))
    return(NA_integer_)
  }

  # Modern: starts with feature version, possibly followed by ".0.1", "+35", "-ea", etc.
  m <- regexec("^(\\d+)", jv)
  g <- regmatches(jv, m)[[1L]]
  if (length(g) >= 2L) return(as.integer(g[2L]))

  NA_integer_
}

.onAttach = function(libname, pkgname){
	jv <- tryCatch(
	  .jcall("java/lang/System", "S", "getProperty", "java.runtime.version"),
	  error = function(e) NA_character_
	)

	fv <- java_feature_version_from_runtime(jv)

	if (is.na(fv)) {
	  packageStartupMessage(
	    sprintf("Could not determine Java version from java.runtime.version='%s'. Java 21+ is required.", jv)
	  )
	} else if (fv < 21L) {
	  packageStartupMessage(
	    sprintf("Java 21+ is required, but detected feature version %d (java.runtime.version='%s').", fv, jv)
	  )
	}

  java_params = getOption("java.parameters")
  if (is.null(java_params) || !any(grepl("--add-modules=jdk.incubator.vector", java_params))){
	  packageStartupMessage("You did not set `options` correctly so bartMachine will not work. Restart R and run\n\noptions(java.parameters = c(\"-Xmx20g\", \"--add-modules=jdk.incubator.vector\", \"-XX:+UseZGC\"))\nlibrary(bartMachine)\n")
  } else {
	  num_gigs_ram_available = .jcall(.jnew("java/lang/Runtime"), "J", "maxMemory") / 1e9
	  more_memory_message = if (num_gigs_ram_available < 10){
												  	"\n\nIf you run out of memory, restart R, and use \n\n'options(java.parameters = \"-Xmx20g\")' for 20GB of RAM before you call\n'library(bartMachine)'.\n"
												  } else {
												  	""
												  }
	  packageStartupMessage(
			paste("Welcome to bartMachine v", packageVersion("bartMachine"), 
					"! You have ", round(num_gigs_ram_available, 2), 
					"GB memory available.", 
					more_memory_message,
					"You can safely ignore the \"WARNING: Using incubator modules: jdk.incubator.vector\" above.",
					sep = "")
	  )
  }
}

options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))
options(warn = 2)

library(bartMachine)

if (!exists("BART_TESTS")) {
  BART_TESTS <- list(
    core_count = 4,
    pdf_width = 8,
    pdf_height = 6,
    seed = 11,
    verbose = FALSE,
    default_n = 120,
    default_p = 4,
    default_class_p = 3,
    reg_coef_1 = 3,
    reg_coef_2 = 2,
    reg_noise_sd = 0.2,
    model_num_trees = 20,
    model_burn_in = 10,
    model_iter = 20,
    class_threshold = 0.5,
    small_data_n = 30,
    small_data_sd = 0.1,
    small_model_num_trees = 10,
    small_model_burn_in = 5,
    small_model_iter = 10,
    dummify_b_values = c(1, 2),
    arr_r = 2,
    posterior_subset_n = 5,
    credible_interval_cols = 2,
    prediction_samples_per_point = 20,
    covariate_index = 1,
    reg_feature_index_1 = 1,
    reg_feature_index_2 = 2,
    class_feature_index = 1,
    first_dim_index = 1,
    num_permute_samples = 5,
    linearity_num_trees = 10,
    linearity_num_burn_in = 5,
    linearity_num_iter = 10,
    num_replicates_for_avg = 1,
    num_trees_bottleneck = 5,
    num_var_plot = 3,
    min_plot_count = 1,
    pd_plot_prop_data = 0.2,
    rmse_tree_list = c(5, 10),
    rmse_num_replicates = 1,
    var_sel_k_folds = 2,
    var_sel_num_reps_for_avg = 1,
    var_sel_num_permute_samples = 5,
    var_sel_num_trees_for_permute = 5,
    var_sel_num_trees_pred_cv = 5,
    cv_data_n = 40,
    cv_data_sd = 0.1,
    cv_num_tree_cvs = c(5, 10),
    cv_num_tree_cvs_single = c(5),
    cv_k_cvs = c(2),
    cv_k_folds = 2,
    cv_num_burn_in = 5,
    cv_num_iter = 10,
    cv_num_trees = 5,
    raw_node_g = 1,
    test_idx_n = 20,
    test_files = c(
      "test-exports-basic.R",
      "test-exports-diagnostics.R",
      "test-exports-model-selection.R"
    ),
    np_grid = list(
      list(n = 60, p = 3, num_trees = 20, num_burn_in = 10, num_iter = 20),
      list(n = 120, p = 4, num_trees = 20, num_burn_in = 10, num_iter = 20),
      list(n = 200, p = 6, num_trees = 30, num_burn_in = 15, num_iter = 30),
      list(n = 240, p = 8, num_trees = 35, num_burn_in = 15, num_iter = 35),
      list(n = 320, p = 10, num_trees = 40, num_burn_in = 20, num_iter = 40),
      list(n = 400, p = 12, num_trees = 45, num_burn_in = 20, num_iter = 45),
      list(n = 480, p = 15, num_trees = 50, num_burn_in = 20, num_iter = 50)
    )
  )
}

get_test_env_int <- function(name, default) {
  val <- Sys.getenv(name, "")
  if (nzchar(val)) {
    as.integer(val)
  } else {
    default
  }
}

with_plot_device <- function(expr) {
  path <- tempfile(fileext = ".pdf")
  grDevices::pdf(path, width = BART_TESTS$pdf_width, height = BART_TESTS$pdf_height)
  on.exit(grDevices::dev.off(), add = TRUE)
  force(expr)
  invisible(path)
}

build_regression_model <- function() {
  set.seed(BART_TESTS$seed)
  n <- as.integer(Sys.getenv("N_TEST", as.character(BART_TESTS$default_n)))
  p <- as.integer(Sys.getenv("P_TEST", as.character(BART_TESTS$default_p)))
  X <- data.frame(matrix(runif(n * p), ncol = p))
  y <- BART_TESTS$reg_coef_1 * X[, BART_TESTS$reg_feature_index_1] +
    BART_TESTS$reg_coef_2 * X[, BART_TESTS$reg_feature_index_2] +
    rnorm(n, sd = BART_TESTS$reg_noise_sd)
  set_bart_machine_num_cores(BART_TESTS$core_count, verbose = BART_TESTS$verbose)
  model <- bartMachine(
    X,
    y,
    num_trees = get_test_env_int("MODEL_NUM_TREES", BART_TESTS$model_num_trees),
    num_burn_in = get_test_env_int("MODEL_BURN_IN", BART_TESTS$model_burn_in),
    num_iterations_after_burn_in = get_test_env_int("MODEL_ITER", BART_TESTS$model_iter),
    flush_indices_to_save_RAM = FALSE,
    verbose = BART_TESTS$verbose
  )
  set_bart_machine_num_cores(1, verbose = BART_TESTS$verbose)
  model
}

build_classification_model <- function() {
  set.seed(BART_TESTS$seed)
  n <- as.integer(Sys.getenv("N_CLASS_TEST", Sys.getenv("N_TEST", as.character(BART_TESTS$default_n))))
  p <- as.integer(Sys.getenv("P_CLASS_TEST", Sys.getenv("P_TEST", as.character(BART_TESTS$default_class_p))))
  X <- data.frame(matrix(runif(n * p), ncol = p))
  y <- factor(ifelse(
    X[, BART_TESTS$class_feature_index] > BART_TESTS$class_threshold,
    "yes",
    "no"
  ))
  set_bart_machine_num_cores(BART_TESTS$core_count, verbose = BART_TESTS$verbose)
  bartMachine(
    X,
    y,
    num_trees = get_test_env_int("MODEL_NUM_TREES", BART_TESTS$model_num_trees),
    num_burn_in = get_test_env_int("MODEL_BURN_IN", BART_TESTS$model_burn_in),
    num_iterations_after_burn_in = get_test_env_int("MODEL_ITER", BART_TESTS$model_iter),
    flush_indices_to_save_RAM = FALSE,
    verbose = BART_TESTS$verbose
  )
}

##variable selection procedures from Bleich et al. (2013)
#' Perform Variable Selection using Three Threshold-based Procedures
#'
#' @description
#' Performs variable selection using the three thresholding methods introduced in Bleich et al. (2013).
#'
#' @details
#' See Bleich et al. (2013) for a complete description of the procedures outlined above as well as the corresponding vignette for a brief summary with examples.
#' @param bart_machine An object of class ``bartMachine''.
#' @param num_reps_for_avg Number of replicates to over over to for the BART model's variable inclusion proportions.
#' @param num_permute_samples Number of permutations of the response to be made to generate the ``null'' permutation distribution.
#' @param num_trees_for_permute Number of trees to use in the variable selection procedure. As with \cr \code{\link{investigate_var_importance}}, a small number of trees should be used to force variables to compete for entry into the model. Note that this number is used to estimate both the ``true'' and ``null'' variable inclusion proportions.
#' @param alpha Cut-off level for the thresholds.
#' @param plot If TRUE, a plot showing which variables are selected by each of the procedures is generated.
#' @param num_var_plot Number of variables (in order of decreasing variable inclusion proportion) to be plotted.
#' @param bottom_margin A display parameter that adjusts the bottom margin of the graph if labels are clipped. The scale of this parameter is the same as set with \code{par(mar = c(....))} in R.
#'   Higher values allow for more space if the crossed covariate names are long. Note that making this parameter too large will prevent plotting and the plot function in R will throw an error.
#' @param verbose If TRUE, prints progress messages.
#'
#' @return
#' Invisibly, returns a list with the following components:
#' 
#'   \item{important_vars_local_names}{Names of the variables chosen by the Local procedure.}
#'   \item{important_vars_global_max_names}{Names of the variables chosen by the Global Max procedure.}
#'     \item{important_vars_global_se_names}{Names of the variables chosen by the Global SE procedure.}
#'     \item{important_vars_local_col_nums}{Column numbers of the variables chosen by the Local procedure.}
#'   \item{important_vars_global_max_col_nums}{Column numbers of the variables chosen by the Global Max procedure.}
#'     \item{important_vars_global_se_col_nums}{Column numbers of the variables chosen by the Global SE procedure.}
#'     \item{var_true_props_avg}{The variable inclusion proportions for the actual data.}
#'   \item{permute_mat}{The permutation distribution generated by permuting the response vector.}
#'
#' @references
#' J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
#' Additive Regression Trees. ArXiv e-prints, 2013.
#' 
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#'
#' @seealso
#' \code{\link{var_selection_by_permute}}, \code{\link{investigate_var_importance}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' Although the reference only explores regression settings, this procedure is applicable to both regression and classification problems.
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 300
#' p = 20 ##15 useless predictors
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model (not actually used in variable selection)
#' bart_machine = bartMachine(X, y)
#' 
#' #variable selection
#' var_sel = var_selection_by_permute(bart_machine)
#' print(var_sel$important_vars_local_names)
#' print(var_sel$important_vars_global_max_names)
#' }
#' @export
var_selection_by_permute = function(bart_machine, num_reps_for_avg = 10, num_permute_samples = 100, num_trees_for_permute = 20, alpha = 0.05, plot = TRUE, num_var_plot = Inf, bottom_margin = 10, verbose = TRUE){	
  assert_class(bart_machine, "bartMachine")
  assert_int(num_reps_for_avg, lower = 1)
  assert_int(num_permute_samples, lower = 1)
  assert_int(num_trees_for_permute, lower = 1)
  assert_number(alpha, lower = 0, upper = 1)
  assert_flag(plot)
  assert_number(num_var_plot, lower = 1)
  assert_number(bottom_margin, lower = 0)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not

	permute_mat = matrix(NA, nrow = num_permute_samples, ncol = bart_machine$p) ##set up permute mat
	colnames(permute_mat) = bart_machine$training_data_features_with_missing_features

	if (bart_machine$pred_type == "regression") {
		# Java path: builds all numRepsForAvg + numPermuteSamples models in parallel
		raw_mat = .jcall(bart_machine$java_bart_machine, "[[D", "runVarImportancePermutations",
				as.integer(num_reps_for_avg), as.integer(num_permute_samples),
				as.integer(num_trees_for_permute), "splits", simplify = TRUE)
		# raw_mat is (num_permute_samples+1) x p; row 1 = avg props, rows 2..end = permutations
		var_true_props_avg = raw_mat[1, ]
		names(var_true_props_avg) = bart_machine$training_data_features_with_missing_features
		permute_mat = raw_mat[2:(num_permute_samples + 1), , drop = FALSE]
		colnames(permute_mat) = bart_machine$training_data_features_with_missing_features
	} else {
		if (verbose){ cat("avg") }
		var_true_props_avg = get_averaged_true_var_props(bart_machine, num_reps_for_avg, num_trees_for_permute, verbose = verbose) ##get props from actual data
		if (verbose){ cat("null") }
		for (b in 1 : num_permute_samples){
			permute_mat[b, ] = get_null_permute_var_importances(bart_machine, num_trees_for_permute, verbose = verbose) ##build null permutation distribution
		}
		if (verbose){ cat("\n") }
	}

	#now sort from high to low
	var_true_props_avg = sort(var_true_props_avg, decreasing = TRUE) ##sort props

	#sort permute mat
	permute_mat = permute_mat[, names(var_true_props_avg)]
	
    ##use local cutoff
	pointwise_cutoffs = as.numeric(matrixStats::colQuantiles(permute_mat, probs = 1 - alpha))
	important_vars_pointwise_names = names(var_true_props_avg[var_true_props_avg > pointwise_cutoffs & var_true_props_avg > 0])
	important_vars_pointwise_col_nums = match(important_vars_pointwise_names, bart_machine$training_data_features_with_missing_features)
	
    ##use global max cutoff
	row_maxes = matrixStats::rowMaxs(permute_mat)
	max_cut = quantile(row_maxes, 1 - alpha)
	important_vars_simul_max_names = names(var_true_props_avg[var_true_props_avg >= max_cut & var_true_props_avg > 0])	
	important_vars_simul_max_col_nums = match(important_vars_simul_max_names, bart_machine$training_data_features_with_missing_features)
	
    #use global se cutoff
	perm_mean = colMeans(permute_mat)
	perm_se = matrixStats::colSds(permute_mat)
	cover_constant = bisectK(tol = .01 , coverage = 1 - alpha, permute_mat = permute_mat, x_left = 1, x_right = 20, countLimit = 100, perm_mean = perm_mean, perm_se = perm_se)
	important_vars_simul_se_names = names(var_true_props_avg[which(var_true_props_avg >= perm_mean + cover_constant * perm_se & var_true_props_avg > 0)])	
	important_vars_simul_se_col_nums = match(important_vars_simul_se_names, bart_machine$training_data_features_with_missing_features)
	
	if (plot){
		if (is.infinite(num_var_plot) || num_var_plot > bart_machine$p){
			num_var_plot = bart_machine$p
		}
		non_zero_idx = which(var_true_props_avg > 0)
		if (length(non_zero_idx) < length(var_true_props_avg)) {
			warning(paste(length(which(var_true_props_avg == 0)), "covariates with inclusion proportions of 0 omitted from plots."))
		}
		if (length(non_zero_idx) > 0){
			non_zero_idx = non_zero_idx[seq_len(min(num_var_plot, length(non_zero_idx)))]
			plot_df = data.frame(
				index = seq_len(length(non_zero_idx)),
				variable = names(var_true_props_avg)[non_zero_idx],
				prop = as.numeric(var_true_props_avg[non_zero_idx]),
				pointwise_cutoff = as.numeric(pointwise_cutoffs[non_zero_idx]),
				perm_mean = as.numeric(perm_mean[non_zero_idx]),
				perm_se = as.numeric(perm_se[non_zero_idx])
			)
			plot_df$pointwise_selected = plot_df$prop > plot_df$pointwise_cutoff
			plot_df$simul_threshold = plot_df$perm_mean + cover_constant * plot_df$perm_se
			plot_df$simul_marker = ifelse(
				plot_df$prop >= max_cut,
				"global_max",
				ifelse(plot_df$prop > plot_df$simul_threshold, "global_se", "none")
			)

			y_max = max(c(plot_df$prop, plot_df$pointwise_cutoff, plot_df$simul_threshold, max_cut * 1.1), na.rm = TRUE)
			margin_cfg = ggplot2::margin(t = 5.5, r = 5.5, b = bottom_margin, l = 5.5)

			local_plot = ggplot2::ggplot(plot_df, ggplot2::aes(x = index, y = prop)) +
				ggplot2::geom_segment(
					ggplot2::aes(xend = index, y = 0, yend = pointwise_cutoff),
					color = "forestgreen"
				) +
				ggplot2::geom_point(ggplot2::aes(shape = pointwise_selected)) +
				ggplot2::scale_shape_manual(values = c(`TRUE` = 16, `FALSE` = 1)) +
				ggplot2::scale_x_continuous(breaks = plot_df$index, labels = plot_df$variable) +
				ggplot2::scale_y_continuous(limits = c(0, y_max)) +
				ggplot2::labs(title = "Local Procedure", y = "proportion included", x = NULL) +
				ggplot2::theme_minimal() +
				ggplot2::theme(
					legend.position = "none",
					axis.text.x = ggplot2::element_text(angle = 90, hjust = 1, vjust = 0.5),
					plot.margin = margin_cfg
				)

			simul_plot = ggplot2::ggplot(plot_df, ggplot2::aes(x = index, y = prop)) +
				ggplot2::geom_segment(
					ggplot2::aes(xend = index, y = 0, yend = simul_threshold),
					color = "blue"
				) +
				ggplot2::geom_hline(yintercept = max_cut, color = "red") +
				ggplot2::geom_point(ggplot2::aes(shape = simul_marker)) +
				ggplot2::scale_shape_manual(values = c(global_max = 16, global_se = 8, none = 1)) +
				ggplot2::scale_x_continuous(breaks = plot_df$index, labels = plot_df$variable) +
				ggplot2::scale_y_continuous(limits = c(0, y_max)) +
				ggplot2::labs(title = "Simul. Max and SE Procedures", y = "proportion included", x = NULL) +
				ggplot2::theme_minimal() +
				ggplot2::theme(
					legend.position = "none",
					axis.text.x = ggplot2::element_text(angle = 90, hjust = 1, vjust = 0.5),
					plot.margin = margin_cfg
				)

			plot_list = list(local_plot, simul_plot)
			if (verbose){
				grid::grid.newpage()
				grid::pushViewport(grid::viewport(layout = grid::grid.layout(2, 1)))
				for (i in seq_len(length(plot_list))){
					suppressMessages(print(
						plot_list[[i]],
						vp = grid::viewport(layout.pos.row = i, layout.pos.col = 1)
					))
				}
				grid::popViewport()
			}
		}
	}
	
  #return an invisible list
	invisible(list(
		important_vars_local_names = important_vars_pointwise_names,
		important_vars_global_max_names = important_vars_simul_max_names,
		important_vars_global_se_names = important_vars_simul_se_names,
		important_vars_local_col_nums = as.numeric(important_vars_pointwise_col_nums),
		important_vars_global_max_col_nums = as.numeric(important_vars_simul_max_col_nums),
		important_vars_global_se_col_nums = as.numeric(important_vars_simul_se_col_nums),		
		var_true_props_avg = var_true_props_avg,
		permute_mat = permute_mat
	))
}

##private
get_averaged_true_var_props = function(bart_machine, num_reps_for_avg, num_trees_for_permute, verbose = TRUE){
	var_props = rep(0, bart_machine$p)
	for (i in 1 : num_reps_for_avg){
		bart_machine_dup = bart_machine_duplicate(bart_machine, num_trees = num_trees_for_permute)
		var_props = var_props + get_var_props_over_chain(bart_machine_dup)
		if (verbose){
			cat(".")
		}
	}
	#average over many runs
	var_props / num_reps_for_avg
}

##private
get_null_permute_var_importances = function(bart_machine, num_trees_for_permute, verbose = TRUE){
	#permute the responses to disconnect x and y
	y_permuted = sample(bart_machine$y, replace = FALSE)
	
	#build BART on this permuted training data
	bart_machine_with_permuted_y = build_bart_machine(bart_machine$X, y_permuted, 
			num_trees = as.numeric(num_trees_for_permute), 
			num_burn_in = bart_machine$num_burn_in, 
			num_iterations_after_burn_in = bart_machine$num_iterations_after_burn_in,
			run_in_sample = FALSE,
			use_missing_data = bart_machine$use_missing_data,
			use_missing_data_dummies_as_covars = bart_machine$use_missing_data_dummies_as_covars,
			num_rand_samps_in_library = bart_machine$num_rand_samps_in_library,
			replace_missing_data_with_x_j_bar = bart_machine$replace_missing_data_with_x_j_bar,
			impute_missingness_with_rf_impute = bart_machine$impute_missingness_with_rf_impute,
			impute_missingness_with_x_j_bar_for_lm = bart_machine$impute_missingness_with_x_j_bar_for_lm,					
			verbose = FALSE)
	#just return the variable proportions	
	var_props = get_var_props_over_chain(bart_machine_with_permuted_y)
	if (verbose){
		cat(".")
	}
	var_props
}

##private - used to compute constant for global se method. simple bisection algo.
bisectK = function(tol, coverage, permute_mat, x_left, x_right, countLimit, perm_mean, perm_se){
	count = 0
	guess = (x_left + x_right) / 2
	# Pre-center and scale the permutation matrix to speed up the loop
	# We want to check if all(permute_mat[s,] - perm_mean <= guess * perm_se)
	# which is equivalent to all((permute_mat[s,] - perm_mean) / perm_se <= guess)
	# which is equivalent to max((permute_mat[s,] - perm_mean) / perm_se) <= guess
	
	# Precompute row-wise maximums of the standardized differences
	row_max_std_diffs = apply(sweep(sweep(permute_mat, 2, perm_mean, "-"), 2, perm_se, "/"), 1, max)
	
	while ((x_right - x_left) / 2 >= tol & count < countLimit){
		empirical_coverage = mean(row_max_std_diffs <= guess)
		if (empirical_coverage == coverage){
			break
		} else if (empirical_coverage < coverage){
			x_left = guess
		} else {
			x_right = guess
		}
		guess = (x_left + x_right) / 2
		count = count + 1
	}
	guess
}

##var selection -- choose best method via CV
#' Perform Variable Selection Using Cross-validation Procedure
#'
#' @description
#' Performs variable selection by cross-validating over the three threshold-based procedures outlined in Bleich et al. (2013) and selecting the single procedure that returns the lowest cross-validation RMSE.
#'
#' @details
#' See Bleich et al. (2013) for a complete description of the procedures outlined above as well as the corresponding vignette for a brief summary with examples.
#' @param bart_machine An object of class ``bartMachine''.
#' @param k_folds Number of folds to be used in cross-validation.
#' @param folds_vec An integer vector of indices specifying which fold each observation belongs to.
#' @param num_reps_for_avg Number of replicates to over over to for the BART model's variable inclusion proportions.
#' @param num_permute_samples Number of permutations of the response to be made to generate the ``null'' permutation distribution.
#' @param num_trees_for_permute Number of trees to use in the variable selection procedure. As with \cr \code{\link{investigate_var_importance}}, a small number of trees should be used to force variables to compete for entry into the model. Note that this number is used to estimate both the ``true'' and ``null'' variable inclusion proportions.
#' @param alpha Cut-off level for the thresholds.
#' @param num_trees_pred_cv Number of trees to use for prediction on the hold-out portion of each fold. Once variables have been selected using the training portion of each fold, a new model is built using only those variables with \code{num_trees_pred_cv} trees in the sum-of-trees model. Forecasts for the holdout sample are made using this model. A larger number of trees is recommended to exploit the full forecasting power of BART.
#' @param verbose If TRUE, prints progress messages.
#'
#' @return
#' Returns a list with the following components:
#' 
#'   \item{best_method}{The name of the best variable selection procedure, as chosen via cross-validation.}
#'   \item{important_vars_cv}{The variables chosen by the \code{best_method} above.}
#'
#' @references
#' J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
#' Additive Regression Trees. ArXiv e-prints, 2013.
#' 
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#'
#' @seealso
#' \code{\link{var_selection_by_permute}}, \code{\link{investigate_var_importance}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function can have substantial run-time.
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 150
#' p = 100 ##95 useless predictors
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model (not actually used in variable selection)
#' bart_machine = bartMachine(X, y)
#' 
#' #variable selection via cross-validation
#' var_sel_cv = var_selection_by_permute_cv(bart_machine, k_folds = 3)
#' print(var_sel_cv$best_method)
#' print(var_sel_cv$important_vars_cv)
#' }
#' @export
var_selection_by_permute_cv = function(bart_machine, k_folds = 5, folds_vec = NULL, num_reps_for_avg = 5, num_permute_samples = 100, num_trees_for_permute = 20, alpha = 0.05, num_trees_pred_cv = 50, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_number(k_folds, lower = 2) # Inf ok
  assert_integerish(folds_vec, null.ok = TRUE)
  assert_int(num_reps_for_avg, lower = 1)
  assert_int(num_permute_samples, lower = 1)
  assert_int(num_trees_for_permute, lower = 1)
  assert_number(alpha, lower = 0, upper = 1)
  assert_int(num_trees_pred_cv, lower = 1)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (k_folds <= 1 || k_folds > bart_machine$n){
		stop("The number of folds must be at least 2 and less than or equal to n, use \"Inf\" for leave one out")
	}	
	
	if (k_folds == Inf){ #leave-one-out
		k_folds = bart_machine$n
	}	
	if (!is.null(folds_vec) & !inherits(folds_vec, "integer")){
		stop("folds_vec must be an a vector of integers specifying the indexes of each folds.")  
	}
	
	#set up k folds
	if (is.null(folds_vec)){ ##if folds were not pre-set:
		n = nrow(bart_machine$X)
		if (k_folds == Inf){ #leave-one-out
			k_folds = n
		}
		
		if (k_folds <= 1 || k_folds > n){
			stop("The number of folds must be at least 2 and less than or equal to n, use \"Inf\" for leave one out")
		}
		
		temp = rnorm(n)
		
		folds_vec = cut(temp, breaks = quantile(temp, seq(0, 1, length.out = k_folds + 1)), 
				include.lowest= T, labels = FALSE)
	} else {
		k_folds = length(unique(folds_vec)) ##otherwise we know the folds, so just get k
	}
	
	L2_err_mat = matrix(NA, nrow = k_folds, ncol = 3)
	colnames(L2_err_mat) = c("important_vars_local_names", "important_vars_global_max_names", "important_vars_global_se_names")
	
	for (k in 1 : k_folds){
		if (verbose){
			cat("cv #", k, "\n", sep = "")
		}
		#find out the indices of the holdout sample
		train_idx = which(folds_vec != k)
		test_idx = setdiff(1 : n, train_idx)
		
		#pull out training data
		training_X_k = bart_machine$model_matrix_training_data[train_idx, -ncol(bart_machine$model_matrix_training_data)] ##toss last col bc its response
		training_y_k = bart_machine$y[train_idx]		
		
		#make a temporary bart machine just so we can run the var selection for all three methods
		bart_machine_temp = bart_machine_duplicate(bart_machine, X = as.data.frame(training_X_k), y = training_y_k, run_in_sample = FALSE, verbose = FALSE)
    
        ##do variable selection
		bart_variables_select_obj_k = var_selection_by_permute(bart_machine_temp, 
				num_permute_samples = num_permute_samples, 
				num_trees_for_permute = num_trees_for_permute,
        		num_reps_for_avg = num_reps_for_avg,                                                                          
				alpha = alpha, 
				plot = FALSE,
				verbose = verbose)
		
		#pull out test data
		test_X_k = bart_machine$model_matrix_training_data[test_idx, -ncol(bart_machine$model_matrix_training_data)]
		test_y_k = bart_machine$y[test_idx]
		
		if (verbose){
			cat("method")
		}
		for (method in colnames(L2_err_mat)){
			if (verbose){
				cat(".")
			}
			#pull out the appropriate vars
			vars_selected_by_method = bart_variables_select_obj_k[[method]]
  
			if (length(vars_selected_by_method) == 0){
				#we just predict ybar
				ybar_est = mean(training_y_k)
				#and we take L2 error against ybar
				L2_err_mat[k, method] = sum((test_y_k - ybar_est)^2)
			} else {
				#now build the bart machine based on reduced model
			  training_X_k_red_by_vars_picked_by_method = data.frame(training_X_k[, vars_selected_by_method])
			  colnames(training_X_k_red_by_vars_picked_by_method) = vars_selected_by_method #bug fix for single column - maybe drop  = F?
        
        ##need to account for cov_prior_vec update
        
				bart_machine_temp = bart_machine_duplicate(bart_machine, X = training_X_k_red_by_vars_picked_by_method, y = training_y_k,
						num_trees = num_trees_pred_cv,
						run_in_sample = FALSE,
            			cov_prior_vec = rep(1, times = ncol(training_X_k_red_by_vars_picked_by_method)),   ##do not want old vec -- standard here                                   
						verbose = FALSE)
				#and calculate oos-L2 and cleanup
				test_X_k_red_by_vars_picked_by_method = data.frame(test_X_k[, vars_selected_by_method])
        		colnames(test_X_k_red_by_vars_picked_by_method) = vars_selected_by_method #bug fix for single column

        		predict_obj = bart_predict_for_test_data(
					bart_machine_temp,
					test_X_k_red_by_vars_picked_by_method,
					test_y_k,
					verbose = verbose
				)
				#now record it
				L2_err_mat[k, method] = predict_obj$L2_err
			}
		}
		if (verbose){
			cat("\n")
		}
	}
	
	#now extract the lowest oos-L2 to find the "best" method for variable selection
	L2_err_by_method = colSums(L2_err_mat)
	min_var_selection_method = colnames(L2_err_mat)[which(L2_err_by_method == min(L2_err_by_method))]
	min_var_selection_method = min_var_selection_method[1]

	#now (finally) do var selection on the entire data and then return the vars from the best method found via cross-validation
	if (verbose){
		cat("final", "\n")
	}
	bart_variables_select_obj = var_selection_by_permute(bart_machine, 
			num_permute_samples = num_permute_samples, 
			num_trees_for_permute = num_trees_for_permute, 
	    	num_reps_for_avg = num_reps_for_avg,                                                                        
			alpha = alpha, 
			plot = FALSE,
			verbose = verbose)
	
    #return vars from best method and method name
	list(best_method = min_var_selection_method, important_vars_cv = sort(bart_variables_select_obj[[min_var_selection_method]]))
}

source("../testthat/helper-bartmachine-models.R")


test_that("variable selection helpers return expected structures", {
  reg_model <- build_regression_model()
  var_sel <- var_selection_by_permute(
    reg_model,
    num_reps_for_avg = BART_TESTS$var_sel_num_reps_for_avg,
    num_permute_samples = BART_TESTS$var_sel_num_permute_samples,
    num_trees_for_permute = BART_TESTS$var_sel_num_trees_for_permute,
    plot = FALSE,
    verbose = BART_TESTS$verbose
  )
  expect_true(is.list(var_sel))
  expect_true("permute_mat" %in% names(var_sel))

  var_sel_cv <- var_selection_by_permute_cv(
    reg_model,
    k_folds = BART_TESTS$var_sel_k_folds,
    num_reps_for_avg = BART_TESTS$var_sel_num_reps_for_avg,
    num_permute_samples = BART_TESTS$var_sel_num_permute_samples,
    num_trees_for_permute = BART_TESTS$var_sel_num_trees_for_permute,
    num_trees_pred_cv = BART_TESTS$var_sel_num_trees_pred_cv,
    verbose = BART_TESTS$verbose
  )
  expect_true(is.list(var_sel_cv))
  expect_true("best_method" %in% names(var_sel_cv))
})

test_that("cross-validation helpers run with small grids", {
  set.seed(BART_TESTS$seed)
  X <- data.frame(
    x1 = runif(BART_TESTS$cv_data_n),
    x2 = runif(BART_TESTS$cv_data_n)
  )
  y <- X$x1 + rnorm(BART_TESTS$cv_data_n, sd = BART_TESTS$cv_data_sd)

  cv_model <- bartMachineCV(
    X,
    y,
    num_tree_cvs = BART_TESTS$cv_num_tree_cvs,
    k_cvs = BART_TESTS$cv_k_cvs,
    k_folds = BART_TESTS$cv_k_folds,
    num_burn_in = BART_TESTS$cv_num_burn_in,
    num_iterations_after_burn_in = BART_TESTS$cv_num_iter,
    verbose = BART_TESTS$verbose
  )
  expect_s3_class(cv_model, "bartMachine")

  cv_model2 <- build_bart_machine_cv(
    X,
    y,
    num_tree_cvs = BART_TESTS$cv_num_tree_cvs_single,
    k_cvs = BART_TESTS$cv_k_cvs,
    k_folds = BART_TESTS$cv_k_folds,
    num_burn_in = BART_TESTS$cv_num_burn_in,
    num_iterations_after_burn_in = BART_TESTS$cv_num_iter,
    verbose = BART_TESTS$verbose
  )
  expect_s3_class(cv_model2, "bartMachine")

  kfold <- k_fold_cv(
    X,
    y,
    k_folds = BART_TESTS$cv_k_folds,
    num_trees = BART_TESTS$cv_num_trees,
    num_burn_in = BART_TESTS$cv_num_burn_in,
    num_iterations_after_burn_in = BART_TESTS$cv_num_iter,
    verbose = BART_TESTS$verbose
  )
  expect_true(is.list(kfold))
  expect_true("rmse" %in% names(kfold))
})

test_that("node/raw data extraction works", {
  reg_model <- build_regression_model()
  raw_nodes <- extract_raw_node_data(reg_model, g = BART_TESTS$raw_node_g)
  expect_true(is.list(raw_nodes))
})

test_that("training/test prediction helper works", {
  reg_model <- build_regression_model()
  idx <- seq_len(BART_TESTS$test_idx_n)
  Xtest <- reg_model$X[idx, , drop = FALSE]
  ytest <- reg_model$y[idx]
  oos <- bart_predict_for_test_data(reg_model, Xtest, ytest, verbose = BART_TESTS$verbose)
  expect_true(is.list(oos))
  expect_true("rmse" %in% names(oos))
})

#' Dummify Design Matrix
#'
#' @description
#' Create a data frame with factors converted to dummies.
#'
#' @details
#' The column names of the dummy variables are given by the ``FactorName_LevelName'' and are augmented to the end of the design matrix. See the example below.
#' @param data Data frame to be dummified.
#'
#' @return
#' Returns a data frame with factors converted to dummy indicator variables.
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' BART handles dummification internally. This function is provided as a utility function.
#'
#' @examples
#' \dontrun{
#' #generate data
#' set.seed(11)
#' x1 = rnorm(20)
#' x2 = as.factor(ifelse(x1 > 0, "A", "B"))
#' x3 = runif(20)
#' X = data.frame(x1,x2,x3)
#' #dummify data
#' X_dummified = dummify_data(X)
#' print(X_dummified)
#' }
#' @export
dummify_data = function(data){
  assert_data_frame(data)

	as.data.frame(pre_process_training_data(data)$data)
}

##private function that handles all pre-processing (dummification, missing data, etc.)
pre_process_training_data = function(data, use_missing_data_dummies_as_covars = FALSE, imputations = NULL){

	#first convert characters to factors
	character_vars = names(which(sapply(data, class) == "character"))
	for (character_var in character_vars){
		data[, character_var] = as.factor(data[, character_var])
	}
	
	factors = names(which(sapply(data, is.factor)))
	
	factor_lengths = c()
	if (length(factors) > 0) {
		# Pre-allocate a list for dummified matrices to avoid repeated cbind
		dummied_list = lapply(factors, function(fac) {
			levs = levels(data[, fac])
			dummied = matrix(0, nrow = nrow(data), ncol = length(levs))
			for (i in seq_along(levs)) {
				dummied[, i] = as.numeric(data[, fac] == levs[i])
			}
			colnames(dummied) = paste(fac, levs, sep = "_")
			dummied
		})
		
		# Record lengths
		factor_lengths = vapply(dummied_list, ncol, integer(1))
		
		# Combine all dummied matrices and non-factor columns
		non_factors = setdiff(names(data), factors)
		data = do.call(cbind, c(list(data[, non_factors, drop = FALSE]), dummied_list))
	}

	if (use_missing_data_dummies_as_covars){		
		#now take care of missing data - add each column as a missingness dummy
		predictor_columns_with_missingness = as.numeric(which(colSums(is.na(data)) > 0))
		
		#only do something if there are predictors with missingness
		if (length(predictor_columns_with_missingness) > 0){
			M = is.na(data[, predictor_columns_with_missingness, drop = FALSE]) + 0
			colnames(M) = paste("M_", colnames(data)[predictor_columns_with_missingness], sep = "")
			
			#now we may want to add imputations before the missingness dummies
			if (!is.null(imputations)){
				data = cbind(data, imputations)
			}
			#append the missing dummy columns to data as if they're real attributes themselves
			data = cbind(data, M)			
		}
	} else if (!is.null(imputations)){ #now we may want to add imputations
		data = cbind(data, imputations) 
	}
	#make sure to cast it as a data matrix and return it along with the factor_lengths
	list(data = data.matrix(data), factor_lengths = factor_lengths)
}

is.missing = function(x){
	is.na(x) || is.nan(x)
}

pre_process_new_data = function(new_data, bart_machine){
	new_data = as.data.frame(new_data)
	if (bart_machine$replace_missing_data_with_x_j_bar){
		new_data = imputeMatrixByXbarjContinuousOrModalForBinary(new_data, bart_machine$X)
	}
	n = nrow(new_data)
	
	imputations = NULL #global namespace?
	if (bart_machine$impute_missingness_with_rf_impute){
		#we ahve to impute with missForest since we don't have y's for the test data we want to predict
		imputations = missForest(rbind(new_data, bart_machine$X), verbose = bart_machine$verbose)$ximp
		imputations = imputations[1 : nrow(new_data), ]
		colnames(imputations) = paste(colnames(imputations), "_imp", sep = "")
	}
	
	#ensure factor levels include training levels to match dummies without binding all rows
	training_factors = names(which(sapply(bart_machine$X, is.factor)))
	if (length(training_factors) > 0){
		for (predictor in training_factors){
			if (!predictor %in% names(new_data)){
				next
			}
			training_levels = levels(bart_machine$X[, predictor])
			new_levels = levels(factor(new_data[, predictor]))
			combined_levels = union(training_levels, new_levels)
			new_data[, predictor] = factor(new_data[, predictor], levels = combined_levels)
		}
	}
	
	new_data = pre_process_training_data(new_data, bart_machine$use_missing_data_dummies_as_covars, imputations)$data
		
	if (bart_machine$use_missing_data){
		training_data_features = bart_machine$training_data_features_with_missing_features
	} else {
		training_data_features = bart_machine$training_data_features
	}
	
	#The new data features has to be a superset of the training data features, so pare it down even more
	new_data_features_before = colnames(new_data)
	missing_features = setdiff(training_data_features, new_data_features_before)
	if (length(missing_features) > 0){
		missing_matrix = matrix(0, nrow = n, ncol = length(missing_features))
		colnames(missing_matrix) = missing_features
		new_data = cbind(new_data, missing_matrix)
	}
	
	new_data = new_data[1 : n, training_data_features, drop = FALSE]
	
	differences = setdiff(new_data_features_before, training_data_features)
	
	if (length(differences) > 0){
		warning("The following features were found in records for prediction which were not found in the original training data:\n    ", paste(differences, collapse = ", "), "\n  These features will be ignored during prediction.")
	}
	
	new_data_features = colnames(new_data)
	
	if (!all(new_data_features == training_data_features)){
		# Re-order and fill missing columns in one go
		final_new_data = matrix(0, nrow = n, ncol = length(training_data_features))
		colnames(final_new_data) = training_data_features
		
		common_features = intersect(training_data_features, new_data_features)
		final_new_data[, common_features] = new_data[, common_features]
		new_data = final_new_data
	}
	#coerce to a numeric matrix
	new_data = data.matrix(new_data)
	mode(new_data) = "numeric"
	new_data
}

##function to permute columns of X and check BART's performance
#' Importance Test for Covariate(s) of Interest
#'
#' @description
#' This function tests the null hypothesis \eqn{H_0}: These covariates of interest
#' do not affect the response under the assumptions of the BART
#' model.
#'
#' @details
#' To test the importance of a covariate or a set of covariates of interest on the response, this function generates
#' \code{num_permutations} BART models with the covariate(s) of interest permuted (differently each time).
#' On each run, a measure of fit is recorded. For regression, the metric is Pseudo-Rsq; for classification, it is
#' total misclassification error.\cr A
#' p-value can then be generated as follows. For regression, the p-value is the number of
#' permutation-sampled Pseudo-Rsq's greater than the observed Pseudo-Rsq divided by
#' \code{num_permutations + 1}. For classification, the p-value is the number of permutation-sampled
#' total misclassification errors less than the observed total misclassification error divided by \code{num_permutations + 1}.
#' @param bart_machine An object of class ``bart_machine''.
#' @param covariates A vector of names of covariates of interest to be tested for having an effect on the response. A value of NULL
#'   indicates an omnibus test for all covariates having an effect on the response. If the name of a covariate is a factor,
#'   the entire factor will be permuted. We do not recommend entering the names of factor covariate dummies.
#' @param num_permutation_samples The number of times to permute the covariates of interest and create a corresponding new BART model (see details).
#' @param plot If \code{TRUE}, this produces a histogram of the Pseudo-Rsq's / total misclassification error rates from
#'   the \code{num_permutations} BART models created with the \code{covariates} permuted. The plot also illustrates
#'   the observed Pseudo-Rsq's / total misclassification error rate from the original training data and indicates
#'   the test's p-value.
#' @param verbose If TRUE, prints progress and summary messages.
#'
#' @return
#' \item{permutation_samples_of_error}{A vector which records the error metric of the BART models with the covariates permuted (see details).}
#' \item{observed_error_estimate}{For regression, this is the Pseudo-Rsq on the original
#' training data set. For classification, this is the observed total misclassification error
#' on the original training data set.}
#' \item{pval}{The approximate p-value for this test (see details).
#' }
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' ##regression example
#' 
#' ##generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' ##now test if X[, 1] affects Y nonparametrically under the BART model assumptions
#' cov_importance_test(bart_machine, covariates = c(1))
#' ## note the plot and the printed p-value
#' 
#' }
#' @export
cov_importance_test = function(bart_machine, covariates = NULL, num_permutation_samples = 100, plot = TRUE, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  # covariates can be numeric (indices) or character (names) or NULL
  assert(
    check_numeric(covariates, null.ok = TRUE, min.len = 1, lower = 1),
    check_character(covariates, null.ok = TRUE, min.len = 1),
    combine = "or"
  )
  assert_int(num_permutation_samples, lower = 1)
  assert_flag(plot)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	#be able to handle regular expressions to find the covariates
	
	all_covariates = bart_machine$training_data_features_with_missing_features
	
	if (is.null(covariates)){
		title = "bartMachine omnibus test for covariate importance\n"
	} else if (length(covariates) <= 3){
		if (inherits(covariates[1], "numeric")){
			cov_names = paste(all_covariates[covariates], collapse = ", ")
		} else {
			cov_names = paste(covariates, collapse = ", ")
		}
		title = paste("bartMachine test for importance of covariate(s):", cov_names, "\n")
	} else {
		title = paste("bartMachine test for importance of", length(covariates), "covariates", "\n")
	}
	if (verbose){
		cat(title)
	}
	observed_error_estimate = ifelse(bart_machine$pred_type == "regression", bart_machine$PseudoRsq, bart_machine$misclassification_error)

	permutation_samples_of_error = array(NA, num_permutation_samples)

	if (bart_machine$pred_type == "regression") {
		# Resolve covariates to 0-based column indices in the processed feature matrix
		all_feats = bart_machine$training_data_features_with_missing_features
		if (is.null(covariates)) {
			java_cols = integer(0)  # empty = permute y (omnibus test)
			use_java = TRUE
		} else if (is.numeric(covariates)) {
			java_cols = as.integer(covariates) - 1L
			use_java = TRUE
		} else {
			idxs = match(covariates, all_feats)
			if (any(is.na(idxs))) {
				use_java = FALSE  # factor expansion or unknown names — fall back to R
			} else {
				java_cols = as.integer(idxs) - 1L
				use_java = TRUE
			}
		}
		if (use_java) {
			permutation_samples_of_error = .jcall(bart_machine$java_bart_machine, "[D",
					"runCovariateImportancePermutations",
					.jarray(java_cols), as.integer(num_permutation_samples))
		}
	} else {
		use_java = FALSE
	}

	if (!use_java) {
	for (nsim in 1 : num_permutation_samples){
		if (verbose){
			cat(".")
			if (nsim %% 50 == 0){
				cat("\n")
			}
		}
		#omnibus F-like test - just permute y (same as permuting ALL the columns of X and it's faster)
		if (is.null(covariates)){
			bart_machine_samp = bart_machine_duplicate(bart_machine, y = sample(bart_machine$y), run_in_sample = TRUE, verbose = FALSE) #we have to turn verbose off otherwise there would be too many outputs
		#partial F-like test - permute the columns that we're interested in seeing if they matter
		} else {
			X_samp = bart_machine$X #copy original design matrix

			covariates_left_to_permute = c()
			for (cov in covariates){
				if (cov %in% colnames(X_samp)){
					X_samp[, cov] = sample(X_samp[, cov])
				} else {
					covariates_left_to_permute = c(covariates_left_to_permute, cov)
				}
			}

			bart_machine_samp = bart_machine_duplicate(bart_machine, X = X_samp, covariates_to_permute = covariates_left_to_permute, run_in_sample = TRUE, verbose = FALSE) #we have to turn verbose off otherwise there would be too many outputs
		}
		#record permutation result
		permutation_samples_of_error[nsim] = ifelse(bart_machine$pred_type == "regression", bart_machine_samp$PseudoRsq, bart_machine_samp$misclassification_error)
	}
	if (verbose){
		cat("\n")
	}
	}
  
	##compute p-value
	pval = ifelse(bart_machine$pred_type == "regression", sum(observed_error_estimate < permutation_samples_of_error), sum(observed_error_estimate > permutation_samples_of_error)) / (num_permutation_samples + 1)
	
	plot_obj = NULL
	if (plot){
		x_min = min(c(permutation_samples_of_error, 0.99 * observed_error_estimate), na.rm = TRUE)
		x_max = max(c(permutation_samples_of_error, 1.01 * observed_error_estimate), na.rm = TRUE)
		bins = max(1, floor(num_permutation_samples / 10))
		plot_df = data.frame(error = permutation_samples_of_error)
		plot_obj = ggplot2::ggplot(plot_df, ggplot2::aes(x = error)) +
			ggplot2::geom_histogram(bins = bins, fill = "gray70", color = "white") +
			ggplot2::geom_vline(xintercept = observed_error_estimate, color = "blue", linewidth = 1) +
			ggplot2::coord_cartesian(xlim = c(x_min, x_max)) +
			ggplot2::labs(
				title = paste(title, "Null Samples of", ifelse(bart_machine$pred_type == "regression", "Pseudo-R^2's", "Misclassification Errors")),
				x = paste("permutation samples\n pval =", round(pval, 3)),
				y = "count"
			) +
			ggplot2::theme_minimal()
		if (verbose){
			print(plot_obj)
		}
	}
	if (verbose){
		cat("p_val = ", pval, "\n")
	}
	result = list(
		permutation_samples_of_error = permutation_samples_of_error,
		observed_error_estimate = observed_error_estimate,
		pval = pval
	)
	if (!is.null(plot_obj)){
		attr(result, "plot") = plot_obj
	}
	invisible(result)
}

#' Test of Linearity
#'
#' @description
#' Test to investigate \eqn{H_0:} the functional relationship between the response and the
#' regressors is linear. We fit a linear model and then test if the residuals are a function
#' of the regressors using the
#' @param lin_mod A linear model you can pass in if you do not want to use the default which is \code{lm(y ~ X)}. Default is \code{NULL} which should be used if you pass in \code{X} and \code{y}.
#' @param X Data frame of predictors. Factors are automatically converted to dummies internally. Default is \code{NULL} which should be used if you pass in \code{lin_mode}.
#' @param y Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.
#'   Default is \code{NULL} which should be used if you pass in \code{lin_mode}.
#' @param num_permutation_samples This function relies on \code{\link{cov_importance_test}} (see documentation there for details).
#' @param plot This function relies on \code{\link{cov_importance_test}} (see documentation there for details).
#' @param verbose If TRUE, prints progress and summary messages.
#' @param ... Additional parameters to be passed to \code{bartMachine}, the model constructed on the residuals of the linear model.
#'
#' @return
#' \item{permutation_samples_of_error}{	This function relies on \code{\link{cov_importance_test}} (see documentation there for details).
#' }
#' \item{observed_error_estimate}{	This function relies on \code{\link{cov_importance_test}} (see documentation there for details).
#' }
#' \item{pval}{The approximate p-value for this test. See the documentation at \code{\link{cov_importance_test}}.
#' }
#'
#' @seealso
#' \code{\link{cov_importance_test}}
#'
#' @author
#' Adam Kapelner
#'
#' @examples
#' \dontrun{
#' ##regression example
#' 
#' ##generate Friedman data i.e. a nonlinear response model
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##now test if there is a nonlinear relationship between X1, ..., X5 and y.
#' linearity_test(X = X, y = y)
#' ## note the plot and the printed p-value.. should be approx 0
#' 
#' #generate a linear response model
#' y = 1 * X[ ,1] + 3 * X[,2] + 5 * X[,3] + 7 * X[ ,4] + 9 * X[,5] + rnorm(n)
#' linearity_test(X = X, y = y)
#' ## note the plot and the printed p-value.. should be > 0.05
#' 
#' }
#' @export
linearity_test = function(lin_mod = NULL, X = NULL, y = NULL, num_permutation_samples = 100, plot = TRUE, verbose = TRUE, ...){
  assert_class(lin_mod, "lm", null.ok = TRUE)
  assert_data_frame(X, null.ok = TRUE)
  if (!is.null(y)) assert_atomic_vector(y)
  assert_int(num_permutation_samples, lower = 1)
  assert_flag(plot)
  assert_flag(verbose)
  if (is.null(lin_mod) && (is.null(X) || is.null(y))) {
      stop("If lin_mod is not provided, both X and y must be provided.")
  }

	if (is.null(lin_mod)){
		lin_mod = lm(y ~ as.matrix(X))
	}
	y_hat = predict(lin_mod, X)
	bart_mod = bartMachine(X, y - y_hat, verbose = verbose, ...)
	cov_importance_test(bart_mod, num_permutation_samples = num_permutation_samples, plot = plot, verbose = verbose)	
}

if (getRversion() >= "2.15.1") {
	utils::globalVariables(
		c(
			"acceptance",
			"actual",
			"avg",
			"core",
			"depth",
			"error",
			"index",
			"in_interval",
			"iter",
			"lower",
			"num_nodes",
			"num_trees",
			"pointwise_cutoff",
			"pointwise_selected",
			"prop",
			"residual",
			"rmse",
			"simul_marker",
			"simul_threshold",
			"sigsq",
			"theoretical",
			"tree",
			"upper",
			"value",
			"var_est",
			"variable",
			"x"
		)
	)
}

BART_MAX_MEM_MB_DEFAULT = 1100 #1.1GB is the most a 32bit machine can give without throwing an error or crashing
BART_NUM_CORES_DEFAULT = 1 #Stay conservative as a default

#' @rdname bartMachine
#' @export
##build a BART model
build_bart_machine = function(X = NULL, y = NULL, Xy = NULL, 
		num_trees = 50, #found many times to not get better after this value... so let it be the default, it's faster too 
		num_burn_in = 250, 
		num_iterations_after_burn_in = 1000, 
		alpha = 0.95,
		beta = 2,
		k = 2,
		q = 0.9,
		nu = 3.0,
		prob_rule_class = 0.5,
		mh_prob_steps = c(2.5, 2.5, 4) / 9, #only the first two matter
		debug_log = FALSE,
		run_in_sample = TRUE,
		s_sq_y = "mse", # "mse" or "var"
		sig_sq_est = NULL, #you can pass this in to speed things up if you have an idea about what you want to use a priori
		print_tree_illustrations = FALSE, #POWER USERS ONLY
		cov_prior_vec = NULL,
		interaction_constraints = NULL,
		use_missing_data = FALSE,
		covariates_to_permute = NULL, #PRIVATE
		num_rand_samps_in_library = 10000, #give the user the option to make a bigger library of random samples of normals and inv-gammas
		use_missing_data_dummies_as_covars = FALSE,
		replace_missing_data_with_x_j_bar = FALSE,
		impute_missingness_with_rf_impute = FALSE,
		impute_missingness_with_x_j_bar_for_lm = TRUE,
		mem_cache_for_speed = TRUE,
		flush_indices_to_save_RAM = TRUE,
		serialize = FALSE,
		seed = NULL,
		use_xoshiro = FALSE,
		verbose = TRUE){

  # Validate arguments
  assert_data_frame(X, null.ok = TRUE)
  if (!is.null(y)) assert_atomic_vector(y)
  assert_data_frame(Xy, null.ok = TRUE)
  assert_int(num_trees, lower = 1)
  assert_int(num_burn_in, lower = 0)
  assert_int(num_iterations_after_burn_in, lower = 1)
  assert_number(alpha, lower = 0, upper = 1)
  assert_number(beta, lower = 0)
  assert_number(k, lower = .Machine$double.eps)
  assert_number(q, lower = .Machine$double.eps, upper = 1 - .Machine$double.eps)
  assert_number(nu, lower = .Machine$double.eps)
  assert_number(prob_rule_class, lower = 0, upper = 1)
  assert_numeric(mh_prob_steps, len = 3, lower = .Machine$double.eps)
  assert_flag(debug_log)
  assert_flag(run_in_sample)
  assert_choice(s_sq_y, c("mse", "var"))
  assert_number(sig_sq_est, lower = .Machine$double.eps, null.ok = TRUE)
  assert_flag(print_tree_illustrations)
  assert_numeric(cov_prior_vec, lower = .Machine$double.eps, null.ok = TRUE)
  assert_list(interaction_constraints, null.ok = TRUE)
  assert_flag(use_missing_data)
  assert_int(num_rand_samps_in_library, lower = 1)
  assert_flag(use_missing_data_dummies_as_covars)
  assert_flag(replace_missing_data_with_x_j_bar)
  assert_flag(impute_missingness_with_rf_impute)
  assert_flag(impute_missingness_with_x_j_bar_for_lm)
  assert_flag(mem_cache_for_speed)
  assert_flag(flush_indices_to_save_RAM)
  assert_flag(serialize)
  assert_int(seed, null.ok = TRUE)
  assert_flag(use_xoshiro)
  assert_flag(verbose)

	if (verbose){
		cat("bartMachine initializing with", num_trees, "trees...\n")	
	}	
	t0 = Sys.time()
	
	if (use_missing_data_dummies_as_covars && replace_missing_data_with_x_j_bar){
		stop("You cannot impute by averages and use missing data as dummies simultaneously.")
	}
	
	if ((is.null(X) && is.null(Xy)) || is.null(y) && is.null(Xy)){
		stop("You need to give bartMachine a training set either by specifying X and y or by specifying a matrix Xy which contains the response named \"y.\"\n")
	} else if (!is.null(X) && !is.null(y) && !is.null(Xy)){
		stop("You cannot specify both X,y and Xy simultaneously.")		
	} else if (is.null(X) && is.null(y)){ #they specified Xy, so now just pull out X,y
		#first ensure it's a dataframe
		if (!inherits(Xy, "data.frame")){
			stop(paste("The training data Xy must be a data frame."), call. = FALSE)	
		}
		y = Xy[, ncol(Xy)]
		for (cov in 1 : (ncol(Xy) - 1)){
			if (colnames(Xy)[cov] == ""){
				colnames(Xy)[cov] = paste("V", cov, sep = "")
			}
		}
		X = as.data.frame(Xy[, 1 : (ncol(Xy) - 1)])
		colnames(X) = colnames(Xy)[1 : (ncol(Xy) - 1)]
	}
	
	#make sure it's a data frame
	if (!inherits(X, "data.frame")){
		stop(paste("The training data X must be a data frame."), call. = FALSE)	
	}
	if (verbose){
		cat("bartMachine vars checked...\n")
	}	
	#we are about to construct a bartMachine object. First, let R garbage collect
	#to clean up previous bartMachine objects that are no longer in use. This is important
	#because R's garbage collection system does not "see" the size of Java objects. Thus,
	#you are at risk of running out of memory without this invocation. 
	gc() #Delete at your own risk!	

	#now take care of classification or regression
	y_levels = levels(y)
	if (inherits(y, "numeric") || inherits(y, "integer")){ #if y is numeric, then it's a regression problem
		if (inherits(y, "integer")){
			if (verbose){
				cat("Warning: The response y is integer, bartMachine will run regression.\n")
			}
		}
		#java expects doubles, not ints, so we need to cast this now to avoid errors later
		if (inherits(y, "integer")){
			y = as.numeric(y)
		}		
		java_bart_machine = .jnew("bartMachine.bartMachineRegressionMultThread")
		y_remaining = y
		pred_type = "regression"
	} else if (inherits(y, "factor") & length(y_levels) == 2){ #if y is a factor and binary
		#convenience for users that use 0/1 variables to ensure positive category is first as a level and a label (i.e. the naive expectation)
		if (all(sort(levels(factor(y))) == c("0", "1"))){
			y = factor(y, levels = c(1, 0), labels = c(1, 0))
			y_levels = levels(y)
		}		
		java_bart_machine = .jnew("bartMachine.bartMachineClassificationMultThread")
		y_remaining = ifelse(y == y_levels[1], 1, 0)
		pred_type = "classification"
	} else { #otherwise throw an error
		stop("Your response must be either numeric, an integer or a factor with two levels.\n")
	}
	
	num_gibbs = num_burn_in + num_iterations_after_burn_in
	
	if (ncol(X) == 0){
		stop("Your data matrix must have at least one attribute.")
	}
	if (nrow(X) == 0){
		stop("Your data matrix must have at least one observation.")
	}
	if (length(y) != nrow(X)){
		stop("The number of responses must be equal to the number of observations in the training data.")
	}
	if (verbose){
		cat("bartMachine java init...\n")
	}
	
	#if no column names, make up names
	if (is.null(colnames(X))){
		colnames(X) = paste("V", seq(from = 1, to = ncol(X), by = 1), sep = "")
	}
	
	if (any(mh_prob_steps < 0)){
		stop("The grow, prune, change ratio parameter vector must all be greater than 0.")
	}
	
	#now we should regenerate the factors for the factor columns
	predictors_which_are_factors = names(which(sapply(X, is.factor)))
	for (predictor in predictors_which_are_factors){
		X[, predictor] = factor(X[, predictor])
	}
	if (verbose){
		cat("bartMachine factors created...\n")
	}
	
	if (sum(is.na(y_remaining)) > 0){
		stop("You cannot have any missing data in your response vector.")
	}
	
	rf_imputations_for_missing = NULL
	if (impute_missingness_with_rf_impute){
		if (nrow(na.omit(X)) == nrow(X)){ #for the cases where it doesn't impute
			warning("No missing entries in the training data to impute.")
			rf_imputations_for_missing = X
		} else {
			#just use cols that HAVE missing data
			predictor_colnums_with_missingness = names(which(colSums(is.na(X)) > 0))
			
			rf_imputations_for_missing = rfImpute(X, y)
			rf_imputations_for_missing = rf_imputations_for_missing[, 2 : ncol(rf_imputations_for_missing), drop = FALSE]
			rf_imputations_for_missing = rf_imputations_for_missing[, predictor_colnums_with_missingness, drop = FALSE]
		}
		colnames(rf_imputations_for_missing) = paste(colnames(rf_imputations_for_missing), "_imp", sep = "")
		if (verbose){
			cat("bartMachine after rf imputations...\n")
		}	
	}
	
	#if we're not using missing data, go on and get rid of it
	if (!use_missing_data && !replace_missing_data_with_x_j_bar){
		rows_before = nrow(X)
		X = na.omit(X)
		rows_after = nrow(X)
		if (rows_before - rows_after > 0){
			stop("You have ", rows_before - rows_after, " observations with missing data. \nYou must either omit your missing data using \"na.omit()\" or turn on the\n\"use_missing_data\" or \"replace_missing_data_with_x_j_bar\" feature in order to use bartMachine.\n")
		}
	} else if (replace_missing_data_with_x_j_bar){
		X = imputeMatrixByXbarjContinuousOrModalForBinary(X, X)
		if (verbose){
			cat("Imputed missing data using attribute averages.\n")
		}
	}
	if (verbose){
		cat("bartMachine before preprocess...\n")
	}
	
	pre_process_obj = pre_process_training_data(X, use_missing_data_dummies_as_covars, rf_imputations_for_missing)
	model_matrix_training_data = cbind(pre_process_obj$data, y_remaining)
	p = ncol(model_matrix_training_data) - 1 # we subtract one because we tacked on the response as the last column
	factor_lengths = pre_process_obj$factor_lengths
	if (verbose){
		cat("bartMachine after preprocess...", p, "total features...\n")
	}
	#now create a default cov_prior_vec that factors in the levels of the factors
	null_cov_prior_vec = is.null(cov_prior_vec)
	if (null_cov_prior_vec && length(factor_lengths) > 0){
		#begin with the uniform
		cov_prior_vec = rep(1, p)
		j_factor_begin = p - sum(factor_lengths) + 1
		for (l in 1 : length(factor_lengths)){
			factor_length = factor_lengths[l]
			cov_prior_vec[j_factor_begin : (j_factor_begin + factor_length - 1)] = 1 / factor_length
			j_factor_begin = j_factor_begin + factor_length
		}
	}	
	
	if (!is.null(interaction_constraints)){
		if (!mem_cache_for_speed){
			stop("In order to use interaction constraints, \"mem_cache_for_speed\" must be set to TRUE.")
		}
		if (!inherits(interaction_constraints, "list")){
			stop("specified parameter \"interaction_constraints\" must be a list")
		} else if (length(interaction_constraints) == 0){
			stop("interaction_constraints list cannot be empty")
		}
		
		col_names = colnames(X)
		for (a in 1 : length(interaction_constraints)){
			vars_a = interaction_constraints[[a]]
			vars_a_converted = integer(length(vars_a))
			#check if the constraint components are valid features
			for (b in 1 : length(vars_a)){
				var = vars_a[[b]]
				if (inherits(var, "factor")){
					var = as.character(var)
				}
				if (inherits(var, "character")){
					idx = match(var, col_names)
					if (is.na(idx)){
						stop(paste("Element", var, "in interaction_constraints vector number", a, "is a string but not one of the column names of X."))
					}
					vars_a_converted[b] = idx - 1
				} else if (inherits(var, "numeric") || inherits(var, "integer")){
					if (var %% 1 != 0 || var < 1 || var > p){
						stop(paste("Element", var, "in interaction_constraints vector number", a, "is numeric but not one of 1, ...,", p, "where", p, "is the number of columns in X."))
					}
					vars_a_converted[b] = as.integer(var) - 1
				} else {
					stop(paste("Element", var, "in interaction_constraints vector number", a, "has unsupported type."))
				}
			}
			interaction_constraints[[a]] = vars_a_converted
		}
	}
	#print("interaction_constraints:"); print(interaction_constraints)

	#this is a private parameter ONLY called by cov_importance_test
	if (!is.null(covariates_to_permute)){
		#first check if these covariates are even in the matrix to begin with
		for (cov in covariates_to_permute){
			if (!(cov %in% colnames(model_matrix_training_data)) && inherits(cov, "character")){
				stop("Covariate \"", cov, "\" not found in design matrix.")
			}
		}
		permuted_order = sample(1 : nrow(model_matrix_training_data), nrow(model_matrix_training_data))
		model_matrix_training_data[, covariates_to_permute] = model_matrix_training_data[permuted_order, covariates_to_permute]
	}
	
	#now set whether we want the program to log to a file
	if (debug_log & verbose){
		cat("warning: printing out the log file will slow down the runtime significantly.\n")
		.jcall(java_bart_machine, "V", "writeStdOutToLogFile")
	}
	#set whether we want there to be tree illustrations
	if (print_tree_illustrations){
		if (verbose){
			cat("warning: printing tree illustrations is excruciatingly slow.\n")
		}
		.jcall(java_bart_machine, "V", "printTreeIllustations")
	}
	
	#set the std deviation of y to use
	if (ncol(model_matrix_training_data) - 1 >= nrow(model_matrix_training_data)){
		if (verbose){
			cat("warning: cannot use MSE of linear model for s_sq_y if p > n. bartMachine will use sample var(y) instead.\n")
		}
		s_sq_y = "var"
		
	}
	
  	##estimate sigma^2 to be given to the BART model
	if (is.null(sig_sq_est)){
		if (pred_type == "regression"){		
			y_range = max(y) - min(y)
			y_trans = (y - min(y)) / y_range - 0.5
			if (s_sq_y == "mse"){
				X_for_lm = as.data.frame(model_matrix_training_data)[1 : (ncol(model_matrix_training_data) - 1)]
				if (impute_missingness_with_x_j_bar_for_lm){
					X_for_lm = imputeMatrixByXbarjContinuousOrModalForBinary(X_for_lm, X_for_lm)
				}
				else if (nrow(na.omit(X_for_lm)) == 0){
					stop("The data does not have enough full records to estimate a naive prediction error. Please rerun with \"impute_missingness_with_x_j_bar_for_lm\" set to true.")
				}
				mod = lm(y_trans ~ ., X_for_lm)
				mse = var(mod$residuals)
				sig_sq_est = as.numeric(mse)
				.jcall(java_bart_machine, "V", "setSampleVarY", sig_sq_est)
			} else if (s_sq_y == "var"){
				sig_sq_est = as.numeric(var(y_trans))
				.jcall(java_bart_machine, "V", "setSampleVarY", sig_sq_est)
			} else { #if it's not a valid flag, throw an error
				stop("s_sq_y must be \"mse\" or \"var\"", call. = FALSE)
			}
			sig_sq_est = sig_sq_est * y_range^2		
			if (verbose){
				cat("bartMachine sigsq estimated...\n") #only print for regression
			}
		}		
	} else {
		if (verbose){
			cat("bartMachine using previous sigsq estimated...\n")
		}		
	}

	
	#if the user hasn't set a number of cores, set it here
	if (!exists("BART_NUM_CORES", envir = bartMachine_globals)){
		assign("BART_NUM_CORES", BART_NUM_CORES_DEFAULT, bartMachine_globals)
	}
	#load the number of cores the user set
	num_cores = get("BART_NUM_CORES", bartMachine_globals)

	
	#build bart to spec with what the user wants
	.jcall(java_bart_machine, "V", "setNumCores", as.integer(num_cores)) #this must be set FIRST!!!
	.jcall(java_bart_machine, "V", "setNumTrees", as.integer(num_trees))
	.jcall(java_bart_machine, "V", "setNumGibbsBurnIn", as.integer(num_burn_in))
	.jcall(java_bart_machine, "V", "setNumGibbsTotalIterations", as.integer(num_gibbs))
	.jcall(java_bart_machine, "V", "setAlpha", alpha)
	.jcall(java_bart_machine, "V", "setBeta", beta)
	.jcall(java_bart_machine, "V", "setK", k)
	.jcall(java_bart_machine, "V", "setQ", q)
	.jcall(java_bart_machine, "V", "setNU", nu)
	mh_prob_steps = mh_prob_steps / sum(mh_prob_steps) #make sure it's a prob vec
	.jcall(java_bart_machine, "V", "setProbGrow", mh_prob_steps[1])
	.jcall(java_bart_machine, "V", "setProbPrune", mh_prob_steps[2])
	.jcall(java_bart_machine, "V", "setVerbose", verbose)
	.jcall(java_bart_machine, "V", "setMemCacheForSpeed", mem_cache_for_speed)
	.jcall(java_bart_machine, "V", "setFlushIndicesToSaveRAM", flush_indices_to_save_RAM)
	.jcall(java_bart_machine, "V", "setUseXoshiro", use_xoshiro)
	.jcall(java_bart_machine, "V", "setUseGpu", isTRUE(getOption("bartMachine.use_gpu", TRUE)))
	
#	cat("seed", seed, "\n")
	if (!is.null(seed)){
		#set the seed in Java
		.jcall(java_bart_machine, "V", "setSeed", as.integer(seed))
	}
	
	#now we need to set random samples
	.jcall(java_bart_machine, "V", "setNormSamples", rnorm(num_rand_samps_in_library))
	n_plus_hyper_nu = nrow(model_matrix_training_data) + nu	
	.jcall(java_bart_machine, "V", "setGammaSamples", rchisq(num_rand_samps_in_library, n_plus_hyper_nu))
	
	if (length(cov_prior_vec) != 0){
		#put in checks here for user to make sure the covariate prior vec is the correct length
		offset = length(cov_prior_vec) - (ncol(model_matrix_training_data) - 1) 
		if (offset < 0){
			warning(paste("covariate prior vector length =", length(cov_prior_vec), "has to be equal to p =", ncol(model_matrix_training_data) - 1, "(the vector was lengthened with 1's)"))
			cov_prior_vec = c(cov_prior_vec, rep(1, -offset))
		}
		if (length(cov_prior_vec) != ncol(model_matrix_training_data) - 1){
			warning(paste("covariate prior vector length =", length(cov_prior_vec), "has to be equal to p =", ncol(model_matrix_training_data) - 1, "(the vector was shortened)"))
			cov_prior_vec = cov_prior_vec[1 : (ncol(model_matrix_training_data) - 1)]		
		}		
		if (sum(cov_prior_vec > 0) != ncol(model_matrix_training_data) - 1){
			stop("covariate prior vector has to have all its elements be positive", call. = FALSE)
			return(TRUE)
		}
		.jcall(java_bart_machine, "V", "setCovSplitPrior", .jarray(as.numeric(cov_prior_vec)))
	}
	
	if (!is.null(interaction_constraints)){
		.jcall(java_bart_machine, "V", "intializeInteractionConstraints", length(interaction_constraints))
		for (interaction_constraint_vector in interaction_constraints){
			for (b in 1 : length(interaction_constraint_vector)){
				.jcall(java_bart_machine, "V", "addInteractionConstraint",
						as.integer(interaction_constraint_vector[b]),
						.jarray(as.integer(interaction_constraint_vector[-b])))
			}
		}
	}
	
	#now load the training data into BART
	for (i in 1 : nrow(model_matrix_training_data)){
		row_as_char = as.character(model_matrix_training_data[i, ])
		row_as_char = replace(row_as_char, is.na(row_as_char), "NA") #this seems to be necessary for some R-rJava-linux distro-Java combinations
		.jcall(java_bart_machine, "V", "addTrainingDataRow", row_as_char)
	}
	.jcall(java_bart_machine, "V", "finalizeTrainingData")
	if (verbose){
		cat("bartMachine training data finalized...\n")
	}
	
	#build the bart machine and let the user know what type of BART this is
	if (verbose){
		cat("Now building bartMachine for", pred_type)
		if (pred_type == "classification"){
			cat(" where \"", y_levels[1], "\" is considered the target level", sep = "")
		}
		cat("...")
		if (length(cov_prior_vec) != 0){
			cat("Covariate importance prior ON. ")
		}
		if (use_missing_data){
			cat("Missing data feature ON. ")
		}
		if (use_missing_data_dummies_as_covars){
			cat("Missingness used as covariates. ")
		}
		if (impute_missingness_with_rf_impute){
			cat("Missing values imputed via rfImpute. ")
		}
		cat("\n")
	}
	.jcall(java_bart_machine, "V", "Build")
	
	#now once it's done, let's extract things that are related to diagnosing the build of the BART model
	
	bart_machine = list(java_bart_machine = java_bart_machine,
			training_data_features = colnames(model_matrix_training_data)[1 : ifelse(use_missing_data && use_missing_data_dummies_as_covars, (p / 2), p)],
			training_data_features_with_missing_features = colnames(model_matrix_training_data)[1 : p], #always return this even if there's no missing features
			X = X,
			y = y,
			y_levels = y_levels,
			pred_type = pred_type,
			model_matrix_training_data = model_matrix_training_data,
			n = nrow(model_matrix_training_data),
			p = p,
			num_cores = num_cores,
			num_trees = num_trees,
			num_burn_in = num_burn_in,
			num_iterations_after_burn_in = num_iterations_after_burn_in, 
			num_gibbs = num_gibbs,
			alpha = alpha,
			beta = beta,
			k = k,
			q = q,
			nu = nu,
			prob_rule_class = prob_rule_class,
			mh_prob_steps = mh_prob_steps,
			s_sq_y = s_sq_y,
			run_in_sample = run_in_sample,
			sig_sq_est = sig_sq_est,
			time_to_build = Sys.time() - t0,
			cov_prior_vec = cov_prior_vec,
			interaction_constraints = interaction_constraints,
			use_missing_data = use_missing_data,
			use_missing_data_dummies_as_covars = use_missing_data_dummies_as_covars,
			replace_missing_data_with_x_j_bar = replace_missing_data_with_x_j_bar,
			impute_missingness_with_rf_impute = impute_missingness_with_rf_impute,
			impute_missingness_with_x_j_bar_for_lm = impute_missingness_with_x_j_bar_for_lm,			
			verbose = verbose,
			serialize = serialize,
			mem_cache_for_speed = mem_cache_for_speed,
			flush_indices_to_save_RAM = flush_indices_to_save_RAM,
			debug_log = debug_log,
			seed = seed,
			use_xoshiro = use_xoshiro,
			num_rand_samps_in_library = num_rand_samps_in_library
	)
	#if the user used a cov prior vec, pass it back
	if (!null_cov_prior_vec){
		bart_machine$cov_prior_vec = cov_prior_vec
	}
	
	#once its done gibbs sampling, see how the training data does if user wants
	if (run_in_sample){
		if (verbose){
			cat("evaluating in sample data...")
		}
		if (pred_type == "regression"){
			y_hat_posterior_samples = 
				.jcall(bart_machine$java_bart_machine, "[[D", "getGibbsSamplesForPrediction", .jarray(model_matrix_training_data, dispatch = TRUE), as.integer(num_cores), simplify = TRUE)
			
			#to get y_hat.. just take straight mean of posterior samples
			y_hat_train = rowMeans(y_hat_posterior_samples)
			#return a bunch more stuff
			bart_machine$y_hat_train = y_hat_train
			bart_machine$residuals = y_remaining - bart_machine$y_hat_train
			bart_machine$L1_err_train = sum(abs(bart_machine$residuals))
			bart_machine$L2_err_train = sum(bart_machine$residuals^2)
			bart_machine$PseudoRsq = 1 - bart_machine$L2_err_train / sum((y_remaining - mean(y_remaining))^2) #pseudo R^2 acc'd to our dicussion with Ed and Shane
			bart_machine$rmse_train = sqrt(bart_machine$L2_err_train / bart_machine$n)
		} else if (pred_type == "classification"){
			p_hat_posterior_samples = 
				.jcall(bart_machine$java_bart_machine, "[[D", "getGibbsSamplesForPrediction", .jarray(model_matrix_training_data, dispatch = TRUE), as.integer(num_cores), simplify = TRUE)
			
			#to get y_hat.. just take straight mean of posterior samples
			p_hat_train = rowMeans(p_hat_posterior_samples)
			y_hat_train = labels_to_y_levels(bart_machine, p_hat_train > prob_rule_class)
			#return a bunch more stuff
			bart_machine$p_hat_train = p_hat_train
			bart_machine$y_hat_train = y_hat_train
			
			#calculate confusion matrix
			confusion_matrix = as.data.frame(matrix(NA, nrow = 3, ncol = 3))
			rownames(confusion_matrix) = c(paste("actual", y_levels), "use errors")
			colnames(confusion_matrix) = c(paste("predicted", y_levels), "model errors")
			
			confusion_matrix[1 : 2, 1 : 2] = as.integer(table(y, y_hat_train)) 
			confusion_matrix[3, 1] = round(confusion_matrix[2, 1] / (confusion_matrix[1, 1] + confusion_matrix[2, 1]), 3)
			confusion_matrix[3, 2] = round(confusion_matrix[1, 2] / (confusion_matrix[1, 2] + confusion_matrix[2, 2]), 3)
			confusion_matrix[1, 3] = round(confusion_matrix[1, 2] / (confusion_matrix[1, 1] + confusion_matrix[1, 2]), 3)
			confusion_matrix[2, 3] = round(confusion_matrix[2, 1] / (confusion_matrix[2, 1] + confusion_matrix[2, 2]), 3)
			confusion_matrix[3, 3] = round((confusion_matrix[1, 2] + confusion_matrix[2, 1]) / sum(confusion_matrix[1 : 2, 1 : 2]), 3)
			
			bart_machine$confusion_matrix = confusion_matrix
#			bart_machine$num_classification_errors = confusion_matrix[1, 2] + confusion_matrix[2, 1]
			bart_machine$misclassification_error = confusion_matrix[3, 3]
		}
		if (verbose){
			cat("done\n")
		}
	}
	
	
	#Let's serialize the object if the user wishes
	if (serialize){
		if (verbose){
			cat("serializing in order to be saved for future R sessions...")
		}
		.jcache(bart_machine$java_bart_machine)
		if (verbose){
			cat("done\n")
		}
	}
	
	#use R's S3 object orientation
	class(bart_machine) = "bartMachine"
	bart_machine
}

##private function that creates a duplicate of an existing bartMachine object.
bart_machine_duplicate = function(bart_machine, X = NULL, y = NULL, cov_prior_vec = NULL, num_trees = NULL, run_in_sample = NULL, covariates_to_permute = NULL, verbose = NULL, ...){	
	if (is.null(X)){
		X = bart_machine$X
	}
	if (is.null(y)){
		y = bart_machine$y
	}
	if (is.null(cov_prior_vec)){
		cov_prior_vec = bart_machine$cov_prior_vec
	}
	if (is.null(num_trees)){
		num_trees = bart_machine$num_trees
	}	
	if (is.null(run_in_sample)){
		run_in_sample = FALSE
	}
	if (is.null(covariates_to_permute)){
		covariates_to_permute = bart_machine$covariates_to_permute
	}
	if (is.null(verbose)){
		verbose = FALSE
	}	
	build_bart_machine(X, y,
		num_trees = num_trees, #found many times to not get better after this value... so let it be the default, it's faster too 
		num_burn_in = bart_machine$num_burn_in, 
		num_iterations_after_burn_in = bart_machine$num_iterations_after_burn_in, 
		alpha = bart_machine$alpha,
		beta = bart_machine$beta,
		k = bart_machine$k,
		q = bart_machine$q,
		nu = bart_machine$nu,
		prob_rule_class = bart_machine$prob_rule_class,
		mh_prob_steps = bart_machine$mh_prob_steps, #only the first two matter
		run_in_sample = run_in_sample,
		s_sq_y =  bart_machine$s_sq_y, # "mse" or "var"
		cov_prior_vec = cov_prior_vec,
		use_missing_data = bart_machine$use_missing_data,
		covariates_to_permute = covariates_to_permute, #PRIVATE
		num_rand_samps_in_library = bart_machine$num_rand_samps_in_library, #give the user the option to make a bigger library of random samples of normals and inv-gammas
		use_missing_data_dummies_as_covars = bart_machine$use_missing_data_dummies_as_covars,
		replace_missing_data_with_x_j_bar = bart_machine$replace_missing_data_with_x_j_bar,
		impute_missingness_with_rf_impute = bart_machine$impute_missingness_with_rf_impute,
		impute_missingness_with_x_j_bar_for_lm = bart_machine$impute_missingness_with_x_j_bar_for_lm,
		mem_cache_for_speed = bart_machine$mem_cache_for_speed,
		serialize = FALSE, #we do not want to waste CPU time here since these are created internally by us
		use_xoshiro = bart_machine$use_xoshiro,
		verbose = verbose)
}

#build a BART-cv model
#' @rdname bartMachineCV
#' @export
build_bart_machine_cv = function(X = NULL, y = NULL, Xy = NULL, 
		num_tree_cvs = c(50, 200),
		k_cvs = c(2, 3, 5),
		nu_q_cvs = NULL,
		k_folds = 5, 
		folds_vec = NULL, 
		use_xoshiro = FALSE,
		verbose = TRUE,
		...){
	
  # Validate arguments
  assert_data_frame(X, null.ok = TRUE)
  if (!is.null(y)) assert_atomic_vector(y)
  assert_data_frame(Xy, null.ok = TRUE)
  assert_integerish(num_tree_cvs, lower = 1, min.len = 1)
  assert_numeric(k_cvs, lower = 0, min.len = 1)
  assert_list(nu_q_cvs, null.ok = TRUE)
  assert_count(k_folds, positive = TRUE)
  assert_integerish(folds_vec, null.ok = TRUE)
  assert_flag(use_xoshiro)
  assert_flag(verbose)

	if ((is.null(X) && is.null(Xy)) || is.null(y) && is.null(Xy)){
		stop("You need to give bartMachine a training set either by specifying X and y or by specifying a matrix Xy which contains the response named \"y.\"\n")
	} else if (!is.null(X) && !is.null(y) && !is.null(Xy)){
		stop("You cannot specify both X,y and Xy simultaneously.")	
	} else if (is.null(X) && is.null(y)){ #they specified Xy, so now just pull out X,y
		if (!inherits(Xy, "data.frame")){
			stop(paste("The training data Xy must be a data frame."), call. = FALSE)	
		}
		y = Xy$y
		Xy$y = NULL
		X = Xy
	}	
	if (!is.null(folds_vec) & !inherits(folds_vec, "integer")){
		stop("folds_vec must be an a vector of integers specifying the indexes of each folds.")  
	}
	
	y_levels = levels(y)
	if (inherits(y, "numeric") || inherits(y, "integer")){ #if y is numeric, then it's a regression problem
		pred_type = "regression"
	} else if (inherits(y, "factor") & length(y_levels) == 2){ #if y is a factor and and binary, then it's a classification problem
		pred_type = "classification"
	} else { #otherwise throw an error
		stop("Your response must be either numeric, an integer or a factor with two levels.\n")
	}
	
	if (pred_type == "classification"){
		if (!is.null(nu_q_cvs)){
			stop("For classification, \"nu_q_cvs\" must be set to NULL (the default).")
		}
		nu_q_cvs = list(c(3, 0.9)) #ensure we only do this once, the 3 and the 0.9 don't actually matter, they just need to be valid numbers for the hyperparameters
	} else { #i.e. regression... 
		if (is.null(nu_q_cvs)){ #set it equal to a good default if user didn't specify
			nu_q_cvs = list(c(3, 0.9), c(3, 0.99), c(10, 0.75))
		}
	}
	
	min_rmse_num_tree = NULL
	min_rmse_k = NULL
	min_rmse_nu_q = NULL
	min_oos_rmse = Inf
	min_oos_misclassification_error = Inf
	
	cv_stats = matrix(NA, nrow = length(k_cvs) * length(nu_q_cvs) * length(num_tree_cvs), ncol = 6)
	colnames(cv_stats) = c("k", "nu", "q", "num_trees", "oos_error", "% diff with lowest")
	
	#set up k folds
	if (is.null(folds_vec)){ ##if folds were not pre-set:
		n = nrow(X)
	    if (k_folds == Inf){ #leave-one-out
			k_folds = n
	    }
	  
	    if (k_folds <= 1 || k_folds > n){
			stop("The number of folds must be at least 2 and less than or equal to n, use \"Inf\" for leave one out")
	    }
	  
	    temp = rnorm(n)
	  
	    folds_vec = cut(temp, breaks = quantile(temp, seq(0, 1, length.out = k_folds + 1)), 
			  include.lowest= T, labels = FALSE)
	  } else {
		  k_folds = length(unique(folds_vec)) ##otherwise we know the folds, so just get k
	  }
  
    #cross-validate
	run_counter = 1
	for (k in k_cvs){
		for (nu_q in nu_q_cvs){
			for (num_trees in num_tree_cvs){
				
				if (verbose){
					if (pred_type == "regression"){
						cat(paste("  bartMachine CV try: k:", k, "nu, q:", paste(as.numeric(nu_q), collapse = ", "), "m:", num_trees, "\n"))	
					} else {
						cat(paste("  bartMachine CV try: k:", k, "m:", num_trees, "\n"))
					}					
				}
				
				k_fold_results = k_fold_cv(X, y, 
          			k_folds = k_folds,
					folds_vec = folds_vec, ##will hold the cv folds constant 
					num_trees = num_trees,
					k = k,
					nu = nu_q[1],
					q = nu_q[2], 
					verbose = verbose,
					...)
				
				if (pred_type == "regression" && k_fold_results$rmse < min_oos_rmse){
					min_oos_rmse = k_fold_results$rmse					
					min_rmse_k = k
					min_rmse_nu_q = nu_q
					min_rmse_num_tree = num_trees
				} else if (pred_type == "classification" && k_fold_results$misclassification_error < min_oos_misclassification_error){
					min_oos_misclassification_error = k_fold_results$misclassification_error					
					min_rmse_k = k
					min_rmse_nu_q = nu_q
					min_rmse_num_tree = num_trees					
				}
				
				cv_stats[run_counter, 1 : 5] = c(k, nu_q[1], nu_q[2], num_trees, 
					ifelse(pred_type == "regression", k_fold_results$rmse, k_fold_results$misclassification_error))
				run_counter = run_counter + 1
			}
		}
	}
	if (verbose){
		if (pred_type == "regression"){
			cat(paste("  bartMachine CV win: k:", min_rmse_k, "nu, q:", paste(as.numeric(min_rmse_nu_q), collapse = ", "), "m:", min_rmse_num_tree, "\n"))
		} else {
			cat(paste("  bartMachine CV win: k:", min_rmse_k, "m:", min_rmse_num_tree, "\n"))
		}
	}
	#now that we've found the best settings, return that bart machine. It would be faster to have kept this around, but doing it this way saves RAM for speed.
	bart_machine_cv = build_bart_machine(X, y,
			num_trees = min_rmse_num_tree,
			k = min_rmse_k,
			nu = min_rmse_nu_q[1],
			q = min_rmse_nu_q[2], 
			verbose = verbose,
			use_xoshiro = use_xoshiro,
			...)
	
	#give the user some cv_stats ordered by the best (ie lowest) oosrmse
	cv_stats = cv_stats[order(cv_stats[, "oos_error"]), ]
	cv_stats[, 6] = (cv_stats[, 5] - cv_stats[1, 5]) / cv_stats[1, 5] * 100
	bart_machine_cv$cv_stats = cv_stats
  bart_machine_cv$folds = folds_vec
	bart_machine_cv
}

##private function for filling in missing data with averages for cont. vars and modes for cat. vars
imputeMatrixByXbarjContinuousOrModalForBinary = function(X_with_missing, X_for_calculating_avgs){
	replacements = vector("list", ncol(X_with_missing))
	for (j in seq_len(ncol(X_with_missing))){
		#mode for factors, otherwise average
		if (inherits(X_with_missing[, j], "factor")){
			replacements[[j]] = names(which.max(table(X_for_calculating_avgs[, j])))
		} else {
			replacements[[j]] = mean(X_for_calculating_avgs[, j], na.rm = TRUE)
		}
	}
	for (j in seq_len(ncol(X_with_missing))){
		missing_idx = is.na(X_with_missing[, j])
		if (any(missing_idx)){
			X_with_missing[missing_idx, j] = replacements[[j]]
		}
	}
	#now we have to go through and drop columns that are all NaN's if need be
	bad_cols = vapply(X_with_missing, function(col){
		if (!is.numeric(col)){
			return(FALSE)
		}
		sum(is.nan(col)) == nrow(X_with_missing)
	}, logical(1))
	if (any(bad_cols)){
		X_with_missing = X_with_missing[, !bad_cols, drop = FALSE]
	}
	X_with_missing
}

#' Gets node predictions indices of the training data for new data.
#'
#' @description
#' This returns a binary tensor for all gibbs samples after burn-in for all trees and for all training observations.
#' @param bart_machine An object of class ``bartMachine''.
#' @param new_data Data that you wish to investigate the training sample weights. If \code{NULL}, the original training data is used.
#'
#' @return
#' Returns a binary tensor indicating whether the prediction node contained a training datum or not. For each observation in new data, the size of this tensor is number of gibbs sample after burn-in
#' times the number of trees times the number of training data observations. This the size of the full tensor is the number of observations in the new data times the three dimensional object just explained.
#'
#' @examples
#' \dontrun{
#' set.seed(11)
#' n = 50
#' X = data.frame(x1 = rnorm(n), x2 = runif(n))
#' y = X$x1 + rnorm(n)
#' bart_machine = bartMachine(X, y, flush_indices_to_save_RAM = FALSE)
#' idx = node_prediction_training_data_indices(bart_machine)
#' }
#' @export
node_prediction_training_data_indices = function(bart_machine, new_data = NULL){
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(new_data, null.ok = TRUE)

	if (bart_machine$flush_indices_to_save_RAM){
		stop("Node prediction training data indices cannot be computed if \"flush_indices_to_save_RAM\" was used to construct the BART model.")
	}
	
	call_java <- function(arg){
		res <- .jcall(
			bart_machine$java_bart_machine,
			"[[[[Z",
			"getNodePredictionTrainingIndicies",
			arg,
			simplify = TRUE
		)
		if (is.null(res)){
			stop(
				"Java failed to produce node prediction training data indices. ",
				"Ensure the model was built with flush_indices_to_save_RAM = FALSE ",
				"and that the Java object is still valid.",
				call. = FALSE
			)
		}
		res
	}

	if (is.null(new_data)){
		double_vec_null = .jcast(.jnull(), new.class = "[[D", check = FALSE, convert.array = FALSE)
		call_java(double_vec_null)
	} else {
		call_java(.jarray(new_data, dispatch = TRUE))
	}	
}

#' Gets Training Sample Projection / Weights
#'
#' @description
#' Returns the matrix H where yhat is approximately equal to H y where yhat is the predicted values for \code{new_data}. If \code{new_data} is unspecified, yhat will be the in-sample fits.
#' If BART was the same as OLS, H would be an orthogonal projection matrix. Here it is a projection matrix, but clearly non-orthogonal. Unfortunately, I cannot get
#' this function to work correctly because of three possible reasons (1) BART does not work by averaging tree predictions: it is a sum of trees model where each tree sees the residuals
#' via backfitting (2) the prediction in each node is a bayesian posterior draw which is close to ybar of the observations contained in the node if noise is gauged to be small and
#' (3) there are transformations of the original y variable. I believe I got close and I think I'm off by a constant multiple which is a function of the number of trees. I can
#' use regression to estimate the constant multiple and correct for it. Turn \code{regression_kludge} to \code{TRUE} for this. Note that the weights do not add up to one here.
#' The intuition is because due to the backfitting there is multiple counting. But I'm not entirely sure.
#' @param bart_machine An object of class ``bartMachine''.
#' @param new_data Data that you wish to investigate the training sample projection / weights. If \code{NULL}, the original training data is used.
#' @param regression_kludge See explanation in the description. Default is \code{FALSE}.
#'
#' @return
#' Returns a matrix of proportions with number of rows equal to the number of rows of \code{new_data} and number of columns equal to the number of rows of the original training data, n.
#'
#' @examples
#' \dontrun{
#' options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))
#' pacman::p_load(bartMachine, tidyverse)
#' 
#' seed = 1984
#' set.seed(seed)
#' n = 100
#' x = rnorm(n, 0, 1)
#' sigma = 0.1
#' y = x + rnorm(n, 0, sigma)
#' 
#' num_trees = 200
#' num_iterations_after_burn_in = 1000
#' bart_mod = bartMachine(data.frame(x = x), y,
#' 	flush_indices_to_save_RAM = FALSE,
#' 	num_trees = num_trees,
#' 	num_iterations_after_burn_in = num_iterations_after_burn_in,
#' 	seed = seed)
#' bart_mod
#' 
#' n_star = 100
#' x_star = rnorm(n_star)
#' y_star = as.numeric(x_star + rnorm(n_star, 0, sigma))
#' yhat_star_bart = predict(bart_mod, data.frame(x = x_star))
#' 
#' Hstar = get_projection_weights(bart_mod, data.frame(x = x_star))
#' rowSums(Hstar)
#' yhat_star_projection = as.numeric(Hstar %*% y)
#' 
#' ggplot(data.frame(
#' 	yhat_star = yhat_star_bart,
#' 	yhat_star_projection = yhat_star_projection,
#' 	y_star = y_star)) +
#'   geom_point(aes(x = yhat_star_bart, y = yhat_star_projection), col = "green") +
#'   geom_abline(slope = 1, intercept = 0)
#' 
#' Hstar = get_projection_weights(bart_mod, data.frame(x = x_star), regression_kludge = TRUE)
#' rowSums(Hstar)
#' yhat_star_projection = as.numeric(Hstar %*% y)
#' 
#' ggplot(data.frame(
#' 	yhat_star = yhat_star_bart,
#' 	yhat_star_projection = yhat_star_projection,
#' 	y_star = y_star)) +
#'   geom_point(aes(x = yhat_star_bart, y = yhat_star_projection), col = "green") +
#'   geom_abline(slope = 1, intercept = 0)
#' 
#' }
#' @export
get_projection_weights = function(bart_machine, new_data = NULL, regression_kludge = FALSE){
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(new_data, null.ok = TRUE)
  assert_flag(regression_kludge)

	if (bart_machine$flush_indices_to_save_RAM){
		stop("Node prediction training data indices cannot be computed if \"flush_indices_to_save_RAM\" was used to construct the BART model.")
	}
	if (regression_kludge){
		if (is.null(new_data)){
			new_data = bart_machine$X
		}
		yhats = predict(bart_machine, new_data)
	}
	if (is.null(new_data)){
		double_vec_null = .jcast(.jnull(), new.class = "[[D", check = FALSE, convert.array = FALSE)
		weights = .jcall(bart_machine$java_bart_machine, "[[D", "getProjectionWeights", double_vec_null, simplify = TRUE)
	} else {
		weights = .jcall(bart_machine$java_bart_machine, "[[D", "getProjectionWeights", .jarray(as.matrix(new_data), dispatch = TRUE), simplify = TRUE)
	}	
	if (regression_kludge){
		yhat_star_projection = as.numeric(weights %*% bart_machine$y)
		weights * as.numeric(coef(lm(yhats ~ yhat_star_projection))[2]) #scale it back
	} else {
		weights
	}
}

BAD_FLAG_INT = -2147483647
BAD_FLAG_DOUBLE = -1.7976931348623157e+308
#' Gets Raw Node data
#'
#' @description
#' Returns a list object that contains all the information for all trees in a given Gibbs sample. Daughter nodes are nested
#' in the list structure recursively.
#' @param bart_machine An object of class ``bartMachine''.
#' @param g The gibbs sample number. It must be a natural number between 1 and the number of iterations after burn in. Default is 1.
#'
#' @return
#' Returns a list object that contains all the information for all trees in a given Gibbs sample.
#'
#' @examples
#' \dontrun{
#' options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))
#' pacman::p_load(bartMachine)
#' 
#' seed = 1984
#' set.seed(seed)
#' n = 100
#' x = rnorm(n, 0, 1)
#' sigma = 0.1
#' y = x + rnorm(n, 0, sigma)
#' 
#' num_trees = 200
#' num_iterations_after_burn_in = 1000
#' bart_mod = bartMachine(data.frame(x = x), y,
#' 	flush_indices_to_save_RAM = FALSE,
#' 	num_trees = num_trees,
#' 	num_iterations_after_burn_in = num_iterations_after_burn_in,
#' 	seed = seed)
#' 
#' raw_node_data = extract_raw_node_data(bart_mod)
#' 
#' }
#' @export
extract_raw_node_data = function(bart_machine, g = 1){
  assert_class(bart_machine, "bartMachine")
  assert_int(g, lower = 1)
  
	if (g < 1 | g > bart_machine$num_iterations_after_burn_in){
		stop("g is the gibbs sample number i.e. it must be a natural number between 1 and the number of iterations after burn in")
	}
	raw_data_java = .jcall(bart_machine$java_bart_machine, "[LbartMachine/bartMachineTreeNode;", "extractRawNodeInformation", as.integer(g - 1), simplify = TRUE)
	raw_data = list()
	for (m in 1 : bart_machine$num_trees){
		raw_data[[m]] = extract_node_data(raw_data_java[[m]])	
	}
	raw_data
}

extract_node_data = function(node_java){
	node_data = list()
	node_data$java_obj = node_java
	node_data$left_java_obj = node_java$left
	node_data$right_java_obj = node_java$right
	node_data$depth = node_java$depth
	node_data$isLeaf = node_java$isLeaf

	node_data$sendMissingDataRight = node_java$sendMissingDataRight

	node_data$n_eta = node_java$n_eta
	node_data$string_id = node_java$stringID()
	node_data$is_stump = node_java$isStump()
	node_data$string_location = node_java$stringLocation()
	
	if (node_java$splitAttributeM == BAD_FLAG_INT){
		node_data$splitAttributeM = NA
	} else {
		node_data$splitAttributeM = node_java$splitAttributeM
	}
	
	if (node_java$splitValue == BAD_FLAG_DOUBLE){
		node_data$splitValue = NA
	} else {
		node_data$splitValue = node_java$splitValue
	}
	
	if (node_java$y_pred == BAD_FLAG_DOUBLE){
		node_data$y_pred = NA
	} else {
		node_data$y_pred = node_java$y_pred
	}
	
	if (node_java$y_avg == BAD_FLAG_DOUBLE){
		node_data$y_avg = NA
	} else {
		node_data$y_avg = node_java$y_avg
	}
	
	if (node_java$posterior_var == BAD_FLAG_DOUBLE){
		node_data$posterior_var = NA
	} else {
		node_data$posterior_var = node_java$posterior_var
	}
	
	if (node_java$posterior_mean == BAD_FLAG_DOUBLE){
		node_data$posterior_mean = NA
	} else {
		node_data$posterior_mean = node_java$posterior_mean
	}
	
	if (!is.jnull(node_java$parent)){
		node_data$parent_java_obj = node_java$parent
	} else {
		node_data$parent_java_obj = NA	
	}
	
	if (!is.jnull(node_java$left)){
		node_data$left = extract_node_data(node_java$left)
	} else {
		node_data$left = NA
	}
	if (!is.jnull(node_java$right)){
		node_data$right = extract_node_data(node_java$right)
	} else {
		node_data$right = NA
	}
	node_data
}
		
		

##give summary info about bart
#' Summarizes information about a \code{bartMachine} object.
#'
#' @description
#' Provides a quick summary of the BART model.
#'
#' @details
#' Gives the version number of the \code{bartMachine} package used to build this \code{additiveBartMachine} object and if the object
#' models either ``regression'' or ``classification.'' Gives the amount of training data and the dimension of feature space. Prints
#' the amount of time it took to build the model, how many processor cores were used to during its construction, as well as the
#' number of burn-in and posterior Gibbs samples were used.
#' 
#' If the model is for regression, it prints the estimate of \eqn{\sigma^2} before the model was constructed as well as after so
#' the user can inspect how much variance was explained.
#' 
#' If the model was built using the \code{run_in_sample = TRUE} parameter in \code{\link{build_bart_machine}} and is for regression, the summary L1,
#' L2, rmse, Pseudo-\eqn{R^2} are printed as well as the p-value for the tests of normality and zero-mean noise. If the model is for classification, a confusion matrix is printed.
#' @param object An object of class ``bartMachine''.
#' @param verbose If TRUE, prints summary output.
#' @param ... Parameters that are ignored.
#'
#' @return
#' None.
#'
#' @author
#' Adam Kapelner
#'
#' @examples
#' \dontrun{
#' #Regression example
#' 
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' ##print out details
#' summary(bart_machine)
#' 
#' ##Also, the default print works too
#' bart_machine
#' }
#' @export
summary.bartMachine = function(object, verbose = TRUE, ...){	
  assert_class(object, "bartMachine")
  assert_flag(verbose)

	if (verbose){
		cat(paste("bartMachine v", packageVersion("bartMachine"), ifelse(object$pred_type == "regression", " for regression", " for classification"), "\n\n", sep = ""))
		if (object$use_missing_data){
			cat("Missing data feature ON\n")
		}
		#first print out characteristics of the training data
		if (!is.null(object$interaction_constraints)){
			cat(paste0("number of specified covariate interactivity constraints = ", length(object$interaction_constraints), ".\ntraining data size: n = ", object$n, " and p = ", object$p, "\n"))
		} else {
			cat(paste("training data size: n =", object$n, "and p =", object$p, "\n"))
		}	
		
		##build time
		ttb = as.numeric(object$time_to_build, units = "secs")
		if (ttb > 60){
			ttb = as.numeric(object$time_to_build, units = "mins")
			cat(paste("built in", round(ttb, 2), "mins on", object$num_cores, ifelse(object$num_cores == 1, "core,", "cores,"), object$num_trees, "trees,", object$num_burn_in, "burn-in and", object$num_iterations_after_burn_in, "post. samples\n"))
		} else {
			cat(paste("built in", round(ttb, 1), "secs on", object$num_cores, ifelse(object$num_cores == 1, "core,", "cores,"), object$num_trees, "trees,", object$num_burn_in, "burn-in and", object$num_iterations_after_burn_in, "post. samples\n"))
		}
		
		if (object$pred_type == "regression"){
			sigsq_est = sigsq_est(object) ##call private function
			cat(paste("\nsigsq est for y beforehand:", round(object$sig_sq_est, 3), "\n"))
			cat(paste("avg sigsq estimate after burn-in:", round(sigsq_est, 5), "\n"))
			
			if (object$run_in_sample){
				cat("\nin-sample statistics:\n")
				cat(paste(" L1 =", round(object$L1_err_train, 2), "\n",
								"L2 =", round(object$L2_err_train, 2), "\n",
								"rmse =", round(object$rmse_train, 2), "\n"),
						"Pseudo-Rsq =", round(object$PseudoRsq, 4))
				
				es = object$residuals
				if (length(es) > 5000){
					normal_p_val = shapiro.test(sample(es, 5000))$p.value
				} else {
					normal_p_val = shapiro.test(es)$p.value
				}
				cat("\np-val for shapiro-wilk test of normality of residuals:", round(normal_p_val, 5), "\n")
				
				centered_p_val = t.test(es)$p.value
				cat("p-val for zero-mean noise:", round(centered_p_val, 5), "\n")	
			} else {
				cat("\nno in-sample information available (use option run_in_sample = TRUE next time)\n")
			}		
		} else if (object$pred_type == "classification"){
			if (object$run_in_sample){
				cat("\nconfusion matrix:\n\n")
				print(object$confusion_matrix)
			} else {
				cat("\nno in-sample information available (use option run_in_sample = TRUE next time)\n")
			}		
		}
		cat("\n")
	}
	invisible(object)
}

#alias for summary
#' Summarizes information about a \code{bartMachine} object.
#'
#' @description
#' This is an alias for the \code{\link{summary.bartMachine}} function. See description in that section.
#' @param x An object of class ``bartMachine''.
#' @param verbose If TRUE, prints summary output.
#' @param ... Parameters that are ignored.
#'
#' @return
#' None.
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' #Regression example
#' 
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' ##print out details
#' print(bart_machine)
#' 
#' ##Also, the default print works too
#' bart_machine
#' }
#' @export
print.bartMachine = function(x, verbose = TRUE, ...){
  assert_class(x, "bartMachine")
  assert_flag(verbose)

	summary(x, verbose = verbose, ...)
	invisible(x)
}

##check BART error assumptions via plot
#' Check BART Error Assumptions
#'
#' @description
#' Diagnostic tools to assess whether the errors of the BART model for regression are normally distributed and homoskedastic, as assumed by the model. This function generates a normal quantile plot of the residuals with a Shapiro-Wilks p-value as well as a residual plot.
#' @param bart_machine An object of class ``bartMachine''.
#' @param hetero_plot If ``yhats'', the residuals are plotted against the fitted values of the response. If ``ys'', the residuals are plotted against the actual values of the response.
#' @param verbose If TRUE, prints plots to the active device.
#'
#' @return
#' None.
#'
#' @seealso
#' \code{\link{plot_convergence_diagnostics}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 300
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #check error diagnostics
#' check_bart_error_assumptions(bart_machine)
#' }
#' @export
check_bart_error_assumptions = function(bart_machine, hetero_plot = "yhats", verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_choice(hetero_plot, c("yhats", "ys"))
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (!(hetero_plot %in% c("ys", "yhats"))){
		stop("You must specify the parameter \"hetero_plot\" as \"ys\" or \"yhats\"")
	}
	if (bart_machine$pred_type == "classification"){
		stop("There are no convergence diagnostics for classification.")
	}	
	es = bart_machine$residuals
	y_hat = bart_machine$y_hat
	
	#plot/test for normality
	if (length(es) > 5000){
		normal_p_val = shapiro.test(sample(es, 5000))$p.value
	} else {
		normal_p_val = shapiro.test(es)$p.value
	}

	n = length(es)
	qq_df = data.frame(
		theoretical = qnorm(ppoints(n)),
		sample = sort(es)
	)
	qq_plot = ggplot2::ggplot(qq_df, ggplot2::aes(x = theoretical, y = sample)) +
		ggplot2::geom_point(color = "blue") +
		ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
		ggplot2::labs(
			title = paste("Assessment of Normality\n", "p-val for shapiro-wilk test of normality of residuals:", round(normal_p_val, 3)),
			x = "Theoretical Quantiles",
			y = "Sample Quantiles"
		)
	
	if (hetero_plot == "yhats"){
		hetero_df = data.frame(x = y_hat, residual = es)
		x_label = "Fitted Values"
	} else {
		hetero_df = data.frame(x = bart_machine$y, residual = es)
		x_label = "Actual Values"
	}
	hetero_plot_obj = ggplot2::ggplot(hetero_df, ggplot2::aes(x = x, y = residual)) +
		ggplot2::geom_point(color = "blue") +
		ggplot2::geom_hline(yintercept = 0, color = "black") +
		ggplot2::labs(
			title = "Assessment of Heteroskedasticity\nFitted vs residuals",
			x = x_label,
			y = "Residuals"
		)
	
	if (verbose){
		print(qq_plot)
		print(hetero_plot_obj)
	}
	invisible(list(qq_plot = qq_plot, hetero_plot = hetero_plot_obj))
}

##private function for plotting tree depths
plot_tree_depths = function(bart_machine){
	
	tree_depths_after_burn_in = get_tree_depths(bart_machine)

	num_after_burn_in = nrow(tree_depths_after_burn_in)
	num_trees = ncol(tree_depths_after_burn_in)
	iter = seq_len(num_after_burn_in)
	tree_df = data.frame(
		iter = rep(iter, times = num_trees),
		tree = factor(rep(seq_len(num_trees), each = num_after_burn_in)),
		depth = as.vector(tree_depths_after_burn_in)
	)
	row_mins = matrixStats::rowMins(tree_depths_after_burn_in)
	row_maxs = matrixStats::rowMaxs(tree_depths_after_burn_in)
	summary_df = data.frame(
		iter = iter,
		mean = rowMeans(tree_depths_after_burn_in),
		min = row_mins,
		max = row_maxs
	)
	
	p = ggplot2::ggplot(tree_df, ggplot2::aes(x = iter, y = depth, group = tree)) +
		ggplot2::geom_line(color = "gray70", alpha = 0.4) +
		ggplot2::geom_line(
			data = summary_df,
			ggplot2::aes(x = iter, y = mean),
			color = "blue",
			linewidth = 1,
			inherit.aes = FALSE
		) +
		ggplot2::geom_line(
			data = summary_df,
			ggplot2::aes(x = iter, y = min),
			color = "black",
			inherit.aes = FALSE
		) +
		ggplot2::geom_line(
			data = summary_df,
			ggplot2::aes(x = iter, y = max),
			color = "black",
			inherit.aes = FALSE
		) +
		ggplot2::labs(
			title = "Tree Depth by MCMC Iteration After Burn-in",
			x = "MCMC Iteration",
			y = "Tree Depth for all cores"
		)
	
	if (bart_machine$num_cores > 1){
		core_lines = (1 : (bart_machine$num_cores - 1)) *
			bart_machine$num_iterations_after_burn_in / bart_machine$num_cores
		p = p + ggplot2::geom_vline(xintercept = core_lines, color = "gray")
	}
	
	p
}

#private function for getting tree depths to plot
get_tree_depths = function(bart_machine){
	tree_depths_list = vector("list", bart_machine$num_cores)
	for (c in 1 : bart_machine$num_cores){
		tree_depths_list[[c]] = 
			.jcall(bart_machine$java_bart_machine, "[[I", "getDepthsForTreesInGibbsSampAfterBurnIn", as.integer(c), simplify = TRUE)
	}
	do.call(rbind, tree_depths_list)
}

#private function for plotting number of nodes in the trees
plot_tree_num_nodes = function(bart_machine){
	
	tree_num_nodes_and_leaves_after_burn_in = get_tree_num_nodes_and_leaves(bart_machine)

	num_after_burn_in = nrow(tree_num_nodes_and_leaves_after_burn_in)
	num_trees = ncol(tree_num_nodes_and_leaves_after_burn_in)
	iter = seq_len(num_after_burn_in)
	tree_df = data.frame(
		iter = rep(iter, times = num_trees),
		tree = factor(rep(seq_len(num_trees), each = num_after_burn_in)),
		num_nodes = as.vector(tree_num_nodes_and_leaves_after_burn_in)
	)
	row_mins = matrixStats::rowMins(tree_num_nodes_and_leaves_after_burn_in)
	row_maxs = matrixStats::rowMaxs(tree_num_nodes_and_leaves_after_burn_in)
	summary_df = data.frame(
		iter = iter,
		mean = rowMeans(tree_num_nodes_and_leaves_after_burn_in),
		min = row_mins,
		max = row_maxs
	)
	
	p = ggplot2::ggplot(tree_df, ggplot2::aes(x = iter, y = num_nodes, group = tree)) +
		ggplot2::geom_line(color = "gray70", alpha = 0.4) +
		ggplot2::geom_line(
			data = summary_df,
			ggplot2::aes(x = iter, y = mean),
			color = "blue",
			linewidth = 1,
			inherit.aes = FALSE
		) +
		ggplot2::geom_line(
			data = summary_df,
			ggplot2::aes(x = iter, y = min),
			color = "black",
			inherit.aes = FALSE
		) +
		ggplot2::geom_line(
			data = summary_df,
			ggplot2::aes(x = iter, y = max),
			color = "black",
			inherit.aes = FALSE
		) +
		ggplot2::labs(
			title = "Tree Num Nodes And Leaves by\nMCMC Iteration After Burn-in",
			x = "MCMC Iteration",
			y = "Tree Num Nodes and Leaves for all cores"
		)
	
	if (bart_machine$num_cores > 1){
		core_lines = (1 : (bart_machine$num_cores - 1)) *
			bart_machine$num_iterations_after_burn_in / bart_machine$num_cores
		p = p + ggplot2::geom_vline(xintercept = core_lines, color = "gray")
	}
	
	p
}
##private function for getting the number of nodes in the trees
get_tree_num_nodes_and_leaves = function(bart_machine){
	tree_num_nodes_and_leaves_list = vector("list", bart_machine$num_cores)
	for (c in 1 : bart_machine$num_cores){
		tree_num_nodes_and_leaves_list[[c]] = 
			.jcall(bart_machine$java_bart_machine, "[[I", "getNumNodesAndLeavesForTreesInGibbsSampAfterBurnIn", as.integer(c), simplify = TRUE)
	}
	do.call(rbind, tree_num_nodes_and_leaves_list)
}

#private function for plotting the MH acceptance proportions by core
plot_mh_acceptance_reject = function(bart_machine){
	
	mh_acceptance_reject = get_mh_acceptance_reject(bart_machine)
	a_r_before_burn_in = mh_acceptance_reject[["a_r_before_burn_in"]]
	a_r_before_burn_in_avg_over_trees = rowSums(a_r_before_burn_in) / bart_machine$num_trees
	
	a_r_after_burn_in_avgs_over_trees = list()
	for (c in 1 : bart_machine$num_cores){
		a_r_after_burn_in = mh_acceptance_reject[["a_r_after_burn_in"]][[c]]
		a_r_after_burn_in_avgs_over_trees[[c]] = rowSums(a_r_after_burn_in) / bart_machine$num_trees		
	}	
	
	num_after_burn_in_per_core = length(a_r_after_burn_in_avgs_over_trees[[1]])
	num_gibbs_per_core = bart_machine$num_burn_in + num_after_burn_in_per_core

	burn_in_df = data.frame(
		iter = 1 : bart_machine$num_burn_in,
		acceptance = a_r_before_burn_in_avg_over_trees
	)
	after_iter = (bart_machine$num_burn_in + 1) : num_gibbs_per_core
	after_df = do.call(
		rbind,
		lapply(1 : bart_machine$num_cores, function(c){
			data.frame(
				iter = after_iter,
				acceptance = a_r_after_burn_in_avgs_over_trees[[c]],
				core = factor(c)
			)
		})
	)
	
	p = ggplot2::ggplot() +
		ggplot2::geom_point(
			data = burn_in_df,
			ggplot2::aes(x = iter, y = acceptance),
			color = "grey"
		) +
		ggplot2::geom_line(
			data = burn_in_df,
			ggplot2::aes(x = iter, y = acceptance),
			color = "black",
			linewidth = 0.8
		) +
		ggplot2::geom_point(
			data = after_df,
			ggplot2::aes(x = iter, y = acceptance, color = core),
			alpha = 0.7
		) +
		ggplot2::geom_line(
			data = after_df,
			ggplot2::aes(x = iter, y = acceptance, color = core),
			linewidth = 0.8,
			alpha = 0.7
		) +
		ggplot2::geom_vline(xintercept = bart_machine$num_burn_in, color = "grey") +
		ggplot2::scale_color_manual(values = COLORS[1 : bart_machine$num_cores]) +
		ggplot2::coord_cartesian(ylim = c(0, 1)) +
		ggplot2::labs(
			title = "Percent Acceptance by MCMC Iteration",
			x = "MCMC Iteration",
			y = "% of Trees Accepting",
			color = "Core"
		)

	smooth_layer <- plot_mh_acceptance_smooth(after_df)
	if (!is.null(smooth_layer)) {
		p <- p + smooth_layer
	}

	p
}

## helper to only smooth cores with enough data
plot_mh_acceptance_smooth <- function(after_df) {
  counts <- table(after_df$core)
  valid_cores <- names(counts[counts >= 4])
  if (length(valid_cores) == 0) {
    return(NULL)
  }
  smooth_df <- after_df[after_df$core %in% valid_cores, , drop = FALSE]
  smooth_layer <- tryCatch(
    suppressMessages(
      ggplot2::geom_smooth(
        data = smooth_df,
        ggplot2::aes(x = iter, y = acceptance, color = core),
        method = "loess",
        formula = y ~ x,
        se = FALSE,
        linewidth = 1
      )
    ),
    error = function(e) NULL
  )
  smooth_layer
}

##private function for getting the MH acceptance proportions by core
get_mh_acceptance_reject = function(bart_machine){
	a_r_before_burn_in = 
		.jcall(bart_machine$java_bart_machine, "[[Z", "getAcceptRejectMHsBurnin", simplify = TRUE) * 1
	
	a_r_after_burn_in = list()
	for (c in 1 : bart_machine$num_cores){
		a_r_after_burn_in[[c]] = 
			.jcall(bart_machine$java_bart_machine, "[[Z", "getAcceptRejectMHsAfterBurnIn", as.integer(c), simplify = TRUE) * 1
	}
	
	list(
			a_r_before_burn_in = a_r_before_burn_in,
			a_r_after_burn_in = a_r_after_burn_in	
	)
}

#plot y vs yhat for training or test data
#' Plot the fitted Versus Actual Response
#'
#' @description
#' Generates a plot actual versus fitted values and corresponding credible intervals or prediction intervals for the fitted values.
#' @param bart_machine An object of class ``bartMachine''.
#' @param Xtest Optional argument for test data. If included, BART computes fitted values at the rows of \code{Xtest}. Else, the fitted values from the training data are used.
#' @param ytest Optional argument for test data. Vector of observed values corresponding to the rows of \code{Xtest} to be plotted against the predictions for the rows of \code{Xtest}.
#' @param credible_intervals If TRUE, Bayesian credible intervals are computed using the quantiles of the posterior distribution of \eqn{\hat{f}(x)}. See \code{\link{calc_credible_intervals}} for details.
#' @param prediction_intervals If TRUE, Bayesian predictive intervals are computed using the a draw of from \eqn{\hat{f}(x)}. See \code{\link{calc_prediction_intervals}} for details.
#' @param interval_confidence_level Desired level of confidence for credible or prediction intervals.
#' @param verbose If TRUE, prints plots to the active device.
#'
#' @return
#' None.
#'
#' @seealso
#' \code{\link{bart_machine_get_posterior}}, \code{\link{calc_credible_intervals}}, \code{\link{calc_prediction_intervals}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate linear data
#' set.seed(11)
#' n  = 500
#' p = 3
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 3*X[ ,1] + 2*X[ ,2] +X[ ,3] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' ##generate plot
#' plot_y_vs_yhat(bart_machine)
#' 
#' #generate plot with prediction bands
#' plot_y_vs_yhat(bart_machine, prediction_intervals = TRUE)
#' }
#' @export
plot_y_vs_yhat = function(bart_machine, Xtest = NULL, ytest = NULL, credible_intervals = FALSE, prediction_intervals = FALSE, interval_confidence_level = 0.95, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_data_frame(Xtest, null.ok = TRUE)
  if (!is.null(ytest)) assert_atomic_vector(ytest)
  assert_flag(credible_intervals)
  assert_flag(prediction_intervals)
  assert_number(interval_confidence_level, lower = 0, upper = 1)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if( (!bart_machine$run_in_sample) & (is.null(Xtest) | is.null(ytest)) ){
		stop("To run on training data, you must set \"run_in_sample\" option to TRUE in \"build_bart_machine\"")
	}
  
	if (credible_intervals && prediction_intervals){
		stop("Cannot plot both credibility intervals and prediction intervals simultaneously.")
	}
	if (bart_machine$pred_type == "classification"){
		stop("Cannot plot y vs y_hat for classification.")
	}
  	if( (is.null(Xtest) & !is.null(ytest)) | (!is.null(Xtest) & is.null(ytest)) ){
    	stop("Must pass both X and y to use on test data")
  	}
	
	if (is.null(Xtest) & is.null(ytest)){
		Xtest = bart_machine$X
		ytest = bart_machine$y
		y_hat = bart_machine$y_hat_train
		in_sample = TRUE
	} else {
		predict_obj = bart_predict_for_test_data(bart_machine, Xtest, ytest, verbose = verbose)
		y_hat = predict_obj$y_hat
		in_sample = FALSE
	}
	
	plot_obj = NULL
	if (credible_intervals){
		credible_intervals = calc_credible_intervals(bart_machine, Xtest, interval_confidence_level)
		ci_a = credible_intervals[, 1]
		ci_b = credible_intervals[, 2]
		y_in_ppi = ytest >= ci_a & ytest <= ci_b
		prop_ys_in_ppi = sum(y_in_ppi) / length(y_in_ppi)
		
		plot_title = paste(
			ifelse(in_sample, "In-Sample", "Out-of-Sample"),
			" Fitted vs. Actual Values\nwith ",
			round(interval_confidence_level * 100),
			"% Cred. Int.'s (",
			round(prop_ys_in_ppi * 100, 2),
			"% coverage)",
			sep = ""
		)
		min_val = min(min(ytest), min(y_hat))
		max_val = max(max(ytest), max(y_hat))
		plot_df = data.frame(
			actual = ytest,
			fitted = y_hat,
			lower = ci_a,
			upper = ci_b,
			in_interval = factor(y_in_ppi, levels = c(TRUE, FALSE))
		)
		p = ggplot2::ggplot(plot_df, ggplot2::aes(x = actual, y = fitted)) +
			ggplot2::geom_segment(
				ggplot2::aes(xend = actual, y = lower, yend = upper),
				color = "gray54",
				linewidth = 0.3
			) +
			ggplot2::geom_point(
				ggplot2::aes(color = in_interval, shape = in_interval),
				size = 1.6
			) +
			ggplot2::scale_color_manual(values = c("TRUE" = "darkblue", "FALSE" = "red")) +
			ggplot2::scale_shape_manual(values = c("TRUE" = 16, "FALSE" = 4)) +
			ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
			ggplot2::coord_cartesian(xlim = c(min_val, max_val), ylim = c(min_val, max_val)) +
			ggplot2::labs(
				title = plot_title,
				x = "Actual Values",
				y = "Fitted Values",
				color = "In Interval",
				shape = "In Interval"
			)
		if (verbose){
			print(p)
		}
		plot_obj = p
	} else if (prediction_intervals){
		prediction_intervals = calc_prediction_intervals(bart_machine, Xtest, interval_confidence_level)
		ci_a = prediction_intervals[, 1]
		ci_b = prediction_intervals[, 2]
		y_in_ppi = ytest >= ci_a & ytest <= ci_b
		prop_ys_in_ppi = sum(y_in_ppi) / length(y_in_ppi)
		
		plot_title = paste(
			ifelse(in_sample, "In-Sample", "Out-of-Sample"),
			" Fitted vs. Actual Values\nwith ",
			round(interval_confidence_level * 100),
			"% Pred. Int.'s (",
			round(prop_ys_in_ppi * 100, 2),
			"% coverage)",
			sep = ""
		)
		min_val = min(min(ytest), min(y_hat))
		max_val = max(max(ytest), max(y_hat))
		plot_df = data.frame(
			actual = ytest,
			fitted = y_hat,
			lower = ci_a,
			upper = ci_b,
			in_interval = factor(y_in_ppi, levels = c(TRUE, FALSE))
		)
		p = ggplot2::ggplot(plot_df, ggplot2::aes(x = actual, y = fitted)) +
			ggplot2::geom_segment(
				ggplot2::aes(xend = actual, y = lower, yend = upper),
				color = "gray54",
				linewidth = 0.3
			) +
			ggplot2::geom_point(
				ggplot2::aes(color = in_interval, shape = in_interval),
				size = 1.6
			) +
			ggplot2::scale_color_manual(values = c("TRUE" = "darkblue", "FALSE" = "red")) +
			ggplot2::scale_shape_manual(values = c("TRUE" = 16, "FALSE" = 4)) +
			ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
			ggplot2::coord_cartesian(xlim = c(min_val, max_val), ylim = c(min_val, max_val)) +
			ggplot2::labs(
				title = plot_title,
				x = "Actual Values",
				y = "Fitted Values",
				color = "In Interval",
				shape = "In Interval"
			)
		if (verbose){
			print(p)
		}
		plot_obj = p
	} else {
		min_val = min(min(ytest), min(y_hat))
		max_val = max(max(ytest), max(y_hat))
		plot_df = data.frame(actual = ytest, fitted = y_hat)
		p = ggplot2::ggplot(plot_df, ggplot2::aes(x = actual, y = fitted)) +
			ggplot2::geom_point(color = "blue") +
			ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
			ggplot2::coord_cartesian(xlim = c(min_val, max_val), ylim = c(min_val, max_val)) +
			ggplot2::labs(
				title = "Fitted vs. Actual Values",
				x = "Actual Values",
				y = "Fitted Values"
			)
		if (verbose){
			print(p)
		}
		plot_obj = p
	}
	invisible(plot_obj)
}

##get sigsqs and plot a histogram, if desired
#' Get Posterior Error Variance Estimates
#'
#' @description
#' Returns the posterior estimates of the error variance from the Gibbs samples with an option to create a histogram of the posterior estimates of the error variance  with a credible interval overlaid.
#' @param bart_machine An object of class ``bartMachine''.
#' @param after_burn_in If TRUE, only the \eqn{\sigma^2} draws after the burn-in period are returned.
#' @param plot_hist If TRUE, a histogram of the posterior \eqn{\sigma^2} draws is generated.
#' @param plot_CI Confidence level for credible interval on histogram.
#' @param plot_sigma If TRUE, plots \eqn{\sigma} instead of \eqn{\sigma^2}.
#' @param verbose If TRUE, prints plots to the active device.
#'
#' @return
#' Returns a vector of posterior \eqn{\sigma^2} draws (with or without the burn-in samples).
#'
#' @seealso
#' \code{\link{get_sigsqs}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 300
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #get posterior sigma^2's after burn-in and plot
#' sigsqs = get_sigsqs(bart_machine, plot_hist = TRUE)
#' }
#' @export
get_sigsqs = function(bart_machine, after_burn_in = TRUE, plot_hist = FALSE, plot_CI = .95, plot_sigma = F, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_flag(after_burn_in)
  assert_flag(plot_hist)
  assert_number(plot_CI, lower = 0, upper = 1)
  assert_flag(plot_sigma)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (bart_machine$pred_type == "classification"){
		stop("There are no sigsq's for classification.")
	}
	
	sigsqs = .jcall(bart_machine$java_bart_machine, "[D", "getGibbsSamplesSigsqs")
	
	num_iterations_after_burn_in = bart_machine$num_iterations_after_burn_in
	num_burn_in = bart_machine$num_burn_in
	num_gibbs = bart_machine$num_gibbs
	num_trees = bart_machine$num_trees
	
	sigsqs_after_burnin = tail(sigsqs, num_iterations_after_burn_in)
	avg_sigsqs = mean(sigsqs_after_burnin, na.rm = TRUE)
	
	if(plot_hist){
    if(plot_sigma){
      var_est_to_plot = sqrt(sigsqs_after_burnin)
    }
    else{
      var_est_to_plot = sigsqs_after_burnin
    }

	  ppi_a = quantile(var_est_to_plot, (.5 - plot_CI/2))
	  ppi_b = quantile(var_est_to_plot, (.5 + plot_CI/2))
	  plot_df = data.frame(var_est = var_est_to_plot)
	  plot_title = paste("Histogram of ", ifelse(plot_sigma ==T, "Sigmas", "Sigma^2s"),  " After Burn-in", sep = "")
	  plot_xlab = paste(
	  	"Avg = ",
	  	round(mean(var_est_to_plot, na.rm = TRUE), 2),
	  	", ",
	  	100 * plot_CI,
	  	"% Credible Interval = [",
	  	round(ppi_a, 2),
	  	", ",
	  	round(ppi_b, 2),
	  	"]",
	  	sep = ""
	  )
	  p = ggplot2::ggplot(plot_df, ggplot2::aes(x = var_est)) +
	  	ggplot2::geom_histogram(bins = 100, fill = "gray70", color = "white") +
	  	ggplot2::geom_vline(xintercept = mean(var_est_to_plot, na.rm = TRUE), color = "blue") +
	  	ggplot2::geom_vline(xintercept = c(ppi_a, ppi_b), color = "red") +
	  	ggplot2::labs(title = plot_title, x = plot_xlab, y = "Count")
	  if (verbose){
	  	print(p)
	  }
	}
  
	if(after_burn_in == T){
    sigsqs_out = sigsqs_after_burnin
	}else{
    sigsqs_out = sigsqs
	}
	if (plot_hist){
		attr(sigsqs_out, "plot") = p
	}
	return(sigsqs_out)
}

#private function for plotting convergence diagnostics for sigma^2
plot_sigsqs_convergence_diagnostics = function(bart_machine){	
	if (bart_machine$pred_type == "classification"){
		stop("There are no convergence diagnostics for classification.")
	}	
	
	sigsqs = get_sigsqs(bart_machine, after_burn_in = FALSE)
	
	num_iterations_after_burn_in = bart_machine$num_iterations_after_burn_in
	num_burn_in = bart_machine$num_burn_in
	num_gibbs = bart_machine$num_gibbs
	num_trees = bart_machine$num_trees
	
	#first look at sigsqs
	sigsqs_after_burnin = sigsqs[(length(sigsqs) - num_iterations_after_burn_in) : length(sigsqs)]
	avg_sigsqs_after_burn_in = mean(sigsqs_after_burnin, na.rm = TRUE)

	ppi_sigsqs = quantile(sigsqs[num_burn_in : length(sigsqs)], c(.025, .975))
	plot_df = data.frame(
		iter = seq_along(sigsqs),
		sigsq = sigsqs
	)
	y_limits = c(quantile(sigsqs, 0.01), quantile(sigsqs, 0.99))
	
	p = ggplot2::ggplot(plot_df, ggplot2::aes(x = iter, y = sigsq)) +
		ggplot2::geom_point(color = "gray70", alpha = 0.5, size = 0.6) +
		ggplot2::geom_point(color = "red", alpha = 0.4, size = 0.6) +
		ggplot2::geom_hline(yintercept = ppi_sigsqs, color = "darkgreen") +
		ggplot2::geom_hline(yintercept = avg_sigsqs_after_burn_in, color = "blue") +
		ggplot2::geom_vline(xintercept = num_burn_in, color = "gray") +
		ggplot2::coord_cartesian(ylim = y_limits) +
		ggplot2::labs(
			title = "Sigsq Estimates over MCMC Iteration",
			x = "MCMC Iteration (green lines: after burn-in 95% CI)",
			y = paste("Sigsq by MCMC Iteration, avg after burn-in =", round(avg_sigsqs_after_burn_in, 3))
		)
	
	if (bart_machine$num_cores > 1){
		core_lines = num_burn_in + (1 : (bart_machine$num_cores - 1)) *
			bart_machine$num_iterations_after_burn_in / bart_machine$num_cores
		p = p + ggplot2::geom_vline(xintercept = core_lines, color = "gray")
	}
	
	p
}

##function for investigating variable inclusion proportions
#' Explore Variable Inclusion Proportions in BART Model
#'
#' @description
#' Explore the variable inclusion proportions for a BART model to learn about the relative influence of the different covariates. This function includes an option to generate a plot of the variable inclusion proportions.
#'
#' @details
#' In the plot, the red bars correspond to the standard error of the variable inclusion proportion estimates.
#' @param bart_machine An object of class ``bartMachine''.
#' @param type If ``splits'', then the proportion of times each variable is chosen for a splitting rule is computed. If ``trees'', then the proportion of times each variable appears in a tree is computed.
#' @param plot If TRUE, a plot of the variable inclusion proportions is generated.
#' @param num_replicates_for_avg The number of replicates of BART to be used to generate variable inclusion proportions. Averaging across multiple BART models improves stability of the estimates. See Bleich et al. (2013) for more details.
#' @param num_trees_bottleneck Number of trees to be used in the sum-of-trees for computing the variable inclusion proportions. A small number of trees should be used to force the variables to compete for entry into the model. Chipman et al. (2010) recommend 20. See this reference for more details.
#' @param num_var_plot Number of variables to be shown on the plot. If ``Inf'', all variables are plotted.
#' @param bottom_margin A display parameter that adjusts the bottom margin of the graph if labels are clipped. The scale of this parameter is the same as set with \code{par(mar = c(....))} in R.
#'   Higher values allow for more space if the covariate names are long. Note that making this parameter too large will prevent plotting and the plot function in R will throw an error.
#' @param verbose If TRUE, prints progress messages and plots to the active device.
#'
#' @return
#' Invisibly, returns a list with the following components:
#' \item{avg_var_props}{The average variable inclusion proportions for each variable\cr (across \code{num_replicates_for_avg})}
#' \item{sd_var_props}{The standard deviation of the variable inclusion proportions for each variable (across \code{num_replicates_for_avg})}
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#' 
#' J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
#' Additive Regression Trees. ArXiv e-prints, 2013.
#' 
#' HA Chipman, EI George, and RE McCulloch. BART: Bayesian Additive Regressive Trees.
#' The Annals of Applied Statistics, 4(1): 266--298, 2010.
#'
#' @seealso
#' \code{\link{interaction_investigator}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 10
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y, num_trees = 20)
#' 
#' #investigate variable inclusion proportions
#' investigate_var_importance(bart_machine)
#' }
#' @export
investigate_var_importance = function(bart_machine, type = "splits", plot = TRUE, num_replicates_for_avg = 5, num_trees_bottleneck = 20, num_var_plot = Inf, bottom_margin = 10, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_choice(type, c("splits", "trees"))
  assert_flag(plot)
  assert_int(num_replicates_for_avg, lower = 1)
  assert_int(num_trees_bottleneck, lower = 1)
  assert_number(num_var_plot, lower = 1) # Inf is allowed
  assert_number(bottom_margin, lower = 0)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	var_props = array(0, c(num_replicates_for_avg, bart_machine$p))
	for (i in 1 : num_replicates_for_avg){
		if (i == 1 & num_trees_bottleneck == bart_machine$num_trees){ ##if original BART is using right number of trees
			var_props[i, ] = get_var_props_over_chain(bart_machine, type)
		} else {
			bart_machine_dup = bart_machine_duplicate(bart_machine, num_trees = num_trees_bottleneck, run_in_sample = FALSE, verbose = FALSE)			
			var_props[i, ] = get_var_props_over_chain(bart_machine_dup, type)				
		}
		if (verbose){
			cat(".")
		}
	}
	if (verbose){
		cat("\n")
	}
	
	avg_var_props = colMeans(var_props)
	names(avg_var_props) = bart_machine$training_data_features_with_missing_features
	sd_var_props = matrixStats::colSds(var_props)
	names(sd_var_props) = bart_machine$training_data_features_with_missing_features
	
	if (num_var_plot == Inf){
		num_var_plot = bart_machine$p
	}
	
	avg_var_props_sorted_indices = sort(avg_var_props, decreasing = TRUE, index.return = TRUE)$ix
	avg_var_props = avg_var_props[avg_var_props_sorted_indices][1 : num_var_plot]
	sd_var_props = sd_var_props[avg_var_props_sorted_indices][1 : num_var_plot]		
	
	plot_obj = NULL
	if (plot){
		if (is.na(sd_var_props[1])){
			moe = 0
		} else {
			moe = 1.96 * sd_var_props / sqrt(num_replicates_for_avg)
		}
		conf_upper = avg_var_props + 1.96 * sd_var_props / sqrt(num_replicates_for_avg)
		conf_lower = avg_var_props - 1.96 * sd_var_props / sqrt(num_replicates_for_avg)
		plot_df = data.frame(
			variable = factor(names(avg_var_props), levels = names(avg_var_props)),
			value = as.numeric(avg_var_props),
			conf_lower = as.numeric(conf_lower),
			conf_upper = as.numeric(conf_upper)
		)
		p = ggplot2::ggplot(plot_df, ggplot2::aes(x = variable, y = value)) +
			ggplot2::geom_col(fill = "gray") +
			ggplot2::labs(y = "Inclusion Proportion", x = NULL) +
			ggplot2::coord_cartesian(ylim = c(0, max(avg_var_props + moe))) +
			ggplot2::theme(
				axis.text.x = ggplot2::element_text(angle = 90, vjust = 0.5, hjust = 1),
				plot.margin = ggplot2::margin(t = 5.5, r = 5.5, b = bottom_margin, l = 5.5)
			)
		if (!is.na(sd_var_props[1])){
			p = p + ggplot2::geom_errorbar(
				ggplot2::aes(ymin = conf_lower, ymax = conf_upper),
				color = rgb(0.59, 0.39, 0.39),
				width = 0.2,
				linewidth = 0.6
			)
		}
		if (verbose){
			print(p)
		}
		plot_obj = p
	}
	invisible(list(avg_var_props = avg_var_props, sd_var_props = sd_var_props, plot = plot_obj))	
}

##user function calling private plotting methods
#' Plot Convergence Diagnostics
#'
#' @description
#' A suite of plots to assess convergence diagnostics and features of the BART model.
#'
#' @details
#' The ``sigsqs'' option plots the posterior error variance estimates by the Gibbs sample number. This is a standard tool to assess convergence of MCMC algorithms. This option is not applicable to classification BART models.\cr
#' The ``mh_acceptance'' option plots the proportion of Metropolis-Hastings steps accepted for each Gibbs sample (number accepted divided by number of trees).\cr
#' The ``num_nodes'' option plots the average number of nodes across each tree in the sum-of-trees model by the Gibbs sample number (for post burn-in only). The blue line
#' is the average number of nodes over all trees.\cr
#' The ``tree_depths'' option plots the average tree depth across each tree in the sum-of-trees model by the Gibbs sample number (for post burn-in only). The blue line
#' is the average number of nodes over all trees.
#' @param bart_machine An object of class ``bartMachine''.
#' @param plots The list of plots to be displayed. The four options are: "sigsqs", "mh_acceptance", "num_nodes", "tree_depths".
#' @param verbose If TRUE, prints plots to the active device.
#'
#' @return
#' None.
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' The ``sigsqs'' plot separates the burn-in \eqn{\sigma^2}'s for the first core by post burn-in \eqn{\sigma^2}'s estimates for all cores by grey vertical lines.
#' The ``mh_acceptance'' plot separates burn-in from post-burn in by a grey vertical line. Post burn-in, the different core proportions plot in different colors.
#' The ``num_nodes'' plot separates different core estimates by vertical lines (post burn-in only).
#' The `tree_depths'' plot separates different core estimates by vertical lines (post burn-in only).
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #plot convergence diagnostics
#' plot_convergence_diagnostics(bart_machine)
#' }
#' @export
plot_convergence_diagnostics = function(bart_machine, plots = c("sigsqs", "mh_acceptance", "num_nodes", "tree_depths"), verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_subset(plots, c("sigsqs", "mh_acceptance", "num_nodes", "tree_depths"))
  assert_flag(verbose)

  check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
  plot_list = list()
  if ("sigsqs" %in% plots){
    if (bart_machine$pred_type == "regression"){
      plot_list = c(plot_list, list(plot_sigsqs_convergence_diagnostics(bart_machine)))
    }
  }
	if ("mh_acceptance" %in% plots){
	  plot_list = c(plot_list, list(plot_mh_acceptance_reject(bart_machine)))
	}
	if ("num_nodes" %in% plots){
	  plot_list = c(plot_list, list(plot_tree_num_nodes(bart_machine)))
	}
	if ("tree_depths" %in% plots){
	  plot_list = c(plot_list, list(plot_tree_depths(bart_machine)))
	}
	num_plots = length(plot_list)
	if (num_plots == 0){
		return(invisible(list()))
	}
	if (num_plots == 1){
		if (verbose){
			print(plot_list[[1]])
		}
		return(invisible(plot_list))
	}
	if (num_plots <= 2){
		num_cols = 2
		num_rows = 1
	} else {
		num_cols = 2
		num_rows = 2
	}
	if (verbose){
		grid::grid.newpage()
		grid::pushViewport(grid::viewport(layout = grid::grid.layout(num_rows, num_cols)))
		for (i in seq_len(num_plots)){
			row_idx = ceiling(i / num_cols)
			col_idx = i - (row_idx - 1) * num_cols
			suppressMessages(print(
				plot_list[[i]],
				vp = grid::viewport(layout.pos.row = row_idx, layout.pos.col = col_idx)
			))
		}
		grid::popViewport()
	}
	invisible(plot_list)
}

##private function
shapiro_wilk_p_val = function(vec){
	tryCatch(shapiro.test(vec)$p.value, error = function(e){})
}

##function for investigating interactions
#' Explore Pairwise Interactions in BART Model
#'
#' @description
#' Explore the pairwise interaction counts for a BART model to learn about interactions fit by the model. This function includes an option to generate a plot of the pairwise interaction counts.
#'
#' @details
#' An interaction between two variables is considered to occur whenever a path from any node of a tree to
#' any of its terminal node contains splits using those two variables. See Kapelner and Bleich, 2013, Section 4.11.
#' @param bart_machine An object of class ``bartMachine''.
#' @param plot If TRUE, a plot of the pairwise interaction counts is generated.
#' @param num_replicates_for_avg The number of replicates of BART to be used to generate pairwise interaction inclusion counts.
#'   Averaging across multiple BART models improves stability of the estimates.
#' @param num_trees_bottleneck Number of trees to be used in the sum-of-trees model for computing pairwise interactions counts.
#'   A small number of trees should be used to force the variables to compete for entry into the model.
#' @param num_var_plot Number of variables to be shown on the plot. If ``Inf,'' all variables are plotted (not recommended if
#'   the number of predictors is large). Default is 50.
#' @param cut_bottom A display parameter between 0 and 1 that controls where the y-axis is plotted. A value of 0 would begin the y-axis at 0; a value of 1 begins
#'   the y-axis at the minimum of the average pairwise interaction inclusion count (the smallest bar in the bar plot). Values between 0 and 1 begin the
#'   y-axis as a percentage of that minimum.
#' @param bottom_margin A display parameter that adjusts the bottom margin of the graph if labels are clipped. The scale of this parameter is the same as set with \code{par(mar = c(....))} in R.
#'   Higher values allow for more space if the crossed covariate names are long. Note that making this parameter too large will prevent plotting and the plot function in R will throw an error.
#' @param verbose If TRUE, prints progress messages and plots to the active device.
#'
#' @return
#' \item{interaction_counts}{For each of the \eqn{p \times p}{p times p} interactions, what is the count across all \code{num_replicates_for_avg}
#' BART model replicates' post burn-in Gibbs samples in all trees.}
#' 	\item{interaction_counts_avg}{For each of the \eqn{p \times p}{p times p} interactions, what is the average count across all \code{num_replicates_for_avg}
#' BART model replicates' post burn-in Gibbs samples in all trees.}
#' 	\item{interaction_counts_sd}{For each of the \eqn{p \times p}{p times p} interactions, what is the sd of the interaction counts across the \code{num_replicates_for_avg}
#' BART models replicates.}
#' 	\item{interaction_counts_avg_and_sd_long}{For each of the \eqn{p \times p}{p times p} interactions, what is the average and sd of the interaction counts across the \code{num_replicates_for_avg}
#' BART models replicates. The output is organized as a convenient long table of class \code{data.frame}.}
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#'
#' @seealso
#' \code{\link{investigate_var_importance}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' In the plot, the red bars correspond to the standard error of the variable inclusion proportion estimates (since multiple replicates were used).
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 10
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y, num_trees = 20)
#' 
#' #investigate interactions
#' interaction_investigator(bart_machine)
#' }
#' @export
interaction_investigator = function(bart_machine, plot = TRUE, num_replicates_for_avg = 5, num_trees_bottleneck = 20, num_var_plot = 50, cut_bottom = NULL, bottom_margin = 10, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  assert_flag(plot)
  assert_int(num_replicates_for_avg, lower = 1)
  assert_int(num_trees_bottleneck, lower = 1)
  assert_number(num_var_plot, lower = 1)
  assert_number(cut_bottom, lower = 0, upper = 1, null.ok = TRUE)
  assert_number(bottom_margin, lower = 0)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	interaction_counts = array(NA, c(bart_machine$p, bart_machine$p, num_replicates_for_avg))
	
	for (r in 1 : num_replicates_for_avg){
		if (r == 1 & num_trees_bottleneck == bart_machine$num_trees){
			interaction_counts[, , r] = .jcall(bart_machine$java_bart_machine, "[[I", "getInteractionCounts", simplify = TRUE)
		} else {
			bart_machine_dup = bart_machine_duplicate(bart_machine, num_trees = num_trees_bottleneck)			
			interaction_counts[, , r] = .jcall(bart_machine_dup$java_bart_machine, "[[I", "getInteractionCounts", simplify = TRUE)
			if (verbose){
				cat(".")
				if (r %% 40 == 0){
					cat("\n")
				}
			}
		}
	}
	if (verbose){
		cat("\n")
	}
	
	interaction_counts_avg = matrix(NA, nrow = bart_machine$p, ncol = bart_machine$p)
	interaction_counts_sd = matrix(NA, nrow = bart_machine$p, ncol = bart_machine$p)
	
	if (bart_machine$use_missing_data){
		rownames(interaction_counts_avg) = bart_machine$training_data_features_with_missing_features
	    colnames(interaction_counts_avg) = bart_machine$training_data_features_with_missing_features
		rownames(interaction_counts_sd) = bart_machine$training_data_features_with_missing_features
		colnames(interaction_counts_sd) = bart_machine$training_data_features_with_missing_features
	} else {
	    rownames(interaction_counts_avg) = bart_machine$training_data_features
	    colnames(interaction_counts_avg) = bart_machine$training_data_features
		rownames(interaction_counts_sd) = bart_machine$training_data_features
		colnames(interaction_counts_sd) = bart_machine$training_data_features	
	}
	
	#now vectorize the interaction counts
	n_interactions = bart_machine$p * (bart_machine$p - 1) / 2
	avg_counts = array(NA, n_interactions)
	sd_counts = array(NA, n_interactions)
	interaction_counts_avg_and_sd_long = data.frame(
		var1 = as.character(rep(NA, n_interactions)),
		var2 = as.character(rep(NA, n_interactions)),
		avg_interaction = as.numeric(rep(NA, n_interactions)),
		se_interaction = as.numeric(rep(NA, n_interactions))
	)
	iter = 1
	for (i in 1 : bart_machine$p){
		for (j in 1 : bart_machine$p){
			if (i <= j){
				interactions_i_j = c(interaction_counts[i, j, ], interaction_counts[j, i, ]) #triangularize the data
				interaction_counts_avg[i, j] = mean(interactions_i_j)
				interaction_counts_sd[i, j] = sd(interactions_i_j)
				
				avg_counts[iter] = interaction_counts_avg[i, j]
				sd_counts[iter] = interaction_counts_sd[i, j]
				names(avg_counts)[iter] = paste(rownames(interaction_counts_avg)[i], "x", rownames(interaction_counts_avg)[j])
				
				interaction_counts_avg_and_sd_long[iter, ] = c(
					rownames(interaction_counts_avg)[i],
					rownames(interaction_counts_avg)[j],
					interaction_counts_avg[i, j],
					interaction_counts_sd[i, j] / sqrt(num_replicates_for_avg)
				)
				iter = iter + 1
			}
		}
	}
	
	if (num_var_plot == Inf || num_var_plot > n_interactions){
		num_var_plot = n_interactions
	}
	
	avg_counts_sorted_indices = sort(avg_counts, decreasing = TRUE, index.return = TRUE)$ix
	avg_counts = avg_counts[avg_counts_sorted_indices][1 : num_var_plot]
	sd_counts = sd_counts[avg_counts_sorted_indices][1 : num_var_plot]
	interaction_counts_avg_and_sd_long = interaction_counts_avg_and_sd_long[avg_counts_sorted_indices, ]
	
	if (is.null(cut_bottom)){
		ylim_bottom = 0
	} else {
		ylim_bottom = cut_bottom * min(avg_counts)
	}
	plot_obj = NULL
	if (plot){
		#now create the bar plot
		if (is.na(sd_counts[1])){
			moe = 0
		} else {
			moe = 1.96 * sd_counts / sqrt(num_replicates_for_avg)
		}
		conf_upper = avg_counts + 1.96 * sd_counts / sqrt(num_replicates_for_avg)
		conf_lower = avg_counts - 1.96 * sd_counts / sqrt(num_replicates_for_avg)
		plot_df = data.frame(
			interaction = factor(names(avg_counts), levels = names(avg_counts)),
			value = as.numeric(avg_counts),
			conf_lower = as.numeric(conf_lower),
			conf_upper = as.numeric(conf_upper)
		)
		p = ggplot2::ggplot(plot_df, ggplot2::aes(x = interaction, y = value)) +
			ggplot2::geom_col(fill = "gray") +
			ggplot2::labs(y = "Relative Importance", x = NULL) +
			ggplot2::coord_cartesian(ylim = c(ylim_bottom, max(avg_counts + moe))) +
			ggplot2::theme(
				axis.text.x = ggplot2::element_text(angle = 90, vjust = 0.5, hjust = 1),
				plot.margin = ggplot2::margin(t = 5.5, r = 5.5, b = bottom_margin, l = 5.5)
			)
		if (!is.na(sd_counts[1])){
			p = p + ggplot2::geom_errorbar(
				ggplot2::aes(ymin = conf_lower, ymax = conf_upper),
				color = rgb(0.59, 0.39, 0.39),
				width = 0.2,
				linewidth = 0.6
			)
		}
		if (verbose){
			print(p)
		}
		plot_obj = p
	}
		
	invisible(list(
		interaction_counts = interaction_counts,
		interaction_counts_avg = interaction_counts_avg,
		interaction_counts_sd = interaction_counts_sd,
		interaction_counts_avg_and_sd_long = interaction_counts_avg_and_sd_long,
		plot = plot_obj
	))
}

##partial dependence plot
#' Partial Dependence Plot
#'
#' @description
#' Creates a partial dependence plot for a BART model for regression or classification.
#'
#' @details
#' For regression models, the units on the y-axis are the same as the units of the response. For classification models, the units on the y-axis are probits.
#' @param bart_machine An object of class ``bartMachine''.
#' @param j The number or name of the column in the design matrix for which the partial dependence plot is to be created.
#' @param levs Quantiles at which the partial dependence function should be evaluated. Linear extrapolation is performed between these points.
#' @param lower_ci Lower limit for credible interval
#' @param upper_ci Upper limit for credible interval
#' @param prop_data The proportion of the training data to use. Default is 1. Use a lower proportion for speedier pd_plots. The closer to 1, the more resolution
#'   the PD plot will have; the closer to 0, the lower but faster.
#' @param verbose If TRUE, prints progress messages and plots to the active device.
#'
#' @return
#' Invisibly, returns a list with the following components:
#' 
#'   \item{x_j_quants}{Quantiles at which the partial dependence function is evaluated}
#'   \item{bart_avg_predictions_by_quantile_by_gibbs}{All samples of \eqn{\hat{f}(x)}}
#'   \item{bart_avg_predictions_by_quantile}{Posterior means for \eqn{\hat{f}(x)} at \code{x_j_quants}}
#'   \item{bart_avg_predictions_lower}{Lower bound of the desired confidence of the credible interval of \eqn{\hat{f}(x)}}
#'   \item{bart_avg_predictions_upper}{Upper bound of the desired confidence of the credible interval of \eqn{\hat{f}(x)}}
#'   \item{prop_data}{The proportion of the training data to use as specified when this function was executed}
#' %% ...
#'
#' @references
#' Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
#' with Bayesian Additive Regression Trees. Journal of Statistical
#' Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
#' 
#' HA Chipman, EI George, and RE McCulloch. BART: Bayesian Additive Regressive Trees.
#' The Annals of Applied Statistics, 4(1): 266--298, 2010.
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #Regression example
#' 
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 5
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y)
#' 
#' #partial dependence plot for quadratic term
#' pd_plot(bart_machine, "X3")
#' 
#' 
#' #Classification example
#' 
#' #get data and only use 2 factors
#' data(iris)
#' iris2 = iris[51:150,]
#' iris2$Species = factor(iris2$Species)
#' 
#' #build BART classification model
#' bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)
#' 
#' #partial dependence plot
#' pd_plot(bart_machine, "Petal.Width")
#' }
#' @export
pd_plot = function(bart_machine, j, levs = c(0.05, seq(from = 0.10, to = 0.90, by = 0.10), 0.95), lower_ci = 0.025, upper_ci = 0.975, prop_data = 1, verbose = TRUE){
  assert_class(bart_machine, "bartMachine")
  # j checked manually later because it can be numeric or character
  assert_numeric(levs, lower = 0, upper = 1, min.len = 1)
  assert_number(lower_ci, lower = 0, upper = 1)
  assert_number(upper_ci, lower = 0, upper = 1)
  assert_number(prop_data, lower = 0, upper = 1)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	if (inherits(j, "integer")){
		j = as.numeric(j)
	}
	if (inherits(j, "numeric") && (j < 1 || j > bart_machine$p)){
		stop(paste("You must set j to a number between 1 and p =", bart_machine$p))
	} else if (inherits(j, "character") && !(j %in% bart_machine$training_data_features)){
		stop("j must be the name of one of the training features (see \"<bart_model>$training_data_features\")")
	} else if (!(inherits(j, "numeric") || inherits(j, "character"))){
		stop("j must be a column number or column name")
	}
	
	x_j = bart_machine$model_matrix_training_data[, j]
	
	#fail with a warning if there's only one value (we don't want an error because it would fail on loops).
	if (length(unique(na.omit(x_j))) <= 1){
		warning("There must be more than one unique value in this training feature. PD plot not generated.")
		return()
	}
	x_j_quants = unique(quantile(x_j, levs, na.rm = TRUE))
	
	#fail with a warning if there's only one value (we don't want an error because it would fail on loops).
	if (length(unique(x_j_quants)) <= 1){
		warning("There must be more than one unique value among the quantiles selected. PD plot not generated.")
		return()
	}
	
	n_pd_plot = round(bart_machine$n * prop_data)
	bart_predictions_by_quantile = array(NA, c(length(x_j_quants), n_pd_plot, bart_machine$num_iterations_after_burn_in))
	
	for (q in 1 : length(x_j_quants)){
		#pull out a certain proportion of the data randomly
		indices = sample(1 : bart_machine$n, n_pd_plot)
		#now create test data matrix
		test_data = bart_machine$X[indices, ]
		test_data[, j] = rep(x_j_quants[q], n_pd_plot)
		
		bart_predictions_by_quantile[q, , ] = bart_machine_get_posterior(bart_machine, test_data, verbose = verbose)$y_hat_posterior_samples
		if (verbose){
			cat(".")
		}
	}
	if (verbose){
		cat("\n")
	}
	
  	if (bart_machine$pred_type == "classification"){ ##convert to probits
    	bart_predictions_by_quantile = qnorm(bart_predictions_by_quantile)
  	}
  
	bart_avg_predictions_by_quantile_by_gibbs = apply(bart_predictions_by_quantile, c(1, 3), mean)
	
	bart_avg_predictions_by_quantile = rowMeans(bart_avg_predictions_by_quantile_by_gibbs)
	bart_avg_predictions_lower = as.numeric(matrixStats::rowQuantiles(
		bart_avg_predictions_by_quantile_by_gibbs,
		probs = lower_ci
	))
	bart_avg_predictions_upper = as.numeric(matrixStats::rowQuantiles(
		bart_avg_predictions_by_quantile_by_gibbs,
		probs = upper_ci
	))
	
	var_name = ifelse(class(j) == "character", j, bart_machine$training_data_features[j])
    ylab_name = ifelse(bart_machine$pred_type == "classification", "Partial Effect (Probits)", "Partial Effect")
	pd_df = data.frame(
		quantile = x_j_quants,
		avg = bart_avg_predictions_by_quantile,
		lower = bart_avg_predictions_lower,
		upper = bart_avg_predictions_upper
	)
	y_limits = c(min(bart_avg_predictions_lower, bart_avg_predictions_upper),
		max(bart_avg_predictions_lower, bart_avg_predictions_upper))
	p = ggplot2::ggplot(pd_df, ggplot2::aes(x = quantile, y = avg)) +
		ggplot2::geom_ribbon(ggplot2::aes(ymin = lower, ymax = upper), fill = "gray87") +
		ggplot2::geom_line(color = "blue", ggplot2::aes(y = lower)) +
		ggplot2::geom_line(color = "blue", ggplot2::aes(y = upper)) +
		ggplot2::geom_line(linewidth = 1) +
		ggplot2::geom_point() +
		ggplot2::coord_cartesian(ylim = y_limits) +
		ggplot2::labs(
			title = "Partial Dependence Plot",
			y = ylab_name,
			x = paste(var_name, "plotted at specified quantiles")
		)
	if (verbose){
		print(p)
	}
	
	invisible(list(
		x_j_quants = x_j_quants, 
		bart_avg_predictions_by_quantile_by_gibbs = bart_avg_predictions_by_quantile_by_gibbs, 
		bart_avg_predictions_by_quantile = bart_avg_predictions_by_quantile, 
		bart_avg_predictions_lower = bart_avg_predictions_lower,
		bart_avg_predictions_upper = bart_avg_predictions_upper,
		prop_data = prop_data,
		plot = p
	))
}

##plot and invisibly return out-of-sample RMSE by the number of trees
#' Assess the Out-of-sample RMSE by Number of Trees
#'
#' @description
#' Assess out-of-sample RMSE of a BART model for varying numbers of trees in the sum-of-trees model.
#' @param bart_machine An object of class ``bartMachine''.
#' @param tree_list List of sizes for the sum-of-trees models.
#' @param in_sample If TRUE, the RMSE is computed on in-sample data rather than an out-of-sample holdout.
#' @param plot If TRUE, a plot of the RMSE by the number of trees in the ensemble is created.
#' @param holdout_pctg Percentage of the data to be treated as an out-of-sample holdout.
#' @param num_replicates Number of replicates to average the results over. Each replicate uses a randomly sampled holdout of the data, (which could have overlap).
#' @param verbose If TRUE, prints progress messages and plots to the active device.
#' @param ... Other arguments to be passed to the plot function.
#'
#' @return
#' Invisibly, returns the out-of-sample average RMSEs for each tree size.
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @note
#' Since using a large number of trees can substantially increase computation time, this plot can help assess whether a smaller ensemble size is sufficient to obtain desirable predictive performance.
#' This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 10
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y, num_trees = 20)
#' 
#' #explore RMSE by number of trees
#' rmse_by_num_trees(bart_machine)
#' }
#' @export
rmse_by_num_trees = function(bart_machine, tree_list = c(5, seq(10, 50, 10), 100, 150, 200), in_sample = FALSE, plot = TRUE, holdout_pctg = 0.3, num_replicates = 4, verbose = TRUE, ...){
  assert_class(bart_machine, "bartMachine")
  assert_integerish(tree_list, lower = 1, min.len = 1)
  assert_flag(in_sample)
  assert_flag(plot)
  assert_number(holdout_pctg, lower = 0, upper = 1)
  assert_int(num_replicates, lower = 1)
  assert_flag(verbose)

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (bart_machine$pred_type == "classification"){
		stop("This function does not work for classification.")
	}		
	X = bart_machine$X
	y = bart_machine$y
	n = bart_machine$n
	
	rmses = array(NA, c(num_replicates, length(tree_list)))
	if (verbose){
		cat("num_trees = ")
	}
	for (t in 1 : length(tree_list)){
		for (r in 1 : num_replicates){
			if (in_sample){
				bart_machine_dup = bart_machine_duplicate(bart_machine, num_trees = tree_list[t], run_in_sample = TRUE)
				rmses[r, t] = bart_machine_dup$rmse_train				
			} else {
				holdout_indicies = sample(1 : n, holdout_pctg * n)
				Xtrain = X[setdiff(1 : n, holdout_indicies), ]
				ytrain = y[setdiff(1 : n, holdout_indicies)]
				Xtest = X[holdout_indicies, ]
				ytest = y[holdout_indicies]
				
				bart_machine_dup = bart_machine_duplicate(bart_machine, Xtrain, ytrain, num_trees = tree_list[t])
				predict_obj = suppressWarnings(bart_predict_for_test_data(
					bart_machine_dup,
					Xtest,
					ytest,
					verbose = verbose
				)) ##predict on holdout
				rmses[r, t] = predict_obj$rmse				
			}
			if (verbose){
				cat("..")
				cat(tree_list[t])
			}			
		}
	}
	if (verbose){
		cat("\n")
	}
	
	rmse_means = colMeans(rmses, na.rm = TRUE)

	if (plot){ ##plotting
		if (num_replicates > 1){
			rmse_sds = matrixStats::colSds(rmses)
		} else {
			rmse_sds = rep(0, length(tree_list))
		}
		y_mins = rmse_means - 2 * rmse_sds
		y_maxs = rmse_means + 2 * rmse_sds
		plot_df = data.frame(
			num_trees = tree_list,
			rmse = rmse_means,
			lower = rmse_means - 1.96 * rmse_sds / sqrt(num_replicates),
			upper = rmse_means + 1.96 * rmse_sds / sqrt(num_replicates)
		)
		plot_df = plot_df[is.finite(plot_df$rmse), , drop = FALSE]
		finite_vals = c(plot_df$rmse, plot_df$lower, plot_df$upper)
		finite_vals = finite_vals[is.finite(finite_vals)]
		if (length(finite_vals) == 0){
			warning("RMSE estimates are not finite; skipping plot.")
			return(invisible(rmse_means))
		}
		y_limits = range(finite_vals)
		p = ggplot2::ggplot(plot_df, ggplot2::aes(x = num_trees, y = rmse)) +
			ggplot2::geom_line() +
			ggplot2::geom_point() +
			ggplot2::coord_cartesian(ylim = y_limits) +
			ggplot2::labs(
				title = paste(ifelse(in_sample, "In-Sample", "Out-Of-Sample"), "RMSE by Number of Trees"),
				x = "Number of Trees",
				y = paste(ifelse(in_sample, "In-Sample", "Out-Of-Sample"), "RMSE")
			)
		if (num_replicates > 1){
			p = p + ggplot2::geom_errorbar(
				ggplot2::aes(ymin = lower, ymax = upper),
				color = "grey",
				width = 0.2,
				linewidth = 0.3
			)
		}
		if (verbose){
			print(p)
		}
	}
	if (plot){
		attr(rmse_means, "plot") = p
	}
	invisible(rmse_means)
}

source("../testthat/helper-bartmachine-models.R")


test_that("node prediction and projection helpers return expected shapes", {
  reg_model <- build_regression_model()
  idx <- node_prediction_training_data_indices(reg_model)
  expect_true(is.array(idx))
  expect_equal(dim(idx)[BART_TESTS$first_dim_index], reg_model$n)

  weights <- get_projection_weights(reg_model)
  expect_equal(nrow(weights), reg_model$n)
  expect_equal(ncol(weights), reg_model$n)

  weights_kludge <- get_projection_weights(reg_model, regression_kludge = TRUE)
  expect_equal(dim(weights_kludge), dim(weights))
})

test_that("posterior and interval helpers return expected structures", {
  reg_model <- build_regression_model()
  idx <- seq_len(BART_TESTS$posterior_subset_n)
  posterior <- bart_machine_get_posterior(reg_model, reg_model$X[idx, , drop = FALSE], verbose = BART_TESTS$verbose)
  expect_true(is.list(posterior))
  expect_true("y_hat" %in% names(posterior))

  cred <- calc_credible_intervals(reg_model, reg_model$X[idx, , drop = FALSE])
  expect_equal(nrow(cred), BART_TESTS$posterior_subset_n)
  expect_equal(ncol(cred), BART_TESTS$credible_interval_cols)

  pred <- calc_prediction_intervals(
    reg_model,
    reg_model$X[idx, , drop = FALSE],
    num_samples_per_data_point = BART_TESTS$prediction_samples_per_point
  )
  expect_true(is.matrix(pred))
  expect_equal(nrow(pred), BART_TESTS$posterior_subset_n)

  tol <- 1e-8
  expect_true(all(cred[, 1] >= pred[, 1] - tol))
  expect_true(all(cred[, 2] <= pred[, 2] + tol))
  expect_true(all(posterior$y_hat >= cred[, 1] - tol))
  expect_true(all(posterior$y_hat <= cred[, 2] + tol))
  expect_true(all(posterior$y_hat >= pred[, 1] - tol))
  expect_true(all(posterior$y_hat <= pred[, 2] + tol))
})

test_that("covariate tests and linearity test return p-values", {
  reg_model <- build_regression_model()
  cov_test <- cov_importance_test(
    reg_model,
    covariates = BART_TESTS$covariate_index,
    num_permutation_samples = BART_TESTS$num_permute_samples,
    plot = FALSE,
    verbose = BART_TESTS$verbose
  )
  expect_true(is.list(cov_test))
  expect_true("pval" %in% names(cov_test))

  lin_test <- linearity_test(
    X = reg_model$X,
    y = reg_model$y,
    num_permutation_samples = BART_TESTS$num_permute_samples,
    plot = FALSE,
    num_trees = BART_TESTS$linearity_num_trees,
    num_burn_in = BART_TESTS$linearity_num_burn_in,
    num_iterations_after_burn_in = BART_TESTS$linearity_num_iter,
    verbose = BART_TESTS$verbose
  )
  expect_true(is.list(lin_test))
  expect_true("pval" %in% names(lin_test))
})

test_that("plotting diagnostics return ggplot objects or lists", {
  reg_model <- build_regression_model()
  with_plot_device({
    plots <- check_bart_error_assumptions(reg_model, verbose = BART_TESTS$verbose)
    expect_true(inherits(plots$qq_plot, "ggplot"))
    expect_true(inherits(plots$hetero_plot, "ggplot"))
  })

  with_plot_device({
    plot_obj <- plot_y_vs_yhat(reg_model, verbose = BART_TESTS$verbose)
    expect_true(inherits(plot_obj, "ggplot"))
  })

  with_plot_device({
    plot_obj <- plot_y_vs_yhat(reg_model, credible_intervals = TRUE, verbose = BART_TESTS$verbose)
    expect_true(inherits(plot_obj, "ggplot"))
  })

  with_plot_device({
    sigsqs <- get_sigsqs(reg_model, plot_hist = TRUE, verbose = BART_TESTS$verbose)
    expect_true(inherits(attr(sigsqs, "plot"), "ggplot"))
  })

  with_plot_device({
    var_imp <- investigate_var_importance(
      reg_model,
      plot = TRUE,
      num_replicates_for_avg = BART_TESTS$num_replicates_for_avg,
      num_trees_bottleneck = BART_TESTS$num_trees_bottleneck,
      num_var_plot = BART_TESTS$num_var_plot,
      verbose = BART_TESTS$verbose
    )
    expect_true(inherits(var_imp$plot, "ggplot"))
  })

  with_plot_device({
    plots <- plot_convergence_diagnostics(reg_model, plots = c("sigsqs", "mh_acceptance"), verbose = BART_TESTS$verbose)
    expect_true(length(plots) >= BART_TESTS$min_plot_count)
  })

  with_plot_device({
    inter <- interaction_investigator(
      reg_model,
      plot = TRUE,
      num_replicates_for_avg = BART_TESTS$num_replicates_for_avg,
      num_trees_bottleneck = BART_TESTS$num_trees_bottleneck,
      num_var_plot = BART_TESTS$num_var_plot,
      verbose = BART_TESTS$verbose
    )
    expect_true(inherits(inter$plot, "ggplot"))
  })

  with_plot_device({
    pd <- pd_plot(
      reg_model,
      j = BART_TESTS$covariate_index,
      prop_data = BART_TESTS$pd_plot_prop_data,
      verbose = BART_TESTS$verbose
    )
    expect_true(inherits(pd$plot, "ggplot"))
  })

  with_plot_device({
    rmses <- rmse_by_num_trees(
      reg_model,
      tree_list = BART_TESTS$rmse_tree_list,
      in_sample = TRUE,
      plot = TRUE,
      num_replicates = BART_TESTS$rmse_num_replicates,
      verbose = BART_TESTS$verbose
    )
    expect_true(is.numeric(rmses))
    plot_obj <- attr(rmses, "plot")
    if (!is.null(plot_obj)) {
      expect_true(inherits(plot_obj, "ggplot"))
    }
  })
})

#' Data concerning automobile prices.
#'
#' @description
#' The \code{automobile} data frame has 201 rows and 25 columns and
#'   concerns automobiles in the 1985 Auto Imports Database. The response
#'   variable, \code{price}, is the log selling price of the automobile. There
#'   are 7 categorical predictors and 17 continuous / integer predictors which
#'   are features of the automobiles. 41 automobiles have missing data in one
#'   or more of the feature entries. This dataset is true to the original except
#'   with a few of the predictors dropped.
#'
#' @docType data
#' @name automobile
#' @usage data(automobile)
#' @source
#' K Bache and M Lichman. UCI machine learning repository, 2013.
#' http://archive.ics.uci.edu/ml/datasets/Automobile
#'
#' @keywords
#' datasets
NULL

#' benchmark_datasets
#'
#' @description
#' Nine diverse datasets which were used for benchmarking bartMachine's out of sample performance in
#'   the vignette for this package.
#'
#' @docType data
#' @name benchmark_datasets
#' @usage data(benchmark_datasets)
#' @source
#' See vignette for details.
#'
#' @keywords
#' datasets
#'
#' @aliases ankara baseball boston compactiv ozone pole triazine wine.red wine.white
NULL

#' bartMachine: Bayesian Additive Regression Trees
#'
#' @importFrom rJava .jpackage .jcall .jnew .jevalArray .jarray is.jnull .jcache .jcast .jnull
#' @import bartMachineJARs
#' @import randomForest
#' @import missForest
#' @import stats
#' @importFrom utils packageVersion tail
#' @importFrom grDevices rgb
#' @import grid
#' @import checkmate
#' @keywords internal
"_PACKAGE"

library(withr)

test_files <- paste0("../grid_tests/", c(
  "test-exports-basic.R",
  "test-exports-diagnostics.R",
  "test-exports-model-selection.R"
))

run_test_for_combo <- function(combo) {
  model_num_trees <- ifelse(
    is.null(combo$num_trees),
    BART_TESTS$model_num_trees,
    combo$num_trees
  )
  model_burn_in <- ifelse(
    is.null(combo$num_burn_in),
    BART_TESTS$model_burn_in,
    combo$num_burn_in
  )
  model_iter <- ifelse(
    is.null(combo$num_iter),
    BART_TESTS$model_iter,
    combo$num_iter
  )
  env_vars <- c(
    N_TEST = as.character(combo$n),
    P_TEST = as.character(combo$p),
    N_CLASS_TEST = as.character(combo$n),
    P_CLASS_TEST = as.character(combo$p),
    MODEL_NUM_TREES = as.character(model_num_trees),
    MODEL_BURN_IN = as.character(model_burn_in),
    MODEL_ITER = as.character(model_iter)
  )
  combo_label <- sprintf("n=%s p=%s trees=%s burn_in=%s iter=%s", combo$n, combo$p, model_num_trees, model_burn_in, model_iter)
  message("Running test grid for ", combo_label)
  with_envvar(env_vars, {
    for (test_file in test_files) {
      testthat::test_file(test_file)
    }
  })
}

for (combo in BART_TESTS$np_grid) {
  run_test_for_combo(combo)
}

##color array 
COLORS = array(NA, 500)
for (i in 1 : 500){
	COLORS[i] = rgb(runif(1, 0, 0.7), runif(1, 0, 0.7), runif(1, 0, 0.7))
}

##set number of cores in use
#' Set the Number of Cores for BART
#'
#' @description
#' Sets the number of cores to be used for all parallelized BART functions.
#' @param num_cores Number of cores to use. If the number of cores is more than 1, setting the seed during model construction
#'   cannot be deterministic.
#' @param verbose If TRUE, prints the updated core count.
#'
#' @return
#' None.
#'
#' @seealso
#' \code{\link{bart_machine_num_cores}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' #set all parallelized functions to use 4 cores
#' set_bart_machine_num_cores(4)
#' }
#' @export
set_bart_machine_num_cores = function(num_cores, verbose = TRUE){
  assert_int(num_cores, lower = 1)
  assert_flag(verbose)

	assign("BART_NUM_CORES", num_cores, bartMachine_globals)
	if (verbose){
		cat("bartMachine now using", num_cores, "cores.\n")
	}
}

##get number of cores in use
DEFAULT_BART_NUM_CORES = 1
#' Get Number of Cores Used by BART
#'
#' @description
#' Returns number of cores used by BART
#'
#' @details
#' Returns the number of cores currently being used by parallelized BART functions
#'
#' @return
#' %%  ~Describe the value returned
#' %%  If it is a LIST, use
#' Number of cores currently being used by parallelized BART functions.
#' %%  \item{comp2 }{Description of 'comp2'}
#' %% ...
#'
#' @seealso
#' \code{\link{set_bart_machine_num_cores}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' bart_machine_num_cores()
#' }
#' @export
bart_machine_num_cores = function(){
	if (exists("BART_NUM_CORES", envir = bartMachine_globals)){
		get("BART_NUM_CORES", bartMachine_globals)
	} else {
		DEFAULT_BART_NUM_CORES
	}
}

##get variable counts
#' Get the Variable Inclusion Counts
#'
#' @description
#' Computes the variable inclusion counts for a BART model.
#' @param bart_machine An object of class ``bartMachine''.
#' @param type If ``splits'', then the number of times each variable is chosen for a splitting rule is computed. If ``trees'', then the number of times each variable appears in a tree is computed.
#'
#' @return
#' Returns a matrix of counts of each predictor across all trees by Gibbs sample. Thus, the dimension is \code{num_iterations_after_burn_in}
#' by \code{p} (where \code{p} is the number of predictors after dummifying factors and adding missingness dummies if specified by \code{use_missing_data_dummies_as_covars}).
#'
#' @seealso
#' \code{\link{get_var_props_over_chain}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' 
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 10
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y, num_trees = 20)
#' 
#' #get variable inclusion counts
#' var_counts = get_var_counts_over_chain(bart_machine)
#' print(var_counts)
#' }
#' @export
get_var_counts_over_chain = function(bart_machine, type = "splits"){
  assert_class(bart_machine, "bartMachine")
  assert_choice(type, c("splits", "trees"))

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not

	if (!(type %in% c("trees", "splits"))){
		stop("type must be \"trees\" or \"splits\"")
	}
	C = .jcall(bart_machine$java_bart_machine, "[[I", "getCountsForAllAttribute", type, simplify = TRUE)
	colnames(C) = colnames(bart_machine$model_matrix_training_data)[1 : bart_machine$p]
	C
}

#get variable inclusion proportions
#' Get the Variable Inclusion Proportions
#'
#' @description
#' Computes the variable inclusion proportions for a BART model.
#' @param bart_machine An object of class ``bartMachine''.
#' @param type If ``splits'', then the proportion of times each variable is chosen for a splitting rule versus all splitting rules is computed. If ``trees'', then the proportion of times each variable appears in a tree versus all appearances of variables in trees is computed.
#'
#' @return
#' Returns a vector of the variable inclusion proportions.
#'
#' @seealso
#' \code{\link{get_var_counts_over_chain}}
#'
#' @author
#' Adam Kapelner and Justin Bleich
#'
#' @examples
#' \dontrun{
#' #generate Friedman data
#' set.seed(11)
#' n  = 200
#' p = 10
#' X = data.frame(matrix(runif(n * p), ncol = p))
#' y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
#' 
#' ##build BART regression model
#' bart_machine = bartMachine(X, y, num_trees = 20)
#' 
#' #Get variable inclusion proportions
#' var_props = get_var_props_over_chain(bart_machine)
#' print(var_props)
#' }
#' @export
get_var_props_over_chain = function(bart_machine, type = "splits"){
  assert_class(bart_machine, "bartMachine")
  assert_choice(type, c("splits", "trees"))

	check_serialization(bart_machine) #ensure the Java object exists and fire an error if not
	
	if (!(type %in% c("trees", "splits"))){
		stop("type must be \"trees\" or \"splits\"")
	}	
	attribute_props = .jcall(bart_machine$java_bart_machine, "[D", "getAttributeProps", type)
	names(attribute_props) = colnames(bart_machine$model_matrix_training_data)[1 : bart_machine$p]
	attribute_props
}

##private function called in summary() 
sigsq_est = function(bart_machine){
	sigsqs = .jcall(bart_machine$java_bart_machine, "[D", "getGibbsSamplesSigsqs")	
	sigsqs_after_burnin = sigsqs[(length(sigsqs) - bart_machine$num_iterations_after_burn_in) : length(sigsqs)]	
	mean(sigsqs_after_burnin)
}

#There's no standard R function for this.
sample_mode = function(data){
	as.numeric(names(sort(-table(data)))[1])
}

check_serialization = function(bart_machine){
	if (is.jnull(bart_machine$java_bart_machine)){
		stop("This bartMachine object was loaded from an R image but was not serialized.\n  Please build bartMachine using the option \"serialize = TRUE\" next time.\n")
	}
}

test_that("bartMachine handles missing data", {
  # Generate data with missing values
  set.seed(BART_TESTS$seed)
  n <- BART_TESTS$small_data_n
  p <- BART_TESTS$default_p
  X <- data.frame(matrix(runif(n * p), ncol = p))
  
  # Introduce missingness
  X[1:5, 1] <- NA
  X[6:10, 2] <- NA
  
  y <- BART_TESTS$reg_coef_1 * ifelse(is.na(X[, 1]), 0.5, X[, 1]) +
       BART_TESTS$reg_coef_2 * ifelse(is.na(X[, 2]), 0.5, X[, 2]) +
       rnorm(n, sd = BART_TESTS$small_data_sd)
  
  # Test with use_missing_data = TRUE
  model_missing <- bartMachine(
    X, 
    y,
    num_trees = BART_TESTS$small_model_num_trees,
    num_burn_in = BART_TESTS$small_model_burn_in,
    num_iterations_after_burn_in = BART_TESTS$small_model_iter,
    use_missing_data = TRUE,
    verbose = BART_TESTS$verbose
  )
  expect_s3_class(model_missing, "bartMachine")
  expect_true(model_missing$use_missing_data)
  
  # Test with use_missing_data_dummies_as_covars = TRUE
  model_dummies <- bartMachine(
    X, 
    y,
    num_trees = BART_TESTS$small_model_num_trees,
    num_burn_in = BART_TESTS$small_model_burn_in,
    num_iterations_after_burn_in = BART_TESTS$small_model_iter,
    use_missing_data = TRUE,
    use_missing_data_dummies_as_covars = TRUE,
    verbose = BART_TESTS$verbose
  )
  expect_s3_class(model_dummies, "bartMachine")
  # Check if dummy columns were created (starts with M_)
  expect_true(any(grepl("^M_", model_dummies$training_data_features_with_missing_features)))
})

test_that("variable counts and proportions extraction works", {
  reg_model <- build_regression_model()
  
  # Check get_var_counts_over_chain
  counts_splits <- get_var_counts_over_chain(reg_model, type = "splits")
  expect_true(is.matrix(counts_splits))
  expect_equal(ncol(counts_splits), reg_model$p)
  expect_equal(nrow(counts_splits), reg_model$num_iterations_after_burn_in)
  
  counts_trees <- get_var_counts_over_chain(reg_model, type = "trees")
  expect_true(is.matrix(counts_trees))
  
  # Check get_var_props_over_chain
  props_splits <- get_var_props_over_chain(reg_model, type = "splits")
  expect_true(is.numeric(props_splits))
  expect_equal(length(props_splits), reg_model$p)
  expect_true(all(props_splits >= 0 & props_splits <= 1))
  
  props_trees <- get_var_props_over_chain(reg_model, type = "trees")
  expect_true(is.numeric(props_trees))
})

test_that("serialization works (save and load)", {
  # Build a model with serialize = TRUE
  set.seed(BART_TESTS$seed)
  X <- data.frame(matrix(runif(BART_TESTS$small_data_n * 2), ncol = 2))
  y <- X[, 1] + rnorm(BART_TESTS$small_data_n, sd = 0.1)
  
  model <- bartMachine(
    X, 
    y, 
    num_trees = BART_TESTS$small_model_num_trees,
    num_burn_in = BART_TESTS$small_model_burn_in,
    num_iterations_after_burn_in = BART_TESTS$small_model_iter,
    serialize = TRUE,
    verbose = BART_TESTS$verbose
  )
  
  # Save to temp file
  tmp_file <- tempfile(fileext = ".rds")
  saveRDS(model, tmp_file)
  
  # Load back
  loaded_model <- readRDS(tmp_file)
  expect_s3_class(loaded_model, "bartMachine")
  
  # Check if prediction works on loaded model
  preds <- predict(loaded_model, X, verbose = BART_TESTS$verbose)
  expect_length(preds, nrow(X))
  
  # Clean up
  unlink(tmp_file)
})

test_that("interaction constraints work", {
    set.seed(BART_TESTS$seed)
    n = 100
    p = 5
    X = data.frame(matrix(runif(n * p), ncol = p))
    y = 10 * X[ ,1] * X[,2] + X[,1] * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)
    
    #constrain so that x1 and x2 interact and x1 and x3 interact and x4 and x5 interact
    interaction_constraints = list(c(1, 2, 3), c(4, 5))
    
    #build BART regression model
    bart_machine = bartMachine(X, y, num_trees = 20, num_burn_in = 100, num_iterations_after_burn_in = 100, interaction_constraints = interaction_constraints, verbose = FALSE)
    expect_s3_class(bart_machine, "bartMachine")
    
    #check invalid constraints throw error
    interaction_constraints_invalid = list(c(1, 2, 3), c(4, 6)) #6 is out of bounds
    expect_error(bartMachine(X, y, num_trees = 20, interaction_constraints = interaction_constraints_invalid, verbose = FALSE))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{investigate_var_importance}
\alias{investigate_var_importance}
\title{Explore Variable Inclusion Proportions in BART Model}
\usage{
investigate_var_importance(
  bart_machine,
  type = "splits",
  plot = TRUE,
  num_replicates_for_avg = 5,
  num_trees_bottleneck = 20,
  num_var_plot = Inf,
  bottom_margin = 10,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{type}{If ``splits'', then the proportion of times each variable is chosen for a splitting rule is computed. If ``trees'', then the proportion of times each variable appears in a tree is computed.}

\item{plot}{If TRUE, a plot of the variable inclusion proportions is generated.}

\item{num_replicates_for_avg}{The number of replicates of BART to be used to generate variable inclusion proportions. Averaging across multiple BART models improves stability of the estimates. See Bleich et al. (2013) for more details.}

\item{num_trees_bottleneck}{Number of trees to be used in the sum-of-trees for computing the variable inclusion proportions. A small number of trees should be used to force the variables to compete for entry into the model. Chipman et al. (2010) recommend 20. See this reference for more details.}

\item{num_var_plot}{Number of variables to be shown on the plot. If ``Inf'', all variables are plotted.}

\item{bottom_margin}{A display parameter that adjusts the bottom margin of the graph if labels are clipped. The scale of this parameter is the same as set with \code{par(mar = c(....))} in R.
Higher values allow for more space if the covariate names are long. Note that making this parameter too large will prevent plotting and the plot function in R will throw an error.}

\item{verbose}{If TRUE, prints progress messages and plots to the active device.}
}
\value{
Invisibly, returns a list with the following components:
\item{avg_var_props}{The average variable inclusion proportions for each variable\cr (across \code{num_replicates_for_avg})}
\item{sd_var_props}{The standard deviation of the variable inclusion proportions for each variable (across \code{num_replicates_for_avg})}
}
\description{
Explore the variable inclusion proportions for a BART model to learn about the relative influence of the different covariates. This function includes an option to generate a plot of the variable inclusion proportions.
}
\details{
In the plot, the red bars correspond to the standard error of the variable inclusion proportion estimates.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 10
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y, num_trees = 20)

#investigate variable inclusion proportions
investigate_var_importance(bart_machine)
}
}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}

J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
Additive Regression Trees. ArXiv e-prints, 2013.

HA Chipman, EI George, and RE McCulloch. BART: Bayesian Additive Regressive Trees.
The Annals of Applied Statistics, 4(1): 266--298, 2010.
}
\seealso{
\code{\link{interaction_investigator}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{benchmark_datasets}
\alias{benchmark_datasets}
\alias{ankara}
\alias{baseball}
\alias{boston}
\alias{compactiv}
\alias{ozone}
\alias{pole}
\alias{triazine}
\alias{wine.red}
\alias{wine.white}
\title{benchmark_datasets}
\source{
See vignette for details.
}
\usage{
data(benchmark_datasets)
}
\description{
Nine diverse datasets which were used for benchmarking bartMachine's out of sample performance in
  the vignette for this package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_arrays.R
\name{bartMachineArr}
\alias{bartMachineArr}
\title{Create an array of BART models for the same data.}
\usage{
bartMachineArr(bart_machine, R = 10)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{R}{The number of replicated BART models in the array.}
}
\value{
A \code{bartMachineArr} object which is just a list of the \code{R} bartMachine models.
}
\description{
If BART creates models that are variable,
running many on the same dataset and averaging is a good strategy.
This function is a convenience method for this procedure.
}
\examples{
\dontrun{
#Regression example
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)
bart_machine_arr = bartMachineArr(bart_machine)

#Classification example
data(iris)
iris2 = iris[51 : 150, ] #do not include the third type of flower for this example
iris2$Species = factor(iris2$Species)
bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)
bart_machine_arr = bartMachineArr(bart_machine)
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_predicts.R
\name{calc_prediction_intervals}
\alias{calc_prediction_intervals}
\title{Calculate Prediction Intervals}
\usage{
calc_prediction_intervals(
  bart_machine,
  new_data,
  pi_conf = 0.95,
  num_samples_per_data_point = 1000
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{new_data}{A data frame containing observations at which prediction intervals for \eqn{\hat{y}} are to be computed.}

\item{pi_conf}{Confidence level for the prediction intervals. The default is 95\%.}

\item{num_samples_per_data_point}{The number of samples taken from the predictive distribution. The default is 1000.}
}
\value{
Returns a matrix of the lower and upper bounds of the prediction intervals for each observation in \code{new_data}.
}
\description{
Generates prediction intervals for \eqn{\hat{y}} for a specified set of observations.
}
\details{
Credible intervals (see \code{\link{calc_credible_intervals}}) are the appropriate quantiles of the prediction
for each of the Gibbs samples post-burn in. Prediction intervals also make use of the noise estimate at each Gibbs
sample and hence are wider. For each Gibbs sample, we record the \eqn{\hat{y}} estimate of the response and the
\eqn{\hat{\sigma^2}} estimate of the noise variance. We then sample \code{normal_samples_per_gibbs_sample} times
from a \eqn{N(\hat{y}, \hat{\sigma^2})} random variable to simulate many possible disturbances for that Gibbs sample.
Then, all \code{normal_samples_per_gibbs_sample} times the number of Gibbs sample post burn-in are collected and the
appropriate quantiles are taken based on the confidence level, \code{pi_conf}.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#get prediction interval
pred_int = calc_prediction_intervals(bart_machine, X)
print(head(pred_int))
}
}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
}
\seealso{
\code{\link{calc_credible_intervals}}, \code{\link{bart_machine_get_posterior}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_predicts.R
\name{bart_machine_get_posterior}
\alias{bart_machine_get_posterior}
\title{Get Full Posterior Distribution}
\usage{
bart_machine_get_posterior(bart_machine, new_data, verbose = TRUE)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{new_data}{A data frame containing observations at which draws from posterior distribution of \eqn{\hat{f}(x)} are to be obtained.}

\item{verbose}{If TRUE, prints preprocessing-related messages.}
}
\value{
Returns a list with the following components:
%%  If it is a LIST, use
  \item{y_hat}{Posterior mean estimates. For regression, the estimates have the same units as the response. For classification, the estimates are probabilities.}
  \item{new_data}{The data frame with rows at which the posterior draws are to be generated. Column names should match that of the training data.}
  \item{y_hat_posterior_samples}{The full set of posterior samples of size \code{num_iterations_after_burn_in} for each observation. For regression, the estimates have the same units as the response. For classification, the estimates are probabilities.}
%% ...
}
\description{
Generates draws from posterior distribution of \eqn{\hat{f}(x)} for a specified set of observations.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#Regression example

#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#get posterior distribution
posterior = bart_machine_get_posterior(bart_machine, X)
print(posterior$y_hat)


#Classification example

#get data and only use 2 factors
data(iris)
iris2 = iris[51:150,]
iris2$Species = factor(iris2$Species)

#build BART classification model
bart_machine = bartMachine(iris2[ ,1 : 4], iris2$Species)

#get posterior distribution
posterior = bart_machine_get_posterior(bart_machine, iris2[ ,1 : 4])
print(posterior$y_hat)
}
}
\seealso{
\code{\link{calc_credible_intervals}}, \code{\link{calc_prediction_intervals}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{plot_y_vs_yhat}
\alias{plot_y_vs_yhat}
\title{Plot the fitted Versus Actual Response}
\usage{
plot_y_vs_yhat(
  bart_machine,
  Xtest = NULL,
  ytest = NULL,
  credible_intervals = FALSE,
  prediction_intervals = FALSE,
  interval_confidence_level = 0.95,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{Xtest}{Optional argument for test data. If included, BART computes fitted values at the rows of \code{Xtest}. Else, the fitted values from the training data are used.}

\item{ytest}{Optional argument for test data. Vector of observed values corresponding to the rows of \code{Xtest} to be plotted against the predictions for the rows of \code{Xtest}.}

\item{credible_intervals}{If TRUE, Bayesian credible intervals are computed using the quantiles of the posterior distribution of \eqn{\hat{f}(x)}. See \code{\link{calc_credible_intervals}} for details.}

\item{prediction_intervals}{If TRUE, Bayesian predictive intervals are computed using the a draw of from \eqn{\hat{f}(x)}. See \code{\link{calc_prediction_intervals}} for details.}

\item{interval_confidence_level}{Desired level of confidence for credible or prediction intervals.}

\item{verbose}{If TRUE, prints plots to the active device.}
}
\value{
None.
}
\description{
Generates a plot actual versus fitted values and corresponding credible intervals or prediction intervals for the fitted values.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate linear data
set.seed(11)
n  = 500
p = 3
X = data.frame(matrix(runif(n * p), ncol = p))
y = 3*X[ ,1] + 2*X[ ,2] +X[ ,3] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

##generate plot
plot_y_vs_yhat(bart_machine)

#generate plot with prediction bands
plot_y_vs_yhat(bart_machine, prediction_intervals = TRUE)
}
}
\seealso{
\code{\link{bart_machine_get_posterior}}, \code{\link{calc_credible_intervals}}, \code{\link{calc_prediction_intervals}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{check_bart_error_assumptions}
\alias{check_bart_error_assumptions}
\title{Check BART Error Assumptions}
\usage{
check_bart_error_assumptions(
  bart_machine,
  hetero_plot = "yhats",
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{hetero_plot}{If ``yhats'', the residuals are plotted against the fitted values of the response. If ``ys'', the residuals are plotted against the actual values of the response.}

\item{verbose}{If TRUE, prints plots to the active device.}
}
\value{
None.
}
\description{
Diagnostic tools to assess whether the errors of the BART model for regression are normally distributed and homoskedastic, as assumed by the model. This function generates a normal quantile plot of the residuals with a Shapiro-Wilks p-value as well as a residual plot.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 300
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#check error diagnostics
check_bart_error_assumptions(bart_machine)
}
}
\seealso{
\code{\link{plot_convergence_diagnostics}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{rmse_by_num_trees}
\alias{rmse_by_num_trees}
\title{Assess the Out-of-sample RMSE by Number of Trees}
\usage{
rmse_by_num_trees(
  bart_machine,
  tree_list = c(5, seq(10, 50, 10), 100, 150, 200),
  in_sample = FALSE,
  plot = TRUE,
  holdout_pctg = 0.3,
  num_replicates = 4,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{tree_list}{List of sizes for the sum-of-trees models.}

\item{in_sample}{If TRUE, the RMSE is computed on in-sample data rather than an out-of-sample holdout.}

\item{plot}{If TRUE, a plot of the RMSE by the number of trees in the ensemble is created.}

\item{holdout_pctg}{Percentage of the data to be treated as an out-of-sample holdout.}

\item{num_replicates}{Number of replicates to average the results over. Each replicate uses a randomly sampled holdout of the data, (which could have overlap).}

\item{verbose}{If TRUE, prints progress messages and plots to the active device.}

\item{...}{Other arguments to be passed to the plot function.}
}
\value{
Invisibly, returns the out-of-sample average RMSEs for each tree size.
}
\description{
Assess out-of-sample RMSE of a BART model for varying numbers of trees in the sum-of-trees model.
}
\note{
Since using a large number of trees can substantially increase computation time, this plot can help assess whether a smaller ensemble size is sufficient to obtain desirable predictive performance.
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 10
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y, num_trees = 20)

#explore RMSE by number of trees
rmse_by_num_trees(bart_machine)
}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_variable_selection.R
\name{var_selection_by_permute_cv}
\alias{var_selection_by_permute_cv}
\title{Perform Variable Selection Using Cross-validation Procedure}
\usage{
var_selection_by_permute_cv(
  bart_machine,
  k_folds = 5,
  folds_vec = NULL,
  num_reps_for_avg = 5,
  num_permute_samples = 100,
  num_trees_for_permute = 20,
  alpha = 0.05,
  num_trees_pred_cv = 50,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{k_folds}{Number of folds to be used in cross-validation.}

\item{folds_vec}{An integer vector of indices specifying which fold each observation belongs to.}

\item{num_reps_for_avg}{Number of replicates to over over to for the BART model's variable inclusion proportions.}

\item{num_permute_samples}{Number of permutations of the response to be made to generate the ``null'' permutation distribution.}

\item{num_trees_for_permute}{Number of trees to use in the variable selection procedure. As with \cr \code{\link{investigate_var_importance}}, a small number of trees should be used to force variables to compete for entry into the model. Note that this number is used to estimate both the ``true'' and ``null'' variable inclusion proportions.}

\item{alpha}{Cut-off level for the thresholds.}

\item{num_trees_pred_cv}{Number of trees to use for prediction on the hold-out portion of each fold. Once variables have been selected using the training portion of each fold, a new model is built using only those variables with \code{num_trees_pred_cv} trees in the sum-of-trees model. Forecasts for the holdout sample are made using this model. A larger number of trees is recommended to exploit the full forecasting power of BART.}

\item{verbose}{If TRUE, prints progress messages.}
}
\value{
Returns a list with the following components:

  \item{best_method}{The name of the best variable selection procedure, as chosen via cross-validation.}
  \item{important_vars_cv}{The variables chosen by the \code{best_method} above.}
}
\description{
Performs variable selection by cross-validating over the three threshold-based procedures outlined in Bleich et al. (2013) and selecting the single procedure that returns the lowest cross-validation RMSE.
}
\details{
See Bleich et al. (2013) for a complete description of the procedures outlined above as well as the corresponding vignette for a brief summary with examples.
}
\note{
This function can have substantial run-time.
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 150
p = 100 ##95 useless predictors
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model (not actually used in variable selection)
bart_machine = bartMachine(X, y)

#variable selection via cross-validation
var_sel_cv = var_selection_by_permute_cv(bart_machine, k_folds = 3)
print(var_sel_cv$best_method)
print(var_sel_cv$important_vars_cv)
}
}
\references{
J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
Additive Regression Trees. ArXiv e-prints, 2013.

Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
}
\seealso{
\code{\link{var_selection_by_permute}}, \code{\link{investigate_var_importance}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_predicts.R
\name{predict.bartMachine}
\alias{predict.bartMachine}
\title{Make a prediction on data using a BART object}
\usage{
\method{predict}{bartMachine}(
  object,
  new_data,
  type = "prob",
  prob_rule_class = NULL,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{An object of class ``bartMachine''.}

\item{new_data}{A data frame where each row is an observation to predict. The column names
should be the same as the column names of the training data.}

\item{type}{Only relevant if the bartMachine model is classification. The type can be ``prob'' which will
return the estimate of \eqn{P(Y = 1)}(the ``positive'' class) or ``class'' which will return the best guess as to the
class of the object, in the original label, based on if the probability estimate is greater
than \code{prob_rule_class}. Default is ``prob.''}

\item{prob_rule_class}{The rule to determine when the class estimate is \eqn{Y = 1} (the ``positive'' class) based on the probability estimate. This
defaults to what was originally specified in the \code{bart_machine} object.}

\item{verbose}{Prints out prediction-related messages. Currently in use only for probability predictions to let the user know which class
is being predicted. Default is \code{TRUE}.}

\item{...}{Parameters that are ignored.}
}
\value{
If regression, a numeric vector of \code{y_hat}, the best guess as to the response. If classification and \code{type = ``prob''},
a numeric vector of \code{p_hat}, the best guess as to the probability of the response class being  the ''positive'' class. If classification and
\code{type = ''class''}, a character vector of the best guess of the response's class labels.
}
\description{
Makes a prediction on new data given a fitted BART model for regression or classification.
}
\examples{
#Regression example
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

##make predictions on the training data
y_hat = predict(bart_machine, X)

#Classification example
data(iris)
iris2 = iris[51 : 150, ] #do not include the third type of flower for this example
iris2$Species = factor(iris2$Species)
bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)

##make probability predictions on the training data
p_hat = predict(bart_machine, X)

##make class predictions on test data
y_hat_class = predict(bart_machine, X, type = "class")

##make class predictions on test data conservatively for ''versicolor''
y_hat_class_conservative = predict(bart_machine, X, type = "class", prob_rule_class = 0.9)
}
}
\seealso{
\code{\link{bart_predict_for_test_data}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{automobile}
\alias{automobile}
\title{Data concerning automobile prices.}
\source{
K Bache and M Lichman. UCI machine learning repository, 2013.
http://archive.ics.uci.edu/ml/datasets/Automobile
}
\usage{
data(automobile)
}
\description{
The \code{automobile} data frame has 201 rows and 25 columns and
  concerns automobiles in the 1985 Auto Imports Database. The response
  variable, \code{price}, is the log selling price of the automobile. There
  are 7 categorical predictors and 17 continuous / integer predictors which
  are features of the automobiles. 41 automobiles have missing data in one
  or more of the feature entries. This dataset is true to the original except
  with a few of the predictors dropped.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_cross_validation.R
\name{k_fold_cv}
\alias{k_fold_cv}
\title{Estimate Out-of-sample Error with K-fold Cross validation}
\usage{
k_fold_cv(X, y, k_folds = 5, folds_vec = NULL, verbose = FALSE, ...)
}
\arguments{
\item{X}{Data frame of predictors. Factors are automatically converted to dummies internally.}

\item{y}{Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.}

\item{k_folds}{Number of folds to cross-validate over. This argument is ignored if \code{folds_vec} is non-null.}

\item{folds_vec}{An integer vector of indices specifying which fold each observation belongs to.}

\item{verbose}{Prints information about progress of the algorithm to the screen.}

\item{\dots}{Additional arguments to be passed to \code{build_bart_machine}.}
}
\value{
For regression models, a list with the following components is returned:
  \item{y_hat}{Predictions for the observations computed on the fold for which the observation was omitted from the training set.}
  \item{L1_err}{Aggregate L1 error across the folds.}
  \item{L2_err}{Aggregate L1 error across the folds.}
  \item{rmse}{Aggregate RMSE across the folds.}
  \item{folds}{Vector of indices specifying which fold each observation belonged to.}

For classification models, a list with the following components is returned:

  \item{y_hat}{Class predictions for the observations computed on the fold for which the observation was omitted from the training set.}
    \item{p_hat}{Probability estimates for the observations computed on the fold for which the observation was omitted from the training set.}
  \item{confusion_matrix}{Aggregate confusion matrix across the folds.}
    \item{misclassification_error}{Total misclassification error across the folds.}
    \item{folds}{Vector of indices specifying which fold each observation belonged to.}
}
\description{
Builds a BART model using a specified set of arguments to \code{build_bart_machine} and estimates the out-of-sample performance by using k-fold cross validation.
}
\details{
For each fold, a new BART model is trained (using the same set of arguments) and its performance is evaluated on the holdout piece of that fold.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

#evaluate default BART on 5 folds
k_fold_val = k_fold_cv(X, y)
print(k_fold_val$rmse)
}
}
\seealso{
\code{\link{bartMachine}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_data_preprocessing.R
\name{dummify_data}
\alias{dummify_data}
\title{Dummify Design Matrix}
\usage{
dummify_data(data)
}
\arguments{
\item{data}{Data frame to be dummified.}
}
\value{
Returns a data frame with factors converted to dummy indicator variables.
}
\description{
Create a data frame with factors converted to dummies.
}
\details{
The column names of the dummy variables are given by the ``FactorName_LevelName'' and are augmented to the end of the design matrix. See the example below.
}
\note{
BART handles dummification internally. This function is provided as a utility function.
}
\examples{
\dontrun{
#generate data
set.seed(11)
x1 = rnorm(20)
x2 = as.factor(ifelse(x1 > 0, "A", "B"))
x3 = runif(20)
X = data.frame(x1,x2,x3)
#dummify data
X_dummified = dummify_data(X)
print(X_dummified)
}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_node_related_methods.R
\name{node_prediction_training_data_indices}
\alias{node_prediction_training_data_indices}
\title{Gets node predictions indices of the training data for new data.}
\usage{
node_prediction_training_data_indices(bart_machine, new_data = NULL)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{new_data}{Data that you wish to investigate the training sample weights. If \code{NULL}, the original training data is used.}
}
\value{
Returns a binary tensor indicating whether the prediction node contained a training datum or not. For each observation in new data, the size of this tensor is number of gibbs sample after burn-in
times the number of trees times the number of training data observations. This the size of the full tensor is the number of observations in the new data times the three dimensional object just explained.
}
\description{
This returns a binary tensor for all gibbs samples after burn-in for all trees and for all training observations.
}
\examples{
\dontrun{
set.seed(11)
n = 50
X = data.frame(x1 = rnorm(n), x2 = runif(n))
y = X$x1 + rnorm(n)
bart_machine = bartMachine(X, y, flush_indices_to_save_RAM = FALSE)
idx = node_prediction_training_data_indices(bart_machine)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{pd_plot}
\alias{pd_plot}
\title{Partial Dependence Plot}
\usage{
pd_plot(
  bart_machine,
  j,
  levs = c(0.05, seq(from = 0.1, to = 0.9, by = 0.1), 0.95),
  lower_ci = 0.025,
  upper_ci = 0.975,
  prop_data = 1,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{j}{The number or name of the column in the design matrix for which the partial dependence plot is to be created.}

\item{levs}{Quantiles at which the partial dependence function should be evaluated. Linear extrapolation is performed between these points.}

\item{lower_ci}{Lower limit for credible interval}

\item{upper_ci}{Upper limit for credible interval}

\item{prop_data}{The proportion of the training data to use. Default is 1. Use a lower proportion for speedier pd_plots. The closer to 1, the more resolution
the PD plot will have; the closer to 0, the lower but faster.}

\item{verbose}{If TRUE, prints progress messages and plots to the active device.}
}
\value{
Invisibly, returns a list with the following components:

  \item{x_j_quants}{Quantiles at which the partial dependence function is evaluated}
  \item{bart_avg_predictions_by_quantile_by_gibbs}{All samples of \eqn{\hat{f}(x)}}
  \item{bart_avg_predictions_by_quantile}{Posterior means for \eqn{\hat{f}(x)} at \code{x_j_quants}}
  \item{bart_avg_predictions_lower}{Lower bound of the desired confidence of the credible interval of \eqn{\hat{f}(x)}}
  \item{bart_avg_predictions_upper}{Upper bound of the desired confidence of the credible interval of \eqn{\hat{f}(x)}}
  \item{prop_data}{The proportion of the training data to use as specified when this function was executed}
%% ...
}
\description{
Creates a partial dependence plot for a BART model for regression or classification.
}
\details{
For regression models, the units on the y-axis are the same as the units of the response. For classification models, the units on the y-axis are probits.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#Regression example

#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#partial dependence plot for quadratic term
pd_plot(bart_machine, "X3")


#Classification example

#get data and only use 2 factors
data(iris)
iris2 = iris[51:150,]
iris2$Species = factor(iris2$Species)

#build BART classification model
bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)

#partial dependence plot
pd_plot(bart_machine, "Petal.Width")
}
}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}

HA Chipman, EI George, and RE McCulloch. BART: Bayesian Additive Regressive Trees.
The Annals of Applied Statistics, 4(1): 266--298, 2010.
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_f_tests.R
\name{cov_importance_test}
\alias{cov_importance_test}
\title{Importance Test for Covariate(s) of Interest}
\usage{
cov_importance_test(
  bart_machine,
  covariates = NULL,
  num_permutation_samples = 100,
  plot = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bart_machine''.}

\item{covariates}{A vector of names of covariates of interest to be tested for having an effect on the response. A value of NULL
indicates an omnibus test for all covariates having an effect on the response. If the name of a covariate is a factor,
the entire factor will be permuted. We do not recommend entering the names of factor covariate dummies.}

\item{num_permutation_samples}{The number of times to permute the covariates of interest and create a corresponding new BART model (see details).}

\item{plot}{If \code{TRUE}, this produces a histogram of the Pseudo-Rsq's / total misclassification error rates from
the \code{num_permutations} BART models created with the \code{covariates} permuted. The plot also illustrates
the observed Pseudo-Rsq's / total misclassification error rate from the original training data and indicates
the test's p-value.}

\item{verbose}{If TRUE, prints progress and summary messages.}
}
\value{
\item{permutation_samples_of_error}{A vector which records the error metric of the BART models with the covariates permuted (see details).}
\item{observed_error_estimate}{For regression, this is the Pseudo-Rsq on the original
training data set. For classification, this is the observed total misclassification error
on the original training data set.}
\item{pval}{The approximate p-value for this test (see details).
}
}
\description{
This function tests the null hypothesis \eqn{H_0}: These covariates of interest
do not affect the response under the assumptions of the BART
model.
}
\details{
To test the importance of a covariate or a set of covariates of interest on the response, this function generates
\code{num_permutations} BART models with the covariate(s) of interest permuted (differently each time).
On each run, a measure of fit is recorded. For regression, the metric is Pseudo-Rsq; for classification, it is
total misclassification error.\cr A
p-value can then be generated as follows. For regression, the p-value is the number of
permutation-sampled Pseudo-Rsq's greater than the observed Pseudo-Rsq divided by
\code{num_permutations + 1}. For classification, the p-value is the number of permutation-sampled
total misclassification errors less than the observed total misclassification error divided by \code{num_permutations + 1}.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
##regression example

##generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

##now test if X[, 1] affects Y nonparametrically under the BART model assumptions
cov_importance_test(bart_machine, covariates = c(1))
## note the plot and the printed p-value

}
}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_predicts.R
\name{calc_credible_intervals}
\alias{calc_credible_intervals}
\title{Calculate Credible Intervals}
\usage{
calc_credible_intervals(bart_machine, new_data, ci_conf = 0.95)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{new_data}{A data frame containing observations at which credible intervals for \eqn{\hat{f}(x)} are to be computed.}

\item{ci_conf}{Confidence level for the credible intervals. The default is 95\%.}
}
\value{
Returns a matrix of the lower and upper bounds of the credible intervals for each observation in \code{new_data}.
}
\description{
Generates credible intervals for \eqn{\hat{f}(x)} for a specified set of observations.
}
\details{
This interval is the appropriate quantiles  based on the confidence level, \code{ci_conf}, of the predictions
for each of the Gibbs samples post-burn in.
}
\note{
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#get credible interval
cred_int = calc_credible_intervals(bart_machine, X)
print(head(cred_int))
}
}
\seealso{
\code{\link{calc_prediction_intervals}}, \code{\link{bart_machine_get_posterior}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{plot_convergence_diagnostics}
\alias{plot_convergence_diagnostics}
\title{Plot Convergence Diagnostics}
\usage{
plot_convergence_diagnostics(
  bart_machine,
  plots = c("sigsqs", "mh_acceptance", "num_nodes", "tree_depths"),
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{plots}{The list of plots to be displayed. The four options are: "sigsqs", "mh_acceptance", "num_nodes", "tree_depths".}

\item{verbose}{If TRUE, prints plots to the active device.}
}
\value{
None.
}
\description{
A suite of plots to assess convergence diagnostics and features of the BART model.
}
\details{
The ``sigsqs'' option plots the posterior error variance estimates by the Gibbs sample number. This is a standard tool to assess convergence of MCMC algorithms. This option is not applicable to classification BART models.\cr
The ``mh_acceptance'' option plots the proportion of Metropolis-Hastings steps accepted for each Gibbs sample (number accepted divided by number of trees).\cr
The ``num_nodes'' option plots the average number of nodes across each tree in the sum-of-trees model by the Gibbs sample number (for post burn-in only). The blue line
is the average number of nodes over all trees.\cr
The ``tree_depths'' option plots the average tree depth across each tree in the sum-of-trees model by the Gibbs sample number (for post burn-in only). The blue line
is the average number of nodes over all trees.
}
\note{
The ``sigsqs'' plot separates the burn-in \eqn{\sigma^2}'s for the first core by post burn-in \eqn{\sigma^2}'s estimates for all cores by grey vertical lines.
The ``mh_acceptance'' plot separates burn-in from post-burn in by a grey vertical line. Post burn-in, the different core proportions plot in different colors.
The ``num_nodes'' plot separates different core estimates by vertical lines (post burn-in only).
The `tree_depths'' plot separates different core estimates by vertical lines (post burn-in only).
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#plot convergence diagnostics
plot_convergence_diagnostics(bart_machine)
}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{interaction_investigator}
\alias{interaction_investigator}
\title{Explore Pairwise Interactions in BART Model}
\usage{
interaction_investigator(
  bart_machine,
  plot = TRUE,
  num_replicates_for_avg = 5,
  num_trees_bottleneck = 20,
  num_var_plot = 50,
  cut_bottom = NULL,
  bottom_margin = 10,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{plot}{If TRUE, a plot of the pairwise interaction counts is generated.}

\item{num_replicates_for_avg}{The number of replicates of BART to be used to generate pairwise interaction inclusion counts.
Averaging across multiple BART models improves stability of the estimates.}

\item{num_trees_bottleneck}{Number of trees to be used in the sum-of-trees model for computing pairwise interactions counts.
A small number of trees should be used to force the variables to compete for entry into the model.}

\item{num_var_plot}{Number of variables to be shown on the plot. If ``Inf,'' all variables are plotted (not recommended if
the number of predictors is large). Default is 50.}

\item{cut_bottom}{A display parameter between 0 and 1 that controls where the y-axis is plotted. A value of 0 would begin the y-axis at 0; a value of 1 begins
the y-axis at the minimum of the average pairwise interaction inclusion count (the smallest bar in the bar plot). Values between 0 and 1 begin the
y-axis as a percentage of that minimum.}

\item{bottom_margin}{A display parameter that adjusts the bottom margin of the graph if labels are clipped. The scale of this parameter is the same as set with \code{par(mar = c(....))} in R.
Higher values allow for more space if the crossed covariate names are long. Note that making this parameter too large will prevent plotting and the plot function in R will throw an error.}

\item{verbose}{If TRUE, prints progress messages and plots to the active device.}
}
\value{
\item{interaction_counts}{For each of the \eqn{p \times p}{p times p} interactions, what is the count across all \code{num_replicates_for_avg}
BART model replicates' post burn-in Gibbs samples in all trees.}
	\item{interaction_counts_avg}{For each of the \eqn{p \times p}{p times p} interactions, what is the average count across all \code{num_replicates_for_avg}
BART model replicates' post burn-in Gibbs samples in all trees.}
	\item{interaction_counts_sd}{For each of the \eqn{p \times p}{p times p} interactions, what is the sd of the interaction counts across the \code{num_replicates_for_avg}
BART models replicates.}
	\item{interaction_counts_avg_and_sd_long}{For each of the \eqn{p \times p}{p times p} interactions, what is the average and sd of the interaction counts across the \code{num_replicates_for_avg}
BART models replicates. The output is organized as a convenient long table of class \code{data.frame}.}
}
\description{
Explore the pairwise interaction counts for a BART model to learn about interactions fit by the model. This function includes an option to generate a plot of the pairwise interaction counts.
}
\details{
An interaction between two variables is considered to occur whenever a path from any node of a tree to
any of its terminal node contains splits using those two variables. See Kapelner and Bleich, 2013, Section 4.11.
}
\note{
In the plot, the red bars correspond to the standard error of the variable inclusion proportion estimates (since multiple replicates were used).
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 10
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y, num_trees = 20)

#investigate interactions
interaction_investigator(bart_machine)
}
}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
}
\seealso{
\code{\link{investigate_var_importance}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_node_related_methods.R
\name{get_projection_weights}
\alias{get_projection_weights}
\title{Gets Training Sample Projection / Weights}
\usage{
get_projection_weights(
  bart_machine,
  new_data = NULL,
  regression_kludge = FALSE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{new_data}{Data that you wish to investigate the training sample projection / weights. If \code{NULL}, the original training data is used.}

\item{regression_kludge}{See explanation in the description. Default is \code{FALSE}.}
}
\value{
Returns a matrix of proportions with number of rows equal to the number of rows of \code{new_data} and number of columns equal to the number of rows of the original training data, n.
}
\description{
Returns the matrix H where yhat is approximately equal to H y where yhat is the predicted values for \code{new_data}. If \code{new_data} is unspecified, yhat will be the in-sample fits.
If BART was the same as OLS, H would be an orthogonal projection matrix. Here it is a projection matrix, but clearly non-orthogonal. Unfortunately, I cannot get
this function to work correctly because of three possible reasons (1) BART does not work by averaging tree predictions: it is a sum of trees model where each tree sees the residuals
via backfitting (2) the prediction in each node is a bayesian posterior draw which is close to ybar of the observations contained in the node if noise is gauged to be small and
(3) there are transformations of the original y variable. I believe I got close and I think I'm off by a constant multiple which is a function of the number of trees. I can
use regression to estimate the constant multiple and correct for it. Turn \code{regression_kludge} to \code{TRUE} for this. Note that the weights do not add up to one here.
The intuition is because due to the backfitting there is multiple counting. But I'm not entirely sure.
}
\examples{
\dontrun{
options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))
pacman::p_load(bartMachine, tidyverse)

seed = 1984
set.seed(seed)
n = 100
x = rnorm(n, 0, 1)
sigma = 0.1
y = x + rnorm(n, 0, sigma)

num_trees = 200
num_iterations_after_burn_in = 1000
bart_mod = bartMachine(data.frame(x = x), y,
	flush_indices_to_save_RAM = FALSE,
	num_trees = num_trees,
	num_iterations_after_burn_in = num_iterations_after_burn_in,
	seed = seed)
bart_mod

n_star = 100
x_star = rnorm(n_star)
y_star = as.numeric(x_star + rnorm(n_star, 0, sigma))
yhat_star_bart = predict(bart_mod, data.frame(x = x_star))

Hstar = get_projection_weights(bart_mod, data.frame(x = x_star))
rowSums(Hstar)
yhat_star_projection = as.numeric(Hstar \%*\% y)

ggplot(data.frame(
	yhat_star = yhat_star_bart,
	yhat_star_projection = yhat_star_projection,
	y_star = y_star)) +
  geom_point(aes(x = yhat_star_bart, y = yhat_star_projection), col = "green") +
  geom_abline(slope = 1, intercept = 0)

Hstar = get_projection_weights(bart_mod, data.frame(x = x_star), regression_kludge = TRUE)
rowSums(Hstar)
yhat_star_projection = as.numeric(Hstar \%*\% y)

ggplot(data.frame(
	yhat_star = yhat_star_bart,
	yhat_star_projection = yhat_star_projection,
	y_star = y_star)) +
  geom_point(aes(x = yhat_star_bart, y = yhat_star_projection), col = "green") +
  geom_abline(slope = 1, intercept = 0)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_predicts.R
\name{bart_predict_for_test_data}
\alias{bart_predict_for_test_data}
\title{Predict for Test Data with Known Outcomes}
\usage{
bart_predict_for_test_data(
  bart_machine,
  Xtest,
  ytest,
  prob_rule_class = NULL,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{Xtest}{Data frame for test data containing rows at which predictions are to be made. Colnames should match that of the training data.}

\item{ytest}{Actual outcomes for test data.}

\item{prob_rule_class}{Threshold for classification.}

\item{verbose}{If TRUE, prints prediction-related messages.}
}
\value{
For regression models, a list with the following components is returned:

  \item{y_hat}{Predictions (as posterior means) for the test observations.}
  \item{L1_err}{L1 error for predictions.}
  \item{L2_err}{L2 error for predictions.}
  \item{rmse}{RMSE for predictions.}

For classification models, a list with the following components is returned:

  \item{y_hat}{Class predictions for the test observations.}
  \item{p_hat}{Probability estimates for the test observations.}
  \item{confusion_matrix}{A confusion matrix for the test observations.}

%% ...
}
\description{
Utility wrapper function for computing out-of-sample metrics for a BART model when the test set outcomes are known.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 250
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##split into train and test
train_X = X[1 : 200, ]
test_X = X[201 : 250, ]
train_y = y[1 : 200]
test_y = y[201 : 250]

##build BART regression model
bart_machine = bartMachine(train_X, train_y)

#explore performance on test data
oos_perf = bart_predict_for_test_data(bart_machine, test_X, test_y)
print(oos_perf$rmse)
}
}
\seealso{
\code{\link{predict}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bartMachine.R, R/bart_package_builders.R
\name{bartMachineCV}
\alias{bartMachineCV}
\alias{build_bart_machine_cv}
\title{Build BART-CV}
\usage{
bartMachineCV(
  X = NULL,
  y = NULL,
  Xy = NULL,
  num_tree_cvs = c(50, 200),
  k_cvs = c(2, 3, 5),
  nu_q_cvs = NULL,
  k_folds = 5,
  folds_vec = NULL,
  use_xoshiro = FALSE,
  verbose = FALSE,
  ...
)

build_bart_machine_cv(
  X = NULL,
  y = NULL,
  Xy = NULL,
  num_tree_cvs = c(50, 200),
  k_cvs = c(2, 3, 5),
  nu_q_cvs = NULL,
  k_folds = 5,
  folds_vec = NULL,
  use_xoshiro = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{X}{Data frame of predictors. Factors are automatically converted to dummies internally.}

\item{y}{Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.}

\item{Xy}{A data frame of predictors and the response. The response column must be named ``y''.}

\item{num_tree_cvs}{Vector of sizes for the sum-of-trees models to cross-validate over.}

\item{k_cvs}{Vector of choices for the hyperparameter \code{k} to cross-validate over.}

\item{nu_q_cvs}{Only for regression. List of vectors containing (\code{nu}, \code{q}) ordered pair choices to cross-validate over. If \code{NULL}, then it defaults to the three values \code{list(c(3, 0.9), c(3, 0.99), c(10, 0.75))}.}

\item{k_folds}{Number of folds for cross-validation}

\item{folds_vec}{An integer vector of indices specifying which fold each observation belongs to.}

\item{use_xoshiro}{if TRUE, use the Xoshiro256PlusPlus random number generator; if FALSE, use the legacy MersenneTwister
random number generator (default is FALSE)}

\item{verbose}{Prints information about progress of the algorithm to the screen.}

\item{\dots}{Additional arguments to be passed to \code{bartMachine}.}
}
\value{
Returns an object of class ``bartMachine'' with the set of hyperparameters chosen via cross-validation. We also return a matrix ``cv_stats''
which contains the out-of-sample RMSE for each hyperparameter set tried and ``folds'' which gives the fold in which each observation fell across the k-folds.
}
\description{
Builds a BART-CV model by cross-validating over a grid of hyperparameter choices.
}
\note{
This function may require significant run-time.
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}} via calling \code{\link{bartMachine}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine_cv = bartMachineCV(X, y)

#information about cross-validated model
summary(bart_machine_cv)
}

}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
}
\seealso{
\code{\link{bartMachine}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bartMachine-package.R
\docType{package}
\name{bartMachine-package}
\alias{bartMachine-package}
\title{bartMachine: Bayesian Additive Regression Trees}
\description{
An advanced implementation of Bayesian Additive Regression Trees with expanded features for data analysis and visualization.
}
\seealso{
Useful links:
\itemize{
  \item Report bugs at \url{https://github.com/kapelner/bartMachine/issues}
}

}
\author{
\strong{Maintainer}: Adam Kapelner \email{kapelner@qc.cuny.edu} (\href{https://orcid.org/0000-0001-5985-6792}{ORCID})

Authors:
\itemize{
  \item Justin Bleich
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_variable_selection.R
\name{var_selection_by_permute}
\alias{var_selection_by_permute}
\title{Perform Variable Selection using Three Threshold-based Procedures}
\usage{
var_selection_by_permute(
  bart_machine,
  num_reps_for_avg = 10,
  num_permute_samples = 100,
  num_trees_for_permute = 20,
  alpha = 0.05,
  plot = TRUE,
  num_var_plot = Inf,
  bottom_margin = 10,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{num_reps_for_avg}{Number of replicates to over over to for the BART model's variable inclusion proportions.}

\item{num_permute_samples}{Number of permutations of the response to be made to generate the ``null'' permutation distribution.}

\item{num_trees_for_permute}{Number of trees to use in the variable selection procedure. As with \cr \code{\link{investigate_var_importance}}, a small number of trees should be used to force variables to compete for entry into the model. Note that this number is used to estimate both the ``true'' and ``null'' variable inclusion proportions.}

\item{alpha}{Cut-off level for the thresholds.}

\item{plot}{If TRUE, a plot showing which variables are selected by each of the procedures is generated.}

\item{num_var_plot}{Number of variables (in order of decreasing variable inclusion proportion) to be plotted.}

\item{bottom_margin}{A display parameter that adjusts the bottom margin of the graph if labels are clipped. The scale of this parameter is the same as set with \code{par(mar = c(....))} in R.
Higher values allow for more space if the crossed covariate names are long. Note that making this parameter too large will prevent plotting and the plot function in R will throw an error.}

\item{verbose}{If TRUE, prints progress messages.}
}
\value{
Invisibly, returns a list with the following components:

  \item{important_vars_local_names}{Names of the variables chosen by the Local procedure.}
  \item{important_vars_global_max_names}{Names of the variables chosen by the Global Max procedure.}
    \item{important_vars_global_se_names}{Names of the variables chosen by the Global SE procedure.}
    \item{important_vars_local_col_nums}{Column numbers of the variables chosen by the Local procedure.}
  \item{important_vars_global_max_col_nums}{Column numbers of the variables chosen by the Global Max procedure.}
    \item{important_vars_global_se_col_nums}{Column numbers of the variables chosen by the Global SE procedure.}
    \item{var_true_props_avg}{The variable inclusion proportions for the actual data.}
  \item{permute_mat}{The permutation distribution generated by permuting the response vector.}
}
\description{
Performs variable selection using the three thresholding methods introduced in Bleich et al. (2013).
}
\details{
See Bleich et al. (2013) for a complete description of the procedures outlined above as well as the corresponding vignette for a brief summary with examples.
}
\note{
Although the reference only explores regression settings, this procedure is applicable to both regression and classification problems.
This function is parallelized by the number of cores set in \code{\link{set_bart_machine_num_cores}}.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 300
p = 20 ##15 useless predictors
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model (not actually used in variable selection)
bart_machine = bartMachine(X, y)

#variable selection
var_sel = var_selection_by_permute(bart_machine)
print(var_sel$important_vars_local_names)
print(var_sel$important_vars_global_max_names)
}
}
\references{
J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
Additive Regression Trees. ArXiv e-prints, 2013.

Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}
}
\seealso{
\code{\link{var_selection_by_permute}}, \code{\link{investigate_var_importance}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_inits.R
\name{bart_machine_num_cores}
\alias{bart_machine_num_cores}
\title{Get Number of Cores Used by BART}
\usage{
bart_machine_num_cores()
}
\value{
%%  ~Describe the value returned
%%  If it is a LIST, use
Number of cores currently being used by parallelized BART functions.
%%  \item{comp2 }{Description of 'comp2'}
%% ...
}
\description{
Returns number of cores used by BART
}
\details{
Returns the number of cores currently being used by parallelized BART functions
}
\examples{
\dontrun{
bart_machine_num_cores()
}
}
\seealso{
\code{\link{set_bart_machine_num_cores}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_node_related_methods.R
\name{extract_raw_node_data}
\alias{extract_raw_node_data}
\title{Gets Raw Node data}
\usage{
extract_raw_node_data(bart_machine, g = 1)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{g}{The gibbs sample number. It must be a natural number between 1 and the number of iterations after burn in. Default is 1.}
}
\value{
Returns a list object that contains all the information for all trees in a given Gibbs sample.
}
\description{
Returns a list object that contains all the information for all trees in a given Gibbs sample. Daughter nodes are nested
in the list structure recursively.
}
\examples{
\dontrun{
options(java.parameters = c("-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"))
pacman::p_load(bartMachine)

seed = 1984
set.seed(seed)
n = 100
x = rnorm(n, 0, 1)
sigma = 0.1
y = x + rnorm(n, 0, sigma)

num_trees = 200
num_iterations_after_burn_in = 1000
bart_mod = bartMachine(data.frame(x = x), y,
	flush_indices_to_save_RAM = FALSE,
	num_trees = num_trees,
	num_iterations_after_burn_in = num_iterations_after_burn_in,
	seed = seed)

raw_node_data = extract_raw_node_data(bart_mod)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_inits.R
\name{get_var_props_over_chain}
\alias{get_var_props_over_chain}
\title{Get the Variable Inclusion Proportions}
\usage{
get_var_props_over_chain(bart_machine, type = "splits")
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{type}{If ``splits'', then the proportion of times each variable is chosen for a splitting rule versus all splitting rules is computed. If ``trees'', then the proportion of times each variable appears in a tree versus all appearances of variables in trees is computed.}
}
\value{
Returns a vector of the variable inclusion proportions.
}
\description{
Computes the variable inclusion proportions for a BART model.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 200
p = 10
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y, num_trees = 20)

#Get variable inclusion proportions
var_props = get_var_props_over_chain(bart_machine)
print(var_props)
}
}
\seealso{
\code{\link{get_var_counts_over_chain}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_arrays.R
\name{predict_bartMachineArr}
\alias{predict_bartMachineArr}
\title{Make a prediction on data using a BART array object}
\usage{
predict_bartMachineArr(object, new_data, ...)
}
\arguments{
\item{object}{An object of class ``bartMachineArr''.}

\item{new_data}{A data frame where each row is an observation to predict. The column names
should be the same as the column names of the training data.}

\item{...}{Not supported. Note that parameters \code{type} and \code{prob_rule_class} for
\code{\link{predict.bartMachine}} are not supported.}
}
\value{
If regression, a numeric vector of \code{y_hat}, the best guess as to the response. If classification and \code{type = ``prob''},
a numeric vector of \code{p_hat}, the best guess as to the probability of the response class being  the ''positive'' class. If classification and
\code{type = ''class''}, a character vector of the best guess of the response's class labels.
}
\description{
Makes a prediction on new data given an array of fitted BART model for
regression or classification. If BART creates models that are variable,
running many and averaging is a good strategy. It is well known that the
Gibbs sampler gets locked into local modes at times. This is a way
to average over many chains.
}
\examples{
\dontrun{
#Regression example
#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)
bart_machine_arr = bartMachineArr(bart_machine)

##make predictions on the training data
y_hat = predict(bart_machine_arr, X)

#Classification example
data(iris)
iris2 = iris[51 : 150, ] #do not include the third type of flower for this example
iris2$Species = factor(iris2$Species)
bart_machine = bartMachine(iris2[ ,1:4], iris2$Species)
bart_machine_arr = bartMachineArr(bart_machine)

##make probability predictions on the training data
p_hat = predict_bartMachineArr(bart_machine_arr, iris2[ ,1:4])
}
}
\seealso{
\code{\link{predict.bartMachine}}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_summaries.R
\name{print.bartMachine}
\alias{print.bartMachine}
\title{Summarizes information about a \code{bartMachine} object.}
\usage{
\method{print}{bartMachine}(x, verbose = TRUE, ...)
}
\arguments{
\item{x}{An object of class ``bartMachine''.}

\item{verbose}{If TRUE, prints summary output.}

\item{...}{Parameters that are ignored.}
}
\value{
None.
}
\description{
This is an alias for the \code{\link{summary.bartMachine}} function. See description in that section.
}
\examples{
\dontrun{
#Regression example

#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

##print out details
print(bart_machine)

##Also, the default print works too
bart_machine
}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_summaries.R
\name{summary.bartMachine}
\alias{summary.bartMachine}
\title{Summarizes information about a \code{bartMachine} object.}
\usage{
\method{summary}{bartMachine}(object, verbose = TRUE, ...)
}
\arguments{
\item{object}{An object of class ``bartMachine''.}

\item{verbose}{If TRUE, prints summary output.}

\item{...}{Parameters that are ignored.}
}
\value{
None.
}
\description{
Provides a quick summary of the BART model.
}
\details{
Gives the version number of the \code{bartMachine} package used to build this \code{additiveBartMachine} object and if the object
models either ``regression'' or ``classification.'' Gives the amount of training data and the dimension of feature space. Prints
the amount of time it took to build the model, how many processor cores were used to during its construction, as well as the
number of burn-in and posterior Gibbs samples were used.

If the model is for regression, it prints the estimate of \eqn{\sigma^2} before the model was constructed as well as after so
the user can inspect how much variance was explained.

If the model was built using the \code{run_in_sample = TRUE} parameter in \code{\link{build_bart_machine}} and is for regression, the summary L1,
L2, rmse, Pseudo-\eqn{R^2} are printed as well as the p-value for the tests of normality and zero-mean noise. If the model is for classification, a confusion matrix is printed.
}
\examples{
\dontrun{
#Regression example

#generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

##print out details
summary(bart_machine)

##Also, the default print works too
bart_machine
}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_f_tests.R
\name{linearity_test}
\alias{linearity_test}
\title{Test of Linearity}
\usage{
linearity_test(
  lin_mod = NULL,
  X = NULL,
  y = NULL,
  num_permutation_samples = 100,
  plot = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{lin_mod}{A linear model you can pass in if you do not want to use the default which is \code{lm(y ~ X)}. Default is \code{NULL} which should be used if you pass in \code{X} and \code{y}.}

\item{X}{Data frame of predictors. Factors are automatically converted to dummies internally. Default is \code{NULL} which should be used if you pass in \code{lin_mode}.}

\item{y}{Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.
Default is \code{NULL} which should be used if you pass in \code{lin_mode}.}

\item{num_permutation_samples}{This function relies on \code{\link{cov_importance_test}} (see documentation there for details).}

\item{plot}{This function relies on \code{\link{cov_importance_test}} (see documentation there for details).}

\item{verbose}{If TRUE, prints progress and summary messages.}

\item{...}{Additional parameters to be passed to \code{bartMachine}, the model constructed on the residuals of the linear model.}
}
\value{
\item{permutation_samples_of_error}{	This function relies on \code{\link{cov_importance_test}} (see documentation there for details).
}
\item{observed_error_estimate}{	This function relies on \code{\link{cov_importance_test}} (see documentation there for details).
}
\item{pval}{The approximate p-value for this test. See the documentation at \code{\link{cov_importance_test}}.
}
}
\description{
Test to investigate \eqn{H_0:} the functional relationship between the response and the
regressors is linear. We fit a linear model and then test if the residuals are a function
of the regressors using the
}
\examples{
\dontrun{
##regression example

##generate Friedman data i.e. a nonlinear response model
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##now test if there is a nonlinear relationship between X1, ..., X5 and y.
linearity_test(X = X, y = y)
## note the plot and the printed p-value.. should be approx 0

#generate a linear response model
y = 1 * X[ ,1] + 3 * X[,2] + 5 * X[,3] + 7 * X[ ,4] + 9 * X[,5] + rnorm(n)
linearity_test(X = X, y = y)
## note the plot and the printed p-value.. should be > 0.05

}
}
\seealso{
\code{\link{cov_importance_test}}
}
\author{
Adam Kapelner
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_plots.R
\name{get_sigsqs}
\alias{get_sigsqs}
\title{Get Posterior Error Variance Estimates}
\usage{
get_sigsqs(
  bart_machine,
  after_burn_in = TRUE,
  plot_hist = FALSE,
  plot_CI = 0.95,
  plot_sigma = F,
  verbose = TRUE
)
}
\arguments{
\item{bart_machine}{An object of class ``bartMachine''.}

\item{after_burn_in}{If TRUE, only the \eqn{\sigma^2} draws after the burn-in period are returned.}

\item{plot_hist}{If TRUE, a histogram of the posterior \eqn{\sigma^2} draws is generated.}

\item{plot_CI}{Confidence level for credible interval on histogram.}

\item{plot_sigma}{If TRUE, plots \eqn{\sigma} instead of \eqn{\sigma^2}.}

\item{verbose}{If TRUE, prints plots to the active device.}
}
\value{
Returns a vector of posterior \eqn{\sigma^2} draws (with or without the burn-in samples).
}
\description{
Returns the posterior estimates of the error variance from the Gibbs samples with an option to create a histogram of the posterior estimates of the error variance  with a credible interval overlaid.
}
\examples{
\dontrun{
#generate Friedman data
set.seed(11)
n  = 300
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)

#get posterior sigma^2's after burn-in and plot
sigsqs = get_sigsqs(bart_machine, plot_hist = TRUE)
}
}
\seealso{
\code{\link{get_sigsqs}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bartMachine.R, R/bart_package_builders.R
\name{bartMachine}
\alias{bartMachine}
\alias{build_bart_machine}
\title{Build a BART Model}
\usage{
bartMachine(
  X = NULL,
  y = NULL,
  Xy = NULL,
  num_trees = 50,
  num_burn_in = 250,
  num_iterations_after_burn_in = 1000,
  alpha = 0.95,
  beta = 2,
  k = 2,
  q = 0.9,
  nu = 3,
  prob_rule_class = 0.5,
  mh_prob_steps = c(2.5, 2.5, 4)/9,
  debug_log = FALSE,
  run_in_sample = TRUE,
  s_sq_y = "mse",
  sig_sq_est = NULL,
  print_tree_illustrations = FALSE,
  cov_prior_vec = NULL,
  interaction_constraints = NULL,
  use_missing_data = FALSE,
  covariates_to_permute = NULL,
  num_rand_samps_in_library = 10000,
  use_missing_data_dummies_as_covars = FALSE,
  replace_missing_data_with_x_j_bar = FALSE,
  impute_missingness_with_rf_impute = FALSE,
  impute_missingness_with_x_j_bar_for_lm = TRUE,
  mem_cache_for_speed = TRUE,
  flush_indices_to_save_RAM = TRUE,
  serialize = FALSE,
  seed = NULL,
  use_xoshiro = FALSE,
  verbose = TRUE
)

build_bart_machine(
  X = NULL,
  y = NULL,
  Xy = NULL,
  num_trees = 50,
  num_burn_in = 250,
  num_iterations_after_burn_in = 1000,
  alpha = 0.95,
  beta = 2,
  k = 2,
  q = 0.9,
  nu = 3,
  prob_rule_class = 0.5,
  mh_prob_steps = c(2.5, 2.5, 4)/9,
  debug_log = FALSE,
  run_in_sample = TRUE,
  s_sq_y = "mse",
  sig_sq_est = NULL,
  print_tree_illustrations = FALSE,
  cov_prior_vec = NULL,
  interaction_constraints = NULL,
  use_missing_data = FALSE,
  covariates_to_permute = NULL,
  num_rand_samps_in_library = 10000,
  use_missing_data_dummies_as_covars = FALSE,
  replace_missing_data_with_x_j_bar = FALSE,
  impute_missingness_with_rf_impute = FALSE,
  impute_missingness_with_x_j_bar_for_lm = TRUE,
  mem_cache_for_speed = TRUE,
  flush_indices_to_save_RAM = TRUE,
  serialize = FALSE,
  seed = NULL,
  use_xoshiro = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{X}{Data frame of predictors. Factors are automatically converted to dummies internally.}

\item{y}{Vector of response variable. If \code{y} is \code{numeric} or \code{integer}, a BART model for regression is built. If \code{y} is a factor with two levels, a BART model for classification is built.}

\item{Xy}{A data frame of predictors and the response. The response column must be named ``y''.}

\item{num_trees}{The number of trees to be grown in the sum-of-trees model.}

\item{num_burn_in}{Number of MCMC samples to be discarded as ``burn-in''.}

\item{num_iterations_after_burn_in}{Number of MCMC samples to draw from the posterior distribution of \eqn{\hat{f}(x)}.}

\item{alpha}{Base hyperparameter in tree prior for whether a node is nonterminal or not.}

\item{beta}{Power hyperparameter in tree prior for whether a node is nonterminal or not.}

\item{k}{For regression, \code{k} determines the prior probability that \eqn{E(Y|X)} is contained in the interval \eqn{(y_{min}, y_{max})}, based on a normal distribution. For example, when \eqn{k=2}, the prior probability is 95\%. For classification, \code{k} determines the prior probability that \eqn{E(Y|X)} is between \eqn{(-3,3)}. Note that a larger value of \code{k} results in more shrinkage and a more conservative fit.}

\item{q}{Quantile of the prior on the error variance at which the data-based estimate is placed. Note that the larger the value of \code{q}, the more aggressive the fit as you are placing more prior weight on values lower than the data-based estimate. Not used for classification.}

\item{nu}{Degrees of freedom for the inverse \eqn{\chi^2} prior. Not used for classification.}

\item{prob_rule_class}{Threshold for classification. Any observation with a conditional probability greater than \code{prob_class_rule} is assigned the ``positive'' outcome. Note that the first level of the response is treated as the ``positive'' outcome and the second is treated as the ``negative'' outcome.}

\item{mh_prob_steps}{Vector of prior probabilities for proposing changes to the tree structures: (GROW, PRUNE, CHANGE)}

\item{debug_log}{If TRUE, additional information about the model construction are printed to a file in the working directory.}

\item{run_in_sample}{If TRUE, in-sample statistics such as \eqn{\hat{f}(x)}, Pseudo-\eqn{R^2}, and RMSE are computed. Setting this to FALSE when not needed can decrease computation time.}

\item{s_sq_y}{If ``mse'', a data-based estimated of the error variance is computed as the MSE from ordinary least squares regression. If ``var''., the data-based estimate is computed as the variance of the response. Not used in classification.}

\item{sig_sq_est}{Pass in an estimate of the maximum sig_sq of the model. This is useful to cache somewhere and then pass in during cross-validation since the default method of estimation is a linear model. In large dimensions, linear model estimation is slow.}

\item{print_tree_illustrations}{For every Gibbs iteration, print out an illustration of the trees side-by-side. This is excruciatingly SLOW!}

\item{cov_prior_vec}{Vector assigning relative weights to how often a particular variable should be proposed as a candidate for a split. The vector is internally normalized so that the weights sum to 1. Note that the length of this vector must equal the length of the design matrix after dummification and augmentation of indicators of missingness (if used). To see what the dummified matrix looks like, use \code{\link{dummify_data}}. See Bleich et al. (2013) for more details on when this feature is most appropriate.}

\item{interaction_constraints}{A list of vectors indicating where the vectors are sets of elements allowed to interact with one another. The elements in each
vector correspond to features in the data frame \code{X} specified by either the column number as a numeric value or the column
name as a string e.g. \code{list(c(1, 2), c("nox", "rm"))}. The elements of the vectors can be reused among components for any
level of interaction complexity you wish. Default is \code{NULL} which corresponds to the vanilla modeling procedure where
all interactions are legal. For a pure generalized added model, use \code{as.list(seq(1 : p))} where \code{p}
is the number of columns in the design matrix \code{X}.}

\item{use_missing_data}{If TRUE, the missing data feature is used to automatically handle missing data without imputation. See Kapelner and Bleich (2013) for details.}

\item{covariates_to_permute}{Private argument for \code{\link{cov_importance_test}}. Not needed by user.}

\item{num_rand_samps_in_library}{Before building a BART model, samples from the Standard Normal and \eqn{\chi^2(\nu)} are drawn to be used in the MCMC steps. This parameter determines the number of samples to be taken.}

\item{use_missing_data_dummies_as_covars}{If TRUE, additional indicator variables for whether or not an observation in a particular column is missing are included. See Kapelner and Bleich (2013) for details.}

\item{replace_missing_data_with_x_j_bar}{If TRUE ,missing entries in \code{X} are imputed with average value or modal category.}

\item{impute_missingness_with_rf_impute}{If TRUE, missing entries are filled in using the rf.impute() function from the \code{randomForest} library.}

\item{impute_missingness_with_x_j_bar_for_lm}{If TRUE, when computing the data-based estimate of \eqn{\sigma^2}, missing entries are imputed with average value or modal category.}

\item{mem_cache_for_speed}{Speed enhancement that caches the predictors and the split values that are available at each node for selecting new rules. If the number
of predictors is large, the memory requirements become large. We recommend keeping this on (default) and turning it off if you experience out-of-memory errors.}

\item{flush_indices_to_save_RAM}{Setting this flag to \code{TRUE} saves memory with the downside that you cannot use the functions \code{node_prediction_training_data_indices} nor \code{get_projection_weights}.}

\item{serialize}{Setting this option to \code{TRUE} will allow serialization of bartMachine objects which allows for persistence between
R sessions if the object is saved and reloaded. Note that serialized objects can take up a large amount of memory.
Thus, the default is \code{FALSE}.}

\item{seed}{Optional: sets the seed in both R and Java. Default is \code{NULL} which does not set the seed in R nor Java.
Setting the seed enforces deterministic behavior only in the case when one core is used (the default before
\code{set_bart_machine_num_cores() was invoked}.}

\item{use_xoshiro}{if TRUE, use the Xoshiro256PlusPlus random number generator; if FALSE, use the legacy MersenneTwister
random number generator (default is FALSE)}

\item{verbose}{Prints information about progress of the algorithm to the screen.}
}
\value{
Returns an object of class ``bartMachine''. The ``bartMachine'' object contains a list of the following components:

  \item{java_bart_machine}{A pointer to the BART Java object.}
  \item{train_data_features}{The names of the variables used in the training data.}
  \item{training_data_features_with_missing_features.}{The names of the variables used in the training data. If \code{use_missing_data_dummies_as_covars = TRUE}, this also includes dummies for any predictors that contain at least one missing entry (named ``M_<feature>'').}
  \item{y}{The values of the response for the training data.}
  \item{y_levels}{The levels of the response (for classification only).}
  \item{pred_type}{Whether the model was build for regression of classification.}
  \item{model_matrix_training_data}{The training data with factors converted to dummies.}
  \item{num_cores}{The number of cores used to build the BART model.}
  \item{sig_sq_est}{The data-based estimate of \eqn{\sigma^2} used to create the prior on the error variance for the BART model.}
  \item{time_to_build}{Total time to build the BART model.}
  \item{y_hat_train}{The posterior means of \eqn{\hat{f}(x)} for each observation. Only returned if \code{run_in_sample = TRUE}.}
  \item{residuals}{The model residuals given by \code{y} - \code{y_hat_train}. Only returned if \code{run_in_sample = TRUE}.}
  \item{L1_err_train}{L1 error on the training set. Only returned if \code{run_in_sample = TRUE}.}
  \item{L2_err_train}{L2 error on the training set. Only returned if \code{run_in_sample = TRUE}.}
  \item{PseudoRsq}{Calculated as 1 - SSE / SST where SSE is the sum of square errors in the training data and SST is the sample variance of the response times \eqn{n-1}. Only returned if \code{run_in_sample = TRUE}.}
  \item{rmse_train}{Root mean square error on the training set. Only returned if \code{run_in_sample = TRUE}.}

Additionally, the parameters passed to the function \code{bartMachine} are also components of the list.
}
\description{
Builds a BART model for regression or classification.
}
\note{
This function is parallelized by the number of cores set by \code{\link{set_bart_machine_num_cores}}. Each core will create an
independent MCMC chain of size \cr
\code{num_burn_in} \eqn{+} \code{num_iterations_after_burn_in / bart_machine_num_cores}.
}
\examples{
\dontrun{
##regression example

##generate Friedman data
set.seed(11)
n  = 200
p = 5
X = data.frame(matrix(runif(n * p), ncol = p))
y = 10 * sin(pi* X[ ,1] * X[,2]) +20 * (X[,3] -.5)^2 + 10 * X[ ,4] + 5 * X[,5] + rnorm(n)

##build BART regression model
bart_machine = bartMachine(X, y)
summary(bart_machine)

##Build another BART regression model
bart_machine = bartMachine(X,y, num_trees = 200, num_burn_in = 500,
num_iterations_after_burn_in = 1000)

##Classification example

#get data and only use 2 factors
data(iris)
iris2 = iris[51:150,]
iris2$Species = factor(iris2$Species)

#build BART classification model
bart_machine = build_bart_machine(iris2[ ,1:4], iris2$Species)

##get estimated probabilities
phat = bart_machine$p_hat_train
##look at in-sample confusion matrix
bart_machine$confusion_matrix
}

}
\references{
Adam Kapelner, Justin Bleich (2016). bartMachine: Machine Learning
with Bayesian Additive Regression Trees. Journal of Statistical
Software, 70(4), 1-40. \doi{10.18637/jss.v070.i04}

HA Chipman, EI George, and RE McCulloch. BART: Bayesian Additive Regressive Trees.
The Annals of Applied Statistics, 4(1): 266--298, 2010.

A Kapelner and J Bleich. Prediction with Missing Data via Bayesian Additive Regression
Trees. Canadian Journal of Statistics, 43(2): 224-239, 2015

J Bleich, A Kapelner, ST Jensen, and EI George. Variable Selection Inference for Bayesian
Additive Regression Trees. ArXiv e-prints, 2013.
}
\seealso{
\code{\link{bartMachineCV}}
}
\author{
Adam Kapelner and Justin Bleich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart_package_inits.R
\name{set_bart_machine_num_cores}
\alias{set_bart_machine_num_cores}
\title{Set the Number of Cores for BART}
\usage{
set_bart_machine_num_cores(num_cores, verbose = TRUE)
}
\arguments{
\item{num_cores}{Number of cores to use. If the number of cores is more than 1, setting the seed during model construction
cannot be deterministic.}

\item{verbose}{If TRUE, prints the updated core count.}
}
\value{
None.
}
\description{
Sets the number of cores to be used for all parallelized BART functions.
}
\examples{
\dontrun{
#set all parallelized functions to use 4 cores
set_bart_machine_num_cores(4)
}
}
\seealso{
\code{\link{bart_machine_num_cores}}
}
\author{
Adam Kapelner and Justin Bleich
}
